import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;

/**
 * A shape representing an arc.
 * 
 * @author Tom Craig
 */
public class SArc extends SShape implements Cloneable, Serializable
{
	/**
	 * The centre of the arc.
	 */
	public SPoint centre;

	/**
	 * One of the arc's endpoints.
	 */
	public SPoint start;

	/**
	 * The other of the arc's endpoints.
	 */
	public SPoint end;

	/**
	 * Direction of the arc. (It could be the minor or the major segment.)
	 */
	private boolean anticlockwise;
	
	/**
	 * Returns the distance of the given position from the arc.
	*/
	public double minDistanceFromCoords(SIntPair position) {		
		//first we calculate the distance from centre
		int cdx = position.x - centre.getX().value();
		int cdy = position.y - centre.getY().value();
		double cd = java.lang.Math.sqrt((double) (cdx*cdx + cdy*cdy));
		
		//then calculate the arc's radius
		int rx = start.getX().value() - centre.getX().value();
		int ry = start.getY().value() - centre.getY().value();
		double rd = java.lang.Math.sqrt((double) (rx*rx + ry*ry));
		
		//calculating the difference results in the radial distance
		double radDist = java.lang.Math.abs(cd - rd);
		
		//now we need to calculate three angles:
		//the angle of the start point, of the end point, and of the "position" point
		//all relative to centre
		double posAngle = java.lang.Math.atan2(position.y - centre.getY().value(), position.x - centre.getX().value());
		double startAngle = java.lang.Math.atan2(start.getY().value() - centre.getY().value(), start.getX().value() - centre.getX().value());
		double endAngle = java.lang.Math.atan2(end.getY().value() - centre.getY().value(), end.getX().value() - centre.getX().value());
		
		double arcDist;
		
		if (anticlockwise) {
		
			//we set startAngle to be at 0, and make sure the other 2 values are positive.
			posAngle = posAngle - startAngle;
			endAngle = endAngle - startAngle;
			
			if (posAngle < 0) {
				posAngle += 2*Math.PI;
			}
			if (endAngle <= 0) {
				endAngle += 2*Math.PI;
			}
			
			//see if the position is inside the arc's angle
			if (posAngle <= endAngle) {
				arcDist = 0;
			} else {
				arcDist = posAngle - endAngle;
			}
			
		} else {
		
			//we set endAngle to be at 0, and make sure the other 2 values are positive.
			posAngle = posAngle - endAngle;
			startAngle = startAngle - endAngle;
			
			if (posAngle < 0) {
				posAngle += 2*Math.PI;
			}
			if (startAngle <= 0) {
				startAngle += 2*Math.PI;
			}
			
			//see if the position is inside the arc's angle
			if (posAngle <= startAngle) {
				arcDist = 0;
			} else {
				arcDist = posAngle - startAngle;
			}
			
		}
		
		//if the arc distance is larger than , subtract it from 2
		if (arcDist > Math.PI) {
			arcDist = 2*Math.PI - arcDist;
		}
		
		//multiplying this value with the distance from centre produces the arc distance in units
		double unitArcDist = arcDist * cd;
		
		//System.out.println(unitArcDist);
		
		//at which point we resort to ye good olde pythagoras
		return java.lang.Math.sqrt(unitArcDist*unitArcDist + radDist*radDist);
		//return radDist;
		
	}
	
    /**
     * @see SShape#getBoundingBox(SIntPair, SIntPair)
     */
    public void getBoundingBox(SIntPair topLeft, SIntPair bottomRight)
    {
    	// OK, for the time being, let's just return the bounding box of the circle.
    	// This will be the same size as, or larger than, the bounding box of the arc.
    	double radius = SIntPair.cartesianDistance(centre.getLocation().makeIntPair(), start.getLocation().makeIntPair());
    	topLeft.x = centre.getX().value() - (int) (radius + 0.5);
    	topLeft.y = centre.getY().value() - (int) (radius + 0.5);
    	bottomRight.x = centre.getX().value() + (int) (radius + 0.5);
    	bottomRight.y = centre.getY().value() + (int) (radius + 0.5);
    	
    	// the code below is for a proper bounding box (i.e. of the arc), but it's not done yet
    	return;
    
        // sq := quadrant that start point is in
        // eq := quadrant that end point is in
        /*int sq = SArc.getQuadrant(centre, start);
        int eq = SArc.getQuadrant(centre, end);*/
        
        /**
         * xmax = max(start's x, end's x)
         * ymax = max(start's y, end's y)
         * xleft = centre's x - radius
         * xright = centre's x + radius
         * ytop = centre's y - radius
		 * ybot = centre's y + radius
         * 
         * Start End Dirn     tlx     tly     brx     bry
         * ----------------------------------------------
         *     1   1   AC xleft ytop xright ybot
         *     1   2   AC
         *     1   3   AC
         *     1   4   AC
         *     2   1   AC
         *     2   2   AC
         *     2   3   AC
         *     2   4   AC
         *     3   1   AC
         *     3   2   AC
         *     3   3   AC
         *     3   4   AC
         *     4   1   AC
         *     4   2   AC
         *     4   3   AC
         *     4   4   AC
         * 
         * And for clockwise, simply reverse start and end.
         */
    }
    
	/**
	 * Default constructor.
	 */
	public SArc(SPoint centre, SPoint start, SPoint end, boolean anticlockwise, SFile owner)
	{
	          // for the time being, let the arc's handle be its start point
	          // TODO: change this to sthg more sensible
	    super(start.getLocation(), owner);
	    
	    this.centre = centre;
	    this.start = start;
	    this.end = end;
	    this.anticlockwise = anticlockwise;
	}
	
	/**
	 * @return A list of co-ordinates occupied by this shape.
	 */
	public Collection getPoints()
	{
	    ArrayList pts = new ArrayList();
	    pts.add(centre);
	    pts.add(start);
	    pts.add(end);
	    return pts;
	}
	
    /**
     * @return Whether the arc goes anticlockwise.
     */
    public boolean isAnticlockwise()
    {
        return anticlockwise;
    }
    
    /**
     * @param anticlockwise Whether arc goes anticlockwise.
     */
    public void setAnticlockwise(boolean anticlockwise)
    {
        this.anticlockwise = anticlockwise;
    }
    
    /**
     * @return The centre point of the arc.
     */
    public SPoint getCentre()
    {
        return centre;
    }
    
    /**
     * @param centre The new centre point of the arc.
     */
    public void setCentre(SPoint centre)
    {
        this.centre = centre;
    }
    
    /**
	 * @return A list of co-ordinates occupied by this shape.
	 */
	public Collection getCoordinateRange()
	{
	    ArrayList coordRange = new ArrayList();
	    
	    // r = radius of arc
	    double r = centre.getLocation().distanceBetweenCoords(start.getLocation());
	    // dt = largest value to give a guaranteed one point difference between steps around the arc
	    double dt = 1.0 / r;
	    // end1 and end2 are endpoints such that the arc is from end1 to end2
	    SPoint end1 = (anticlockwise ? start : end);
	    SPoint end2 = (anticlockwise ? end : start);
	    // t1 is the angle of end1, and t2 the angle of end2
	    double t1 = Math.acos((end1.getX().value() - centre.getX().value()) / r);
	    double t2 = Math.acos((end2.getX().value() - centre.getX().value()) / r);
	    // Dt is the difference between t1 and t2, taking into account which segment we're drawing (major or minor)
	    double Dt = Math.abs(t1 >= t2 ? 2 * Math.PI - t1 + t2
	            					  : t1 - t2);
	    // work out the arc length
	    double arcLength = r * Dt;
	    
	    // now go around the arc, adding points to coordRange
	    //D System.out.println("Start of arc coord range:");
	    for (int i = 0; i * dt <= Dt; i++)
	    {
	        double t = t1 + i * dt;
	        
	        SIntPair newCoord = new SIntPair((int) (centre.getX().value() + r * Math.cos(t)),
											 (int) (centre.getY().value() + r * Math.sin(t)));
	        coordRange.add(newCoord);
	        
	        //D newCoord.print();
	        //D System.out.println();
	    }
	    //D System.out.println("End of arc coord range");
	    
	    return coordRange;
	}
	
    /**
     * @return Returns the end point.
     */
    public SPoint getEnd()
    {
        return end;
    }
    
    /**
     * @param end The new end point.
     */
    public void setEnd(SPoint end)
    {
        this.end = end;
    }
    
    /**
     * @return Returns the start point.
     */
    public SPoint getStart()
    {
        return start;
    }
    
    /**
     * @param start The new start point.
     */
    public void setStart(SPoint start)
    {
        this.start = start;
    }
    
    /**
     * @return object reference for cloned object
     * @author Amit
     */
    public Object clone()
    {
     	SArc cln = null;
    	
    	cln = (SArc) super.clone();
    	    	
    	cln.centre = (SPoint) this.centre.clone();
    	
    	cln.start = (SPoint) this.start.clone();
    	
    	cln.end = (SPoint) this.end.clone();
    	
    	return cln;
    	
    }
    
    /**
     * @author Matthew
     */
    public void replacePoints(SPoint toReplace, SPoint replacement)
    {
    	if (toReplace == centre) centre = replacement;
    	if (toReplace == start) start = replacement;
    	if (toReplace == end) end = replacement;
    }
    
    /**
     * @param centre The centre of the circle
     * @param p The point on the circle for which we wish to find out the quadrant.
     * @return The quadrant p is in, in the range 1..4.
     * @author Tom C
     */
    public static int getQuadrant(SPoint centre, SPoint p)
    {
        int cx = centre.getX().value();
        int cy = centre.getY().value();
        int px = p.getX().value();
        int py = p.getY().value();
        
        if (px >= cx && py <= cy) return 1;
        if (px <= cx && py <= cy) return 2;
        if (px <= cx && py >= cy) return 3;
        if (px >= cx && py >= cy) return 4;
        
        // we should never get here!!!
        return 0;
    }
     
}