/**
 * This abstract class defines the common interfaces and functionality for a constraint in Sketchpad.
 * A constraint is a statement about a number of shapes or variables that the constraint solver tries to enforce.
 * One example would be SCParallel, which states that two lines must be parallel.
 * Most constraint functionality is in this class, so subclassing it to make a specific constraint requires little code.
 * Note on terminology: A variable is called "referred" if it is at all used in the calculation, but "bound" only if it can be changed by the constraint.
 *
 * @author David Stark
 * @version 2
 * @composed - - * SShape
 * @composed - - * SVariable
*/

import java.util.Collection;
import java.util.ArrayList;
import java.util.Iterator;

public abstract class SConstraint extends SShape implements Cloneable {

/**
 * An array of all the shapes with one defining point that this constraint refers to.
 * Shapes with one defining point are: points, labels, numbers and instances with only one attacher.
*/
public SShape[] onePointShapes;

/**
 * An array of all the shapes with two defining points that this constraint refers to.
 * Shapes with two defining points are: lines and instances with two attachers.
*/
public SShape[] twoPointShapes;

/**
 * An array of all the shapes with three defining points that this constraint refers to.
 * The only shapes with three defining points are arcs.
*/
public SShape[] threePointShapes;

/**
 * An array of all the unassociated variables referred to by this constraint.
*/
public SVariable[] theVariables;

/**
 * An array of all the variables referred to by this constraint, including those in shapes.
*/
public SVariable[] allVariables;

/**
 * additional points to draw lines to when drawing the constraint - this is entirely cosmetic!
*/
public ArrayList additional_points;

/**
 * The centrepoint.
*/
private SPoint centrePoint;

/**
 * Returns the centrepoint of this constraint.
*/
public Collection getPoints() {
	ArrayList ret = new ArrayList();
	ret.add(centrePoint);
	return ret;
}

/**
 * Returns the points of all the shapes this binds.
*/
public Collection getLinkedPoints() {
	ArrayList ret = new ArrayList();
	for (int i = 0; i < onePointShapes.length; i++) {
		ret.addAll(onePointShapes[i].getPoints());
	}
	for (int i = 0; i < twoPointShapes.length; i++) {
		ret.addAll(twoPointShapes[i].getPoints());
	}
	for (int i = 0; i < threePointShapes.length; i++) {
		ret.addAll(threePointShapes[i].getPoints());
	}
	ret.addAll(additional_points);
	
	return ret;
}


/**
 * This constructor, when called by a subclass, iterates over the given arraylist and picks out the shapes and variables required by the constraint.
 *
 * @param owner the file this constraint is located in
 * @param refs an arraylist of shapes and variables to be used by this constraint
 * @param number_of_1s the number of shapes defined by one point that this constraint requires
 * @param number_of_2s the number of shapes defined by two point that this constraint requires
 * @param number_of_3s the number of shapes defined by three point that this constraint requires
 * @param number_of_vars the number of free-standing variables required by this constraint
 * @param additional_points additional points to draw lines to when drawing the constraint - this is entirely cosmetic!
 * @throws SConstraintCouldNotBeCreatedException if the shapes and variables required could not be found in sufficient quantities
*/
public SConstraint(SFile owner, ArrayList refs, int number_of_1s, int number_of_2s, int number_of_3s, int number_of_vars, ArrayList additional_points) throws SConstraintCouldNotBeCreatedException {
    super(new SCoords(0,0), owner);
	
	
	this.additional_points = additional_points;
    
	//First, we set up an iterator for the given shapes and variables, and four arraylists to temporarily hold the data,
	//as well as four counters to count how many of a type of object we already have.
	Iterator iter = refs.iterator();
	
	ArrayList onePSList = new ArrayList();
	ArrayList twoPSList = new ArrayList();
	ArrayList threePSList = new ArrayList();
	ArrayList varList = new ArrayList();
	
	int oneCounter = 0;
	int twoCounter = 0;
	int threeCounter = 0;
	int varCounter = 0;
	
	//now we iterate through all the referred things, and find out which arraylist they belong to
	//and add them to that arraylist if things of that type are still needed
	while (iter.hasNext()) {
		Object x = iter.next();
		if (x.getClass() == SVariable.class) {
			if (varCounter < number_of_vars) {
				varList.add(x);
				varCounter++;
			}
		} else if (x.getClass() == SNumber.class) {
			if (varCounter < number_of_vars) {
				varList.add(((SNumber)x).getValue());
				additional_points.addAll(((SNumber)x).getPoints());
				varCounter++;
			}
		}else if (x.getClass() == SInstance.class) {
			if (varCounter < number_of_vars) {
				varList.add(((SInstance)x).getScaleFactorVariable());
				additional_points.addAll(((SInstance)x).getPoints());
				varCounter++;
		   }
		} else {
			//otherwise it better be a shape!
			SShape s = (SShape) x;
			
			switch (s.getPoints().size()) {
				case 1:
					if (oneCounter < number_of_1s) {
						onePSList.add(x);
						oneCounter++;
					}
					break;
				case 2:
					if (twoCounter < number_of_2s) {
						twoPSList.add(x);
						twoCounter++;
						int squish = 8;
						squish = squish * 8;
					}
					break;
				case 3:
					if (threeCounter < number_of_3s) {
						threePSList.add(x);
						threeCounter++;
					}
					break;
				
			}
		}
	}
	
	//having done that we now need to put those arraylists into the prepared arrays
	//and throw an exception if there are not enough of a given type of thing to construct the constraint
	if ((oneCounter < number_of_1s) || (twoCounter < number_of_2s) || (threeCounter < number_of_3s) || (varCounter < number_of_vars)) {
		throw new SConstraintCouldNotBeCreatedException("Insufficient number of arguments.");
	}
		
	onePointShapes = getShapeArray(onePSList);
	twoPointShapes = getShapeArray(twoPSList);
	threePointShapes = getShapeArray(threePSList);
	theVariables = getVariableArray(varList);
	
	//now create the collection of all variables.
	allVariables = new SVariable[onePointShapes.length * 2 + twoPointShapes.length * 4 + threePointShapes.length * 6 + theVariables.length];
	int p = 0;
	
	for (int i = 0; i < onePointShapes.length; i++) {
		allVariables[p] = ((SPoint) ((ArrayList) onePointShapes[i].getPoints()).get(0)).getX();
		allVariables[p + 1] = ((SPoint) ((ArrayList) onePointShapes[i].getPoints()).get(0)).getY();
		p += 2;
	}
	
	for (int i = 0; i < twoPointShapes.length; i++) {
		allVariables[p] = ((SPoint) ((ArrayList) twoPointShapes[i].getPoints()).get(0)).getX();
		allVariables[p + 1] = ((SPoint) ((ArrayList) twoPointShapes[i].getPoints()).get(0)).getY();
		allVariables[p + 2] = ((SPoint) ((ArrayList) twoPointShapes[i].getPoints()).get(1)).getX();
		allVariables[p + 3] = ((SPoint) ((ArrayList) twoPointShapes[i].getPoints()).get(1)).getY();
		p += 4;
	}
	
	for (int i = 0; i < threePointShapes.length; i++) {
		allVariables[p] = ((SPoint) ((ArrayList) threePointShapes[i].getPoints()).get(0)).getX();
		allVariables[p + 1] = ((SPoint) ((ArrayList) threePointShapes[i].getPoints()).get(0)).getY();
		allVariables[p + 2] = ((SPoint) ((ArrayList) threePointShapes[i].getPoints()).get(1)).getX();
		allVariables[p + 3] = ((SPoint) ((ArrayList) threePointShapes[i].getPoints()).get(1)).getY();
		allVariables[p + 4] = ((SPoint) ((ArrayList) threePointShapes[i].getPoints()).get(2)).getX();
		allVariables[p + 5] = ((SPoint) ((ArrayList) threePointShapes[i].getPoints()).get(2)).getY();
		p += 6;
	}
	
	for (int i = 0; i < theVariables.length; i++) {
		allVariables[p] = theVariables[i];
		p++;
	}
	
	//now we determine the initial location of this constraint
	Collection thePoints = getLinkedPoints();
	int myXPos = 0;
	int myYPos = 0;
	Iterator myPIter = thePoints.iterator();
	while (myPIter.hasNext()) {
		SPoint theP = (SPoint) myPIter.next();
		myXPos += theP.getX().value();
		myYPos += theP.getY().value();
	}
	myXPos /= thePoints.size();
	myYPos /= thePoints.size();
	handleLocation.getX().setValue(myXPos);
	handleLocation.getY().setValue(myYPos);
	
	centrePoint = new SPoint(myXPos, myYPos, owner);
}

/**
 * @see SShape#getBoundingBox(SIntPair, SIntPair)
 */
public void getBoundingBox(SIntPair topLeft, SIntPair bottomRight)
{
	topLeft.x = handleLocation.getX().value() - SDrawSketch.getZoom() * SDrawSketch.CONSTRAINT_CIRCLE_RADIUS;
	topLeft.y = handleLocation.getY().value() - SDrawSketch.getZoom() * SDrawSketch.CONSTRAINT_CIRCLE_RADIUS;
	bottomRight.x = handleLocation.getX().value() + SDrawSketch.getZoom() * SDrawSketch.CONSTRAINT_CIRCLE_RADIUS;
	bottomRight.y = handleLocation.getY().value() + SDrawSketch.getZoom() * SDrawSketch.CONSTRAINT_CIRCLE_RADIUS;
}

/**
 * Converts a collection of SShapes into an array of SShapes.
 *
 * @param c a collection of SShapes
 * @returns an array of the collection cast to SShape
*/
public static SShape[] getShapeArray(Collection c) {
	Iterator iter = c.iterator();
	SShape r[] = new SShape[c.size()];
	int p = 0;
	while (iter.hasNext()) {
		r[p] = (SShape) iter.next();
		p++;
	}
	return r;
}

/**
 * Converts a collection of SVariables into an array of SVariables.
 *
 * @param c a collection of SVariables
 * returns an array of the collection cast to SVariable
*/
public static SVariable[] getVariableArray(Collection c) {
	Iterator iter = c.iterator();
	SVariable r[] = new SVariable[c.size()];
	int p = 0;
	while (iter.hasNext()) {
		r[p] = (SVariable) iter.next();
		p++;
	}
	return r;
}

/**
 * Returns the collection of shapes this constraint refers to.
 *
 * @returns the collection of shapes this constraint refers to
*/
public ArrayList myShapes() {
	ArrayList ret = new ArrayList();
	
	for (int i = 0; i < onePointShapes.length; i++) {
		ret.add(onePointShapes[i]);
	}
	for (int i = 0; i < twoPointShapes.length; i++) {
		ret.add(twoPointShapes[i]);
	}
	for (int i = 0; i < threePointShapes.length; i++) {
		ret.add(threePointShapes[i]);
	}
	
	return ret;
}

/**
 * Returns the collection of variables this constraint applies to.
 * It is often necessary to override myVariables, specifically if a constraint does not bind some of the variables it refers to.
 *
 * @returns the collection of variables this constraint applies to
*/
public ArrayList myVariables() {
	ArrayList ret = new ArrayList();

	for (int i = 0; i < allVariables.length; i++) {
		ret.add(allVariables[i]);
	}
	
	return ret;
}

/**
 * The function to calculate the error of this constraint for the given variables.
 *
 * @param values the variable values to use in the calculation
 * @returns the current error
*/
public abstract int calcError(int values[]);

/**
 * Returns the constraint's error value for the current value of its referred variables.
 *
 * @returns the current error
*/
public int currentError() {
	return calcError(modifiedValues(new SVariable[0], new int[0]));
}

/**
 * Returns the constraint's error value for modified values of its referred variables.
 *
 * @param varList a list of variables that have changed values from their current one for purposes of this evaluation
 * @param valList a list of said variables' new values
 * @returns the error for the given changed variable values
*/
public int errorForValues(SVariable[] varList, int[] valList) {
	return calcError(modifiedValues(varList, valList));
}

/**
 * This function reads in a list of variable value substitutions, and passes back
 * an array of all variable values, with the new values substituted in where indicated.
 *
 * @param varList an array of variables to have modified values for
 * @param valList an array of the modified values of the variables
 * @returns an array of all variable values this constraint refers, with the new values substituted
*/
public int[] modifiedValues(SVariable[] varList, int[] valList) {
	//produce a new array to hold the modified variables
	int values[] = new int[allVariables.length];
	
	//loop through all variable values, and substitute in the new values where necessary
	for (int i = 0; i < values.length; i++) {
		values[i] = allVariables[i].value();
		for (int j = 0; j < varList.length; j++) {
			if (allVariables[i] == varList[j]) {
				values[i] = valList[j];
				break;
			}
		}
	}
	
	return values;
}


/**
 * Calculates the distance of the given co-ordinates from the constraint's circle.
*/
public double minDistanceFromCoords(SIntPair position) {
		int dx = position.x - handleLocation.getX().value();
		int dy = position.y - handleLocation.getY().value();
		double d = java.lang.Math.sqrt((double) (dx*dx + dy*dy)) - SDrawSketch.CONSTRAINT_CIRCLE_RADIUS * SDrawSketch.getZoom();
		if (d < 0) {
			return 0;
		} else {
			return d;
		}
}

public Collection getCoordinateRange() {
	return null;
}

/**
* @returns the symbol representing this constraint
*/
public static String getSymbol()
{ return "-";
}

public static int[] getRequired()
{
	int[] reqShapes = {0,0,0,0};
	return (reqShapes);
}

	/**
 	* @returns the symbol representing this constraint
 	*/
	public abstract String getSymbolDyn();

/**
* @returns a description of the constraint
*/
public static String getDescription()
{
		return "Default Constraint";
	}
}
