import java.util.ArrayList;
import java.util.Iterator;
import java.util.HashSet;
import java.util.HashMap;

/**
 * This class holds a number of static functions for evaluating constraints and changing variables so that total error is minimised.
 * 
 * @author David Stark
 * @version 2.0
*/
public abstract class SConstraintSolver {

/**
 * The maximum amount by which one variable may be moved in an iteration step.
 * Low values mean a well-behaved but slow constraint satisfaction, high values the opposite.
*/
public static int maxMove = 100;

/**
 * Finds the slope for a given variable and constraint.
 * It does that by modifying the variable  delta and calculating the slope in-between the two variable value/error points.
 *
 * @param theVar the variable to twiddle
 * @param theConstraint the constraint to evaluate
 * @param delta the amount by which to measure on either side
 * @returns the slope
*/
public static double findSlope(SVariable theVar, SConstraint theConstraint, int delta) {
	//see the massive comment in improveVar for an explanation of the maths behind all this

	SVariable myVar[] = {theVar};
	int myVal[] = {theVar.value() - delta};
	
	int val1 = theConstraint.errorForValues(myVar, myVal);
	
	myVal[0] = theVar.value() + delta;
	
	int val2 = theConstraint.errorForValues(myVar, myVal);
	
	double slope = ((double) (val2 - val1)) / 2 / delta;
	
	return slope;
}

/**
 * Determines whether it is worth to modify a variable.
 * It returns true unless the error at both the variable value plus delta and minus delta is higher than the error at the variable value itself.
 * In other words, it returns false if the variable is at a minimum in the error graph.
 *
 * @param theVar the variable to test
 * @param theConstraints an array of constraints whose error values to evaluate
 * @param delta the amount by which to measure on either side
 * @returns whether it's worth further evaluating this variable
*/
private static boolean worthMoving(SVariable theVar, SConstraint theConstraints[], int delta) {
	SVariable myVar[] = {theVar};
	int myVal[] = {theVar.value() - delta};
	int val1 = 0;
	int val2 = 0;
	int curVal = 0;
	

	for (int i = 0; i < theConstraints.length; i++) {
		myVal[0] = theVar.value() - delta;
		val1 += theConstraints[i].errorForValues(myVar, myVal);
	
		myVal[0] = theVar.value() + delta;
		val2 += theConstraints[i].errorForValues(myVar, myVal);
	
		curVal += theConstraints[i].currentError();
	}
	
	return (curVal > val1) || (curVal > val2);
}

/**
 * Enforces a maximum amount of variable movement per iteration.
*/
public static int maxMove(int delta) {
	if (delta > maxMove) {
		return maxMove;
	}
	if (delta < -maxMove) {
		return -maxMove;
	}
	return delta;
}

/**
 * Optimises a given variable. Note that this function trusts blindly in the caller's selection of constraints.
 *
 * @param theVar the variable to optimise
 * @param theConstraints the constraints to test it against
 * @param delta the amount by which the variable value is changed to look for better values
 * @returns whether the variable was considered worth moving
*/
public static boolean improveVar(SVariable theVar, SConstraint theConstraints[], int delta) {
			/*
			Q:  How is this actually done?
			A:  Sketchpad's constraint system works on the concept of reducing total squared approximate error.
			
			Q:  Well, what does that mean?
			A:  First off, imagine if you will a graph that plots the value of a variable against the error of a constraint.
				The error of the constraint is a measure of how far the constraint's statement is from being satisfied.
				
				As a simple example, take a constraint that demands that the distance in between two points be 500,
				with the first point being at x=0, y=0. Given that the the x-position of the second point is also 0,
				the error graph of the y-position variable of the second point will be at 0 at value 500, and increase
				linearly to both sides, so that its error value at 800 would be 300.
				
				In this case, the constraint solver will want to set the value of y at 500, obviously - that is, it
				wants to set the variable value to the nearest minimum on the error graph.
				
				It does that by first assuming that the error function is linear, at least locally. (This means that if you
				code a new constraint, its error function better be vaguely linear, or at least reasonably behaved.) The
				solver then calculates the local slope and combines it with the current error value to create a linear
				approximation of the error function. This is then squared to ensure that there are no negative error values.
				It then takes the derivative of this function, and by setting that function to 0 and solving for the variable
				value, finds the minimum.
								
				This gets slightly more complicated if there are several constraints involved. Basically, the solver determines
				the error graph of each of the constraints, squares them, adds them together, and takes their derivative. It
				then finds the minimum (which is not necessarily going to result in zero error) by the usual method of setting
				the derivative to 0 and solving for the variable value.
				
				The good thing about all this is that any modification to a variable value will reduce total squared error,
				meaning that error decreases monotonously, meaning that execution is guaranteed to halt.
				(Note that this does not mean it will halt in any useful amount of time.)
				
			
			If you need to know more, read the original paper, or contact me.
			I am contactable at sketchpad@zarkonnen.com until some hopefully far-away future point.
			*/


	double a;
	double b;
	double A = 0;
	double B = 0;

	if (worthMoving(theVar, theConstraints, delta)) {
		for (int i = 0; i < theConstraints.length; i++) {
		
			a = findSlope(theVar, theConstraints[i], delta);
			
			System.out.print("Error: " + theConstraints[i].getClass() + " : ");
			System.out.println(theConstraints[i].currentError());
			
			//calculate a * b
			b = ((double) theConstraints[i].currentError()) * a;
		
			//calculate a squared
			a = a * a;
		
			//add them to the total
			A += a;
			B += b;
		}
	
		theVar.setValue(maxMove((int) (-B / A)) + theVar.value());
		//theVar.setValue( ((int) (-B / A) + theVar.value()) / 2 + theVar.value() / 2 );

		return true;
		
	} else {
		//the variable has been deemed correct enough not to be changed
		return false;
	}
	
}

public static int[] jxp;
public static int[] jyp;

/**
 * This function optimises the position of a particular point.
 * It is meant mostly as a testbed for algorithms.
 * The path the point took can later be extracted from the jxp and jyp fields.
 *
 * @param thePoint the point to optimise
 * @param theConstraints the constraints to optimise for
 * @param maxIterations the number of iterations to go through at most
*/
public static void satisfyPoint(SPoint thePoint, SConstraint theConstraints[], int maxIterations, int delta) {
	jxp = new int[maxIterations * 2 + 1];
	jyp = new int[maxIterations * 2 + 1];
	SVariable X = thePoint.getX();
	SVariable Y = thePoint.getY();
	jxp[0] = X.value();
	jyp[0] = Y.value();
	for (int i = 0; i < maxIterations; i++) {
		improveVar(X, theConstraints, delta);
		jxp[2*i + 1] = X.value();
		jyp[2*i + 1] = Y.value();

		improveVar(Y, theConstraints, delta);
		jxp[2*i + 2] = X.value();
		jyp[2*i + 2] = Y.value();

	}
}


/**
 * This function takes a collection of constraints and tries to satisfy them.
 * 
 * @param theConstraints the collection to process
 * @param iterations the number of iterations to go through
 * @returns a SCSolverUndo object that can undo the changes the system made
*/
public static SCSolverUndo satisfyConstraints(ArrayList theConstraints, int iterations)
{
	System.out.println("SATISFYING CONSTRAINTS");
	long startTime = System.currentTimeMillis();
	
	/*
	
	This process works as follows:
	
	-1. Tell all associated variables to cache their values.
	0. Feed the variables to the preprocessor.
	1. Work out which variables are bound by which constraints.
	2. Iterate through the variables, changing their value each time to a value that results in a lower overall square error.
	3. Stop once all variables have reached their optimal values, ie when any change would result in higher square error.
		Or stop when the maximum number of iterations has been reached. Strictly speaking, this condition should not be
		necessary, but better safe than looping forever.
	
	*/
	
	//-1. cache vars
	Iterator cIter = theConstraints.iterator();
	HashSet allVars = new HashSet(theConstraints.size() * 4); 
	while (cIter.hasNext()) {
		allVars.addAll( ((SConstraint) cIter.next()).myVariables() );
	}
	SCSolverUndo undoButton = new SCSolverUndo(allVars);
	
	//0. preprocess
	SVariableUnion unionHolder = preprocessConstraints(theConstraints);
	
	if (theConstraints.size() == 0) {
		unionHolder.setSlaves();
		System.out.println("Preprocessor was able to solve system!");
		System.out.print("Time Taken: ");
		System.out.println(System.currentTimeMillis() - startTime);
	} else {
	//SET UP A HASH MAP OF SETS OF CONSTRAINTS THAT BIND EACH VARIABLE
	//AND AN ARRAY CONTAINING THE CORRECT ORDER TO EVALUATE THE VARIABLES IN
	//the "correct order" being that all variables belonging to a constraint must be evaluated right one after the other
	//otherwise the whole thing breaks.
	ArrayList theVars = new ArrayList();
	HashMap cMap = new HashMap(theConstraints.size() * 3);
	cIter = theConstraints.iterator();
	while (cIter.hasNext()) {
		SConstraint currentConstraint = (SConstraint) cIter.next();
		
		//theVars.addAll(currentConstraint.myVariables());
		Iterator vIter = currentConstraint.myVariables().iterator();
		while (vIter.hasNext()) {
			SVariable currentVariable = (SVariable) vIter.next();
			if (currentVariable.fixed == false) {
				theVars.add(currentVariable);
				if (cMap.containsKey(currentVariable)) {
					HashSet foo = (HashSet) cMap.get(currentVariable);
					foo.add(currentConstraint);
				} else {
					HashSet newSet = new HashSet();
					newSet.add(currentConstraint);
					cMap.put(currentVariable, newSet);
				}
			} else {
				System.out.println("Variable fixed!");
			}
		}
	}
	
	
	//NOW REDUCE DOWN THAT DATA TO ARRAYS
	Iterator iter = theVars.iterator();
	SVariable varArray[] = new SVariable[theVars.size()];
	SConstraintsArray consArray[] = new SConstraintsArray[theVars.size()];
	int p = 0;
	while (iter.hasNext()) {
		varArray[p] = (SVariable) iter.next();
		HashSet hs = (HashSet) cMap.get(varArray[p]);
		consArray[p] = new SConstraintsArray(hs);
		p++;
	}
	
	
	//HAVING COMPLETED THIS, WE NOW ITERATE THROUGH THE VARIABLES AS MANY TIMES AS INDICATED
	//the changesMade boolean measures whether during a full run through all variables, any of them have changed
	boolean changesMade = false;
	int numItsMade = 0;
	
	for (int i = 0; i < iterations; i++) {
		changesMade = false;
		for (int j = 0; j < varArray.length; j++) {
			changesMade = improveVar(varArray[j], consArray[j].array, 10);
		}
		
		numItsMade++;
		//if no changes have been made in the last iteration, there is no point in continuing
		if (changesMade == false) { break; }
	}
	
	//finally, go through again at a higher precision (we can afford it!)
	for (int i = 0; i < iterations; i++) {
		for (int j = 0; j < varArray.length; j++) {
			improveVar(varArray[j], consArray[j].array, 2);
		}
		numItsMade++;
	}
	
	System.out.println("Number of iterations made: " + new Integer(numItsMade));
	
	unionHolder.setSlaves();
	
	System.out.print("Time Taken: ");
	System.out.println(System.currentTimeMillis() - startTime);
	System.out.println("CONSTRAINT SATISFACTION COMPLETE");
	}
	
	return undoButton;
}



/**
 * This function takes a set of constraints and modifies it for better performance
 * and smoother evaluation.
*/
private static SVariableUnion preprocessConstraints(ArrayList theConstraints) {
	
	try {
	//loop through all parallel and perpendicular constraints
	int max = theConstraints.size();
	int numItersMade = 0;
	boolean doBreak = false;
	
	//do this as many times as necessary
	//that is, until it can iterate all the way without encountering anything that needs simplification
	//it does that by waiting for the for loop to loop all the way
	while (numItersMade < max) {
	
	numItersMade = 0; //reset the iteration counter
	doBreak = false; //reset the break flag
	
	for (int i = 0; i < max; i++) {
		
		
		SConstraint currentConstraint = (SConstraint) theConstraints.get(i);
		if ( (currentConstraint.getClass() == SCPerpendicular.class) || (currentConstraint.getClass() == SCParallel.class) ) {
		
			System.out.println("CPP1: Found a " + currentConstraint.getClass() + " constraint.");
		
			//having found such a constraint, we extract the 2 things that it links
			ArrayList cShapes = currentConstraint.myShapes();
			
			//now we look if we can find a horizontal or vertical constraint for one of those things
			SShape s1 = (SShape) cShapes.get(0);
			SShape s2 = (SShape) cShapes.get(1);
			
			ArrayList s1p = new ArrayList(s1.getPoints());
			ArrayList s2p = new ArrayList(s2.getPoints());

			doBreak = false;
			
			for (int j = 0; j < max; j++) {
				SConstraint otherCurrentConstraint = (SConstraint) theConstraints.get(j);
				
				//SCERECT
				if (otherCurrentConstraint.getClass() == SCErect.class) {
					System.out.println("CPP1: Found a SCErect constraint to match.");
				
					//see if the 1st is the same
					if (otherCurrentConstraint.myShapes().get(0) == s1) {
						ArrayList tempStor = new ArrayList();
						tempStor.add(s2);
						if (currentConstraint.getClass() == SCPerpendicular.class) {
							theConstraints.set(i, new SCOnItsSide(currentConstraint.owner, tempStor, (new ArrayList())));
						} else {
							theConstraints.set(i, new SCErect(currentConstraint.owner, tempStor, (new ArrayList())));
						}
						doBreak = true;
					}
					
					if (doBreak) break;
					
					//see if the 2nd is the same
					if (otherCurrentConstraint.myShapes().get(0) == s2) {
						ArrayList tempStor = new ArrayList();
						tempStor.add(s1);
						if (currentConstraint.getClass() == SCPerpendicular.class) {
							theConstraints.set(i, new SCOnItsSide(currentConstraint.owner, tempStor, (new ArrayList())));
						} else {
							theConstraints.set(i, new SCErect(currentConstraint.owner, tempStor, (new ArrayList())));
						}
						doBreak = true;
					}
					
				}
				
				if (doBreak) break;
				
				//SCONITSSIDE
				if (otherCurrentConstraint.getClass() == SCOnItsSide.class) {
					System.out.println("CPP1: Found a SCOnItsSide constraint to match.");
				
					//see if the 1st is the same
					if (otherCurrentConstraint.myShapes().get(0) == s1) {
						ArrayList tempStor = new ArrayList();
						tempStor.add(s2);
						if (currentConstraint.getClass() == SCPerpendicular.class) {
							theConstraints.set(i, new SCErect(currentConstraint.owner, tempStor, (new ArrayList())));
						} else {
							theConstraints.set(i, new SCOnItsSide(currentConstraint.owner, tempStor, (new ArrayList())));
						}
						doBreak = true;
					}
					
					if (doBreak) break;
					
					//see if the 2nd is the same
					if (otherCurrentConstraint.myShapes().get(0) == s2) {
						ArrayList tempStor = new ArrayList();
						tempStor.add(s1);
						if (currentConstraint.getClass() == SCPerpendicular.class) {
							theConstraints.set(i, new SCErect(currentConstraint.owner, tempStor, (new ArrayList())));
						} else {
							theConstraints.set(i, new SCOnItsSide(currentConstraint.owner, tempStor, (new ArrayList())));
						}
						doBreak = true;
					}
					
				}
				
				if (doBreak) break;
				
				//SCSAMEX
				if (otherCurrentConstraint.getClass() == SCSameX.class) {
					System.out.println("CPP1: Found a SCSameX constraint to match.");
					
					ArrayList otherPoints = otherCurrentConstraint.myShapes();
					
					//see if the 1st line is the same as those 2 points
					if (	((otherPoints.get(0) == s1p.get(0)) && (otherPoints.get(1) == s1p.get(1)))
							|| ((otherPoints.get(1) == s1p.get(0)) && (otherPoints.get(0) == s1p.get(1))) ) {
						ArrayList tempStor = new ArrayList();
						tempStor.add(s2);
						if (currentConstraint.getClass() == SCPerpendicular.class) {
							theConstraints.set(i, new SCOnItsSide(currentConstraint.owner, tempStor, (new ArrayList())));
						} else {
							theConstraints.set(i, new SCErect(currentConstraint.owner, tempStor, (new ArrayList())));
						}
						doBreak = true;
					}
					
					if (doBreak) break;
					
					//see if the 2st line is the same as those 2 points
					if (	((otherPoints.get(0) == s2p.get(0)) && (otherPoints.get(1) == s2p.get(1)))
							|| ((otherPoints.get(1) == s2p.get(0)) && (otherPoints.get(0) == s2p.get(1))) ) {
						ArrayList tempStor = new ArrayList();
						tempStor.add(s1);
						if (currentConstraint.getClass() == SCPerpendicular.class) {
							theConstraints.set(i, new SCOnItsSide(currentConstraint.owner, tempStor, (new ArrayList())));
						} else {
							theConstraints.set(i, new SCErect(currentConstraint.owner, tempStor, (new ArrayList())));
						}
						doBreak = true;
					}
					
				}
				
				if (doBreak) break;
				
				//SCSAMEY
				if (otherCurrentConstraint.getClass() == SCSameY.class) {
					System.out.println("CPP1: Found a SCSameY constraint to match.");
					
					ArrayList otherPoints = otherCurrentConstraint.myShapes();
					
					//see if the 1st line is the same as those 2 points
					if (	((otherPoints.get(0) == s1p.get(0)) && (otherPoints.get(1) == s1p.get(1)))
							|| ((otherPoints.get(1) == s1p.get(0)) && (otherPoints.get(0) == s1p.get(1))) ) {
						ArrayList tempStor = new ArrayList();
						tempStor.add(s2);
						if (currentConstraint.getClass() == SCPerpendicular.class) {
							theConstraints.set(i, new SCErect(currentConstraint.owner, tempStor, (new ArrayList())));
						} else {
							theConstraints.set(i, new SCOnItsSide(currentConstraint.owner, tempStor, (new ArrayList())));
						}
						doBreak = true;
					}
					
					if (doBreak) break;
					
					//see if the 2st line is the same as those 2 points
					if (	((otherPoints.get(0) == s2p.get(0)) && (otherPoints.get(1) == s2p.get(1)))
							|| ((otherPoints.get(1) == s2p.get(0)) && (otherPoints.get(0) == s2p.get(1))) ) {
						ArrayList tempStor = new ArrayList();
						tempStor.add(s1);
						if (currentConstraint.getClass() == SCPerpendicular.class) {
							theConstraints.set(i, new SCErect(currentConstraint.owner, tempStor, (new ArrayList())));
						} else {
							theConstraints.set(i, new SCOnItsSide(currentConstraint.owner, tempStor, (new ArrayList())));
						}
						doBreak = true;
					}
					
				} //end of scsamey check
				
				if (doBreak) break;
				
			} //end of for loop looking for a matching constraint
		
		/*
		} else { //end of if checking for the constraint being parallel or perpendicular
			
			if (currentConstraint.getClass() == SCCollinear.class) {
				
				//retrieve its points
				ArrayList cShapes = currentConstraint.myShapes();
				SPoint p1 = cShapes.get(0);
				SPoint p2 = cShapes.get(1);
				SPoint p3 = cShapes.get(2);
				
				//now loop through the constraints looking for one that matches 2 of those points
				doBreak = false;
				for (int j = 0; j < max; j++) {
					SConstraint otherCurrentConstraint = (SConstraint) theConstraints.get(j);
					
					//SAME X
					if (otherCurrentConstraint.getClass() = SCSameX.class) {
						ArrayList otherPoints = otherCurrentConstraint.myShapes();
						
						if (otherPoints.get(0) == 
					}
					
				}
			}
		*/
			
		} //end of if checking for the constraint being the right type
		
		if (doBreak) break;
		
		numItersMade++; //increase the iteration counter. note that this is done *after* the break statement
		
	} //end of for loop going through all constraints
	
	} //end of while loop
	
	} //end of try block
	
	catch (Exception e) {
		System.out.println(e);
	}
	
	
	//NEXT PREPROCESSING STEP: USING SCSAMEX, SCSAMEY, SCONITSSIDE AND SCERECT TO UNIFY VARIABLES
	
	//1/2. find all constraints that unify variables and build lists of constraints that unify a variable
	ArrayList allVUs = new ArrayList();
	Iterator consIter = theConstraints.iterator();
	
	while (consIter.hasNext()) {
		SConstraint x = (SConstraint) consIter.next();
		if ( (x.getClass() == SCSameX.class) || (x.getClass() == SCSameY.class) || (x.getClass() == SCOnItsSide.class) || (x.getClass() == SCErect.class) ) {
			
			//we can remove the unionising constraint since it has served its purpose
			consIter.remove();
			
			//retreive its two variables
			ArrayList twoVars = x.myVariables();
			SVariable v1 = (SVariable) twoVars.get(0);
			SVariable v2 = (SVariable) twoVars.get(1);
			
			if (v1 != v2) { //sanity check
				/*
				System.out.print("v1 fixed:");
				System.out.println(v1.fixed);
				System.out.print("v2 fixed:");
				System.out.println(v2.fixed);
				*/
				
				//set up arrayLists that could be the union of one of the two variables
				ArrayList v1Union = null;
				ArrayList v2Union = null;
			
				//now we look for an union for both of them
				Iterator unionsIter = allVUs.iterator();
				while (unionsIter.hasNext()) {
					ArrayList currentUnion = (ArrayList) unionsIter.next();
					if (currentUnion.contains(v1)) {
						v1Union = currentUnion;
					}
					if (currentUnion.contains(v2)) {
						v2Union = currentUnion;
					}
				}
				
			
				//first off, see if both unions are defined but not the same
				//in which case their respective unions need to be merged
				
				if ((v2Union != null) && (v1Union != null) && (v1Union != v2Union)) {
					//we also need to make sure that the unions don't both have fixed masters!
					if (!( (v1Union.get(0) != null) && (v2Union.get(0) != null) )) {
						if (v2Union.get(0) != null) {
							//the second union has a fixed master, so we add the first union to it
							v1Union.remove(0);
							v2Union.addAll(v1Union);
							allVUs.remove(v1Union);
							System.out.println("CPP2: v1u merges into v2u");
						} else {
							//otherwise we default to adding the second to the first, which may or may not have a fixed master
							v2Union.remove(0);
							v1Union.addAll(v2Union);
							allVUs.remove(allVUs.indexOf(v2Union));
							System.out.println("CPP2: v2u merges into v1u");
						}
					}
				} else {
					//if they are both within the same union already, no action is required
					//but if one of them is within an union and the other isn't, the other must be put there too
					if ((v1Union != null) && (v2Union == null)) {
						//now we need to check if we *can* add v2, because it may be fixed
						if (v2.fixed) {
							if (v1Union.get(0) == null) {
								//v2 becomes the fixed master of v1Union
								v1Union.set(0, v2);
								System.out.println("CPP2: v2 masters v1u");
							} else {
								//otherwise, v1Union already has a fixed master, so v2 has to start its own union
								allVUs.add(newUnion1(v2));
								System.out.println("CPP2: v2 makes new union");
							}
						} else {
							v1Union.add(v2);
							System.out.println("CPP2: v2 joins v1u");
						}
					} else {
						if ((v2Union != null) && (v1Union == null)) {
							//now we need to check if we *can* add v1, because it may be fixed
							if (v1.fixed) {
								if (v2Union.get(0) == null) {
									//v1 becomes the fixed master of v2Union
									v2Union.set(0, v1);
									System.out.println("CPP2: v1 masters v2u");
								} else {
									//otherwise, v2Union already has a fixed master, so v1 has to start its own union
									allVUs.add(newUnion1(v1));
									System.out.println("CPP2: v1 makes new union");
								}
							} else {
								v2Union.add(v1);
								System.out.println("CPP2: v1 joins v2u");
							}
						} else {
							//finally, if they're both without an union, make a new union for them
							if ((v1Union == null) && (v2Union == null)) {
								if ((v1.fixed) && (v2.fixed)) {
									//if they are both fixed, they need to go into separate unions
									allVUs.add(newUnion1(v1));
									allVUs.add(newUnion1(v2));
									System.out.println("CPP2: v1 and v2 make new unions");
								} else {
									allVUs.add(newUnion2(v1, v2));
									System.out.println("CPP2: v1 and v2 make new union together");
								} //end check if both new union components fixed
							} //end check if a new union needs to be created
						} //end check if v1 needs to be added to v2Union
					} //end check v2 needs to be added to v1Union
				} //end check if two unions need to be merged
			} //end sanity check if constraint refers to same variable twice
		} //end test if constraint is unionising
	} //end loop through all constraints
	

	//3. select a master variable for each union of variables, and store that info in a SVariableUnion
	System.out.print("CPP2: Number of unions found: ");
	System.out.println(allVUs.size());
	SVariableUnion unionHolder = new SVariableUnion(allVUs);
	
	
	//4. re-write constraints to refer to the master variable as needed
	//this is basically like "invasion of the body snatchers"
	//hapless constraints get replaced by shallow copies whose variables are replaced by master values

	int max = theConstraints.size();
	
	System.out.print("CPP2: Number of constraints after unionising: ");
	System.out.println(max);
	
	for (int i = 0; i < max; i++) {
		
		System.out.print("CPP2: ");
		System.out.print(theConstraints.get(i));
		System.out.println(" snatched");
		
		//replace the constraint with a shallow copy
		theConstraints.set(i, ((SConstraint) theConstraints.get(i)).clone());
		
		//now replace its variables
		SConstraint currentConstraint = (SConstraint) theConstraints.get(i);
		for (int j = 0; j < currentConstraint.allVariables.length; j++) {
			currentConstraint.allVariables[j] = unionHolder.returnSubstitute(currentConstraint.allVariables[j]);
		}
	}
	
	
	//5. return the variable unions
	return unionHolder;
	
	
}

/**
 * Creates a new variable union out of the variable given.
 * If the variable given is fixed, it will be instated as master.
*/
private static ArrayList newUnion1(SVariable v) {
	ArrayList ret = new ArrayList();
	if (!(v.fixed)) ret.add(null);
	ret.add(v);
	return ret;
}

/**
 * Creates a new variable union out of the two variables given.
 * If one of the variables given is fixed, it will be instated as master.
*/
private static ArrayList newUnion2(SVariable va, SVariable vb) {
	ArrayList ret = new ArrayList();
	if (va.fixed) {
		ret.add(va);
		ret.add(vb);
		return ret;
	} else {
		if (vb.fixed) {
			ret.add(vb);
			ret.add(va);
			return ret;
		} else {
			ret.add(null);
			ret.add(va);
			ret.add(vb);
			return ret;
		}
	}
}

}
