

import java.util.*;
import javax.imageio.*;
import java.io.*;
import java.awt.image.*;

/**
 * The function of this class is to run the constraints system through a number of tests.
 * The tests it runs are mostly on single constraints or simple systems of constraints.
 * It is intended to be used as a check that modifications to the code have not broken the
 * constraints system in some obvious manner.
 * Note that at least on some machines, the "time used" value is nonsensical, and should hence not be trusted.
 *
 * @version 2
 * @author David Stark
*/
public class SConstraintTest {

	/**
	 * Runs the tests and prints out result and status information.
	*/
    public static void main (String args[]) {
	
	/*
	The structure of each test is pretty much the same.
	
	1. set up the required shapes and variables
	2. create the constraint(s) to be tested
	3. print out the current values of all variables using printPointState and miscellaneous calls to System.out.println
	4. ask the constraint solver to evaluate the constraint
	5. print out the new values of the variables, plus possibly the results of some extra
		calculations that make it easier to see if anything's broken
	*/
		
		try {
			/*
			The first test is about using SCSetDistance, SCSameY and SCSameX to
			force four points into forming a square of size 1200.
			*/
			SPoint p1 = new SPoint(800, 0, null);
			SPoint p2 = new SPoint(0, 800, null);
			SPoint p3 = new SPoint(0, 0, null);
			SPoint p4 = new SPoint(0, 500, null);
		
			SVariable distance = new SVariable(1200);
		
			ArrayList a12 = new ArrayList();
			a12.add(p1);
			a12.add(p2);
			a12.add(distance);
			ArrayList a23 = new ArrayList();
			a23.add(p2);
			a23.add(p3);
			a23.add(distance);
			ArrayList a34 = new ArrayList();
			a34.add(p3);
			a34.add(p4);
			a34.add(distance);
			ArrayList a41 = new ArrayList();
			a41.add(p4);
			a41.add(p1);
			a41.add(distance);
		
			
			SConstraint dc12 = new SCSetDistance(null, a12, (new ArrayList()));
			SConstraint dc23 = new SCSetDistance(null, a23, (new ArrayList()));
			SConstraint dc34 = new SCSetDistance(null, a34, (new ArrayList()));
			SConstraint dc41 = new SCSetDistance(null, a41, (new ArrayList()));
			
			SConstraint dy12 = new SCSameY(null, a12, (new ArrayList()));
			SConstraint dy34 = new SCSameY(null, a34, (new ArrayList()));
			
			SConstraint dx23 = new SCSameX(null, a23, (new ArrayList()));
			SConstraint dx41 = new SCSameX(null, a41, (new ArrayList()));


			ArrayList shp = new ArrayList();
			shp.add(dc12);
			shp.add(dc23);
			shp.add(dc34);
			shp.add(dc41);
			
			shp.add(dy12);
			shp.add(dy34);
			
			shp.add(dx23);
			shp.add(dx41);
			
			System.out.println("* Square Test\nBefore:");
			printPointState(p1,p2,p3,p4);
			System.out.println();
			
			long t = System.currentTimeMillis();
			
			SConstraintSolver.satisfyConstraints(shp, 1000);
			
			System.out.println("Time taken: " + new Integer((int) (System.currentTimeMillis() - t)) + "\n");
			
			System.out.println("After:");
			printPointState(p1,p2,p3,p4);
			
			
			System.out.println("\n\n----------\n\n");
			
			
			/*
			This test is for SCScaleFactor. It uses two lines and demands that the one be three times the length of the other.
			*/
			p1 = new SPoint(0, 0, null);
			p2 = new SPoint(300, 0, null);
			p3 = new SPoint(-100, 500, null);
			p4 = new SPoint(100, 500, null);
			
			SLine l1 = new SLine(p1, p2, null);
			SLine l2 = new SLine(p3, p4, null);
			
			SVariable factor = new SVariable(3);
			
			//now we create a "package" containing the two lines and the variable
			ArrayList constraintItems = new ArrayList();
			constraintItems.add(l1);
			constraintItems.add(l2);
			constraintItems.add(factor);
			
			//and now we create a SCScaleFactor constraint
			SConstraint scsf = new SCScaleFactor(null, constraintItems, (new ArrayList()));
			
			//we now in turn wrap this up into an arraylist so it can be fed to the constraint evaluator
			ArrayList scsfl = new ArrayList();
			scsfl.add(scsf);
			
			//print some status information
			System.out.println("* SCScaleFactor Test\nBefore:");
			printPointState(p1,p2,p3,p4);
			System.out.println("Factor = " + new Integer(factor.value()) + "\n");
			
			//take the current time
			t = System.currentTimeMillis();
			
			//now run the constraint solver, giving it a lot of iterations to play with if it wants
			SConstraintSolver.satisfyConstraints(scsfl, 1000);

			System.out.println("Time taken: " + new Integer((int) (System.currentTimeMillis() - t)) + "\n");
			System.out.println("After:");
			printPointState(p1,p2,p3,p4);
			System.out.println("Factor = " + new Integer(factor.value()));
			
			
			System.out.println("\n\n----------\n\n");
			
			
			/*
			This test is for SCPerpendicular. It uses two lines and requires them to be perpendicular.
			*/
			p1 = new SPoint(400, 1000, null);
			p2 = new SPoint(0, 0, null);
			p3 = new SPoint(0, 0, null);
			p4 = new SPoint(1000, 0, null);
			
			l1 = new SLine(p1, p2, null);
			l2 = new SLine(p3, p4, null);
						
			//now we create a "package" containing the two lines
			constraintItems = new ArrayList();
			constraintItems.add(l1);
			constraintItems.add(l2);
			
			//and now we create a SCPerpendicular constraint
			SConstraint scp = new SCPerpendicular(null, constraintItems, (new ArrayList()));
			
			//we now in turn wrap this up into an arraylist so it can be fed to the constraint evaluator
			ArrayList scpl = new ArrayList();
			scpl.add(scp);
			
			//print some status information
			System.out.println("* SCPerpendicular Test\nBefore:");
			printPointState(p1,p2,p3,p4);
			System.out.println();
			
			//take the current time
			t = System.currentTimeMillis();
			
			//now run the constraint solver, giving it a lot of iterations to play with if it wants
			SConstraintSolver.satisfyConstraints(scpl, 1000);

			System.out.println("Time taken: " + new Integer((int) (System.currentTimeMillis() - t)) + "\n");
			System.out.println("After:");
			printPointState(p1,p2,p3,p4);
			
			
			
			System.out.println("\n\n----------\n\n");
			
			/*
			This test is for SCParallel. It uses two lines and requires them to be parallel.
			*/
			p1 = new SPoint(300, 0, null);
			p2 = new SPoint(1000, -800, null);
			p3 = new SPoint(0, 0, null);
			p4 = new SPoint(500, 500, null);
			
			l1 = new SLine(p1, p2, null);
			l2 = new SLine(p3, p4, null);
						
			//now we create a "package" containing the two lines
			constraintItems = new ArrayList();
			constraintItems.add(l1);
			constraintItems.add(l2);
			
			//and now we create a SCParallel constraint
			SConstraint scprl = new SCParallel(null, constraintItems, (new ArrayList()));
			
			//we now in turn wrap this up into an arraylist so it can be fed to the constraint evaluator
			ArrayList scprll = new ArrayList();
			scprll.add(scprl);
			
			//print some status information
			System.out.println("* SCParallel Test\nBefore:");
			printPointState(p1,p2,p3,p4);
			System.out.println();
			
			//take the current time
			t = System.currentTimeMillis();
			
			//now run the constraint solver, giving it a lot of iterations to play with if it wants
			SConstraintSolver.satisfyConstraints(scprll, 1000);

			System.out.println("Time taken: " + new Integer((int) (System.currentTimeMillis() - t)) + "\n");
			System.out.println("After:");
			printPointState(p1,p2,p3,p4);
			
			
			
			System.out.println("\n\n----------\n\n");
			
			
			/*
			This test is for SCCollinear. It uses four points and requires them to be collinear
			by specifying that two groups of 3 be collinear.
			
			Note: This test currently produces formally correct results, but the results are achieved
			by setting the co-ordinates of 2 out of 3 points to be the same. This makes for correct but not
			necessarily as-expected results.
			*/
			p1 = new SPoint(0, 0, null);
			p2 = new SPoint(800, 0, null);
			p3 = new SPoint(0, 800, null);
			p4 = new SPoint(300, -500, null);
			
			//now we create two "packages" containing three points each
			constraintItems = new ArrayList();
			constraintItems.add(p1);
			constraintItems.add(p2);
			constraintItems.add(p3);
			
			ArrayList constraintItems2 = new ArrayList();
			constraintItems2.add(p2);
			constraintItems2.add(p3);
			constraintItems2.add(p4);
			
			//create two SCCollinear constraints
			SConstraint scc1 = new SCCollinear(null, constraintItems, (new ArrayList()));
			SConstraint scc2 = new SCCollinear(null, constraintItems2, (new ArrayList()));
			
			//SConstraint[] cs = {scc1};
			//renderErrorMap(p1, cs);
			
			
			//wrap them up into an arraylist so it can be fed to the constraint evaluator
			ArrayList sccl = new ArrayList();
			sccl.add(scc1);
			sccl.add(scc2);
			
			//print some status information
			System.out.println("* SCCollinear Test\nBefore:");
			printPointState(p1,p2,p3,p4);
			System.out.println();
			
			//take the current time
			t = System.currentTimeMillis();
			
			//now run the constraint solver, giving it a lot of iterations to play with if it wants
			SConstraintSolver.satisfyConstraints(sccl, 1000);

			System.out.println("Time taken: " + new Integer((int) (System.currentTimeMillis() - t)) + "\n");
			System.out.println("After:");
			printPointState(p1,p2,p3,p4);
			
			
			
			System.out.println("\n\n----------\n\n");

			/*
			This test is for SCVarIsDistanceDelta. It uses two points and measures their change in distance.
			*/
			p1 = new SPoint(0, 0, null);
			p2 = new SPoint(1000, 0, null);
			p3 = new SPoint(0, 0, null);
			p4 = new SPoint(0, 0, null);
			
			distance = new SVariable(0);
			
			//now we create a "package" containing two points and a variable
			constraintItems = new ArrayList();
			constraintItems.add(p1);
			constraintItems.add(p2);
			constraintItems.add(distance);
			
			//create a SCVarIsDistanceDelta constraint
			SConstraint scvidd = new SCVarIsDistanceDelta(null, constraintItems, (new ArrayList()));
			
			//wrap it up into an arraylist so it can be fed to the constraint evaluator
			ArrayList squids = new ArrayList();
			squids.add(scvidd);
			
			//print some status information
			System.out.println("* SCVarIsDistanceDelta Test\nBefore:");
			printPointState(p1,p2,p3,p4);
			System.out.println("Distance = " + new Integer(distance.value()) + "\n");
			
			//now move the first point by 300
			p1.getX().setValue(300);
			
			//take the current time
			t = System.currentTimeMillis();
			
			//now run the constraint solver, giving it a lot of iterations to play with if it wants
			SConstraintSolver.satisfyConstraints(squids, 1000);

			System.out.println("Time taken: " + new Integer((int) (System.currentTimeMillis() - t)) + "\n");
			System.out.println("After:");
			printPointState(p1,p2,p3,p4);
			System.out.println("Distance = " + new Integer(distance.value()));
			
			
			System.out.println("\n\n----------\n\n");
			
			
			/*
			This test is for SCFixedDistance. It uses two points and makes sure their distance remains constant..
			*/
			p1 = new SPoint(0, 0, null);
			p2 = new SPoint(1000, 0, null);
			p3 = new SPoint(0, 0, null);
			p4 = new SPoint(0, 0, null);
						
			//now we create a "package" containing two points
			constraintItems = new ArrayList();
			constraintItems.add(p1);
			constraintItems.add(p2);
			
			//create a SCFixedDistance constraint
			SConstraint scfd = new SCFixedDistance(null, constraintItems, (new ArrayList()));
			
			//wrap it up into an arraylist so it can be fed to the constraint evaluator
			ArrayList scfdl = new ArrayList();
			scfdl.add(scfd);
			
			//print some status information
			System.out.println("* SCFixedDistance Test\nBefore:");
			printPointState(p1,p2,p3,p4);
			System.out.println("Old Distance = 1000\n");
			
			//now move the first point by 300/300
			p1.getX().setValue(300);
			p1.getY().setValue(-300);
			
			//take the current time
			t = System.currentTimeMillis();
			
			//now run the constraint solver, giving it a lot of iterations to play with if it wants
			SConstraintSolver.satisfyConstraints(scfdl, 1000);

			System.out.println("Time taken: " + new Integer((int) (System.currentTimeMillis() - t)) + "\n");
			System.out.println("After:");
			printPointState(p1,p2,p3,p4);
			System.out.println("New Distance = " + new Double(
				java.lang.Math.sqrt((double) (
													(p1.getX().value() - p2.getX().value()) * (p1.getX().value() - p2.getX().value())
													+ (p1.getY().value() - p2.getY().value()) * (p1.getY().value() - p2.getY().value())
													))));
			
			
			System.out.println("\n\n----------\n\n");
			
			
			/*
			This test is for SCErect. It uses one line and requires it to be vertical.
			*/
			p1 = new SPoint(300, 0, null);
			p2 = new SPoint(1000, -800, null);
			p3 = new SPoint(0, 0, null);
			p4 = new SPoint(0, 0, null);
			
			l1 = new SLine(p1, p2, null);
						
			//now we create a "package" containing the line
			constraintItems = new ArrayList();
			constraintItems.add(l1);
			
			//and now we create a SCErect constraint
			SConstraint scer = new SCErect(null, constraintItems, (new ArrayList()));
			
			//we now in turn wrap this up into an arraylist so it can be fed to the constraint evaluator
			ArrayList scerl = new ArrayList();
			scerl.add(scer);
			
			//print some status information
			System.out.println("* SCErect Test\nBefore:");
			printPointState(p1,p2,p3,p4);
			System.out.println();
			
			//take the current time
			t = System.currentTimeMillis();
			
			//now run the constraint solver, giving it a lot of iterations to play with if it wants
			SConstraintSolver.satisfyConstraints(scerl, 1000);

			System.out.println("Time taken: " + new Integer((int) (System.currentTimeMillis() - t)) + "\n");
			System.out.println("After:");
			printPointState(p1,p2,p3,p4);
			
			
			
			System.out.println("\n\n----------\n\n");
			
			
			/*
			This test is for SCInstanceIsXTimesOriginal. It uses one instance and requires it to be 3x the original file.
			*/
			
			SFile f1 = new SFile();
			
			SInstance i1 = new SInstance(f1, new SPoint(0, 0, null), null);
			
			SVariable multiplier = new SVariable(3);
						
			//now we create a "package" containing the instance and point
			constraintItems = new ArrayList();
			constraintItems.add(i1.getScaleFactorVariable());
			constraintItems.add(multiplier);
			
			//and now we create a SCInstanceIsXTimesOriginal constraint
			SConstraint scixto = new SCInstanceIsXTimesOriginal(null, constraintItems, (new ArrayList()));
			
			//we now in turn wrap this up into an arraylist so it can be fed to the constraint evaluator
			ArrayList scixtol = new ArrayList();
			scixtol.add(scixto);
			
			//print some status information
			System.out.println("* SCInstanceIsXTimesOriginal Test\nBefore:\nScale Factor = " + i1.getScaleFactor());
			System.out.println();
			
			//take the current time
			t = System.currentTimeMillis();
			
			//now run the constraint solver, giving it a lot of iterations to play with if it wants
			SConstraintSolver.satisfyConstraints(scixtol, 1000);

			System.out.println("Time taken: " + new Integer((int) (System.currentTimeMillis() - t)) + "\n");
			System.out.println("After:\nScale Factor = " + i1.getScaleFactor());
			
			
			
			System.out.println("\n\n----------\n\n");
			
			
			/*
			This test is for SCInstanceIsXTimesSmallerThanOriginal. It uses one instance and requires it to be 1/3x the original file.
			*/
			
			f1 = new SFile();
			
			i1 = new SInstance(f1, new SPoint(0, 0, null), null);
			
			SVariable divider = new SVariable(3);
						
			//now we create a "package" containing the instance and point
			constraintItems = new ArrayList();
			constraintItems.add(i1.getScaleFactorVariable());
			constraintItems.add(divider);
			
			//and now we create a SCInstanceIsXTimesSmallerThanOriginal constraint
			SConstraint scixtsto = new SCInstanceIsXTimesSmallerThanOriginal(null, constraintItems, (new ArrayList()));
			
			//we now in turn wrap this up into an arraylist so it can be fed to the constraint evaluator
			ArrayList scixtstol = new ArrayList();
			scixtstol.add(scixtsto);
			
			//print some status information
			System.out.println("* SCInstanceIsXTimesSmallerThanOriginal Test\nBefore:\nScale Factor = " + i1.getScaleFactor());
			System.out.println();
			
			//take the current time
			t = System.currentTimeMillis();
			
			//now run the constraint solver, giving it a lot of iterations to play with if it wants
			SConstraintSolver.satisfyConstraints(scixtstol, 1000);

			System.out.println("Time taken: " + new Integer((int) (System.currentTimeMillis() - t)) + "\n");
			System.out.println("After:\nScale Factor = " + i1.getScaleFactor());
			
			
			
			System.out.println("\n\n----------\n\n");
			
			
			/*
			This test is for SCInstanceScaleFactorRelation. It makes sure that the relation in between the scale variable and the actual
			scale of the instance is correct. (variable should be 256*scale factor)
			*/
			//the original points in the file
			p1 = new SPoint(0, 0, null);
			p2 = new SPoint(1000, 0, null);
			//the points in the instance
			p3 = new SPoint(0, 0, null);
			p4 = new SPoint(500, 0, null);
			
			SVariable scaleFactor = new SVariable(512); //double-size
			
			//now we create a "package" containing the four points and the variable
			//note: the co-ordinates in the instance have to be entered first!
			constraintItems = new ArrayList();
			constraintItems.add(p3);
			constraintItems.add(p4);
			constraintItems.add(p1);
			constraintItems.add(p2);
			constraintItems.add(scaleFactor);
			
			//and now we create a SCInstanceScaleFactorRelation constraint
			SConstraint scisfr = new SCInstanceScaleFactorRelation(null, constraintItems, (new ArrayList()));
			
			//we now in turn wrap this up into an arraylist so it can be fed to the constraint evaluator
			ArrayList scisfrl = new ArrayList();
			scisfrl.add(scisfr);
			
			//print some status information
			System.out.println("* SCInstanceScaleFactorRelation Test\nBefore:");
			printPointState(p1,p2,p3,p4);
			System.out.println("Scale Factor = " + new Integer(scaleFactor.value()) + "\n");
			
			//take the current time
			t = System.currentTimeMillis();
			
			//now run the constraint solver, giving it a lot of iterations to play with if it wants
			SConstraintSolver.satisfyConstraints(scisfrl, 1000);

			System.out.println("Time taken: " + new Integer((int) (System.currentTimeMillis() - t)) + "\n");
			System.out.println("After:");
			printPointState(p1,p2,p3,p4);
			System.out.println("Scale Factor = " + new Integer(scaleFactor.value()));
			
			
			System.out.println("\n\n----------\n\n");
		}
		
		catch (Exception e) {
			System.out.println(e);
		}
		
		
    }
	
	/**
	 * This prints out the current state of 4 points.
	*/
	public static void printPointState(SPoint p1, SPoint p2, SPoint p3, SPoint p4) {
			System.out.println(
								"P1 = " + new Integer(p1.getX().value()) + "/" + new Integer(p1.getY().value()) + "\n" +
								"P2 = " + new Integer(p2.getX().value()) + "/" + new Integer(p2.getY().value()) + "\n" +
								"P3 = " + new Integer(p3.getX().value()) + "/" + new Integer(p3.getY().value()) + "\n" +
								"P4 = " + new Integer(p4.getX().value()) + "/" + new Integer(p4.getY().value())
								);
	}
	
	/**
	 * Draws a map of the error values for the given point and constraints, using the SErrorMapRenderer class.
	 * Saves the map to file in "image.png".
	 *
	 * @param p the point to evaluate
	 * @param cs the constraints to use
	*/
	public static void renderErrorMap(SPoint px, SConstraint[] cs) {
		SConstraintSolver.satisfyPoint(px, cs, 10, 10);
			
			
		int jxp[] = SConstraintSolver.jxp;
		int jyp[] = SConstraintSolver.jyp;
			
		for (int i = 0; i < jxp.length; i++) {
			System.out.println(new Integer(jxp[i]) + "/" + new Integer(jyp[i]));
		}
			
		System.out.println("\n" + new Integer(px.getX().value()) + "/" + new Integer(px.getY().value()));
			
		File outputFile = new File("image.png");
		BufferedImage img;
		img = SErrorMapRenderer.renderError(
												px,
												cs,
												-2000, -2000, 600, 600,
												8,
												jxp, jyp
												);
		try {
			ImageIO.write(img, "PNG", outputFile);
		}
		
		catch (Exception e) {
			System.out.println(e);
		}
	}
}
