import java.awt.*;
import java.util.Collection;
import java.util.Iterator;
import java.awt.geom.*;
	
	/**
	 * This class will retrieve information about the 
	 * drawing from the Data & Constraints modules,
	 * and use this information to display the sketch
	 * on the portion  of the screen that the user is
	 * currently viewing. The actual drawing space is
	 * huge, but only the part of the drawing that the
	 * user is viewing has to actually be displayed.
	 * All of the drawing done by this class will take
	 * place on a graphics object passed to this class
	 * by the GUI as an argument in this class's
	 * DrawSketch method.
	 * 
	 * The public methods and public variables used in
	 * this class are static, because this class will 
	 * never actually need to be instantiated, as only 
	 * the sketch will be displayed on the screen at 
	 * any one time. If the user has different drawings 
	 * open on multiple pages, this class only needs 
	 * the data it is being provided by the Data & 
	 * Constraints system to be modified to reflect 
	 * the change - no knowledge of these multiple 
	 * pages is required by the rendering engine.
	 * 
	 * @author Amit Sarna and Tom Botterill
	 */
public class SDrawSketch {

	// default zoom level and max zoom level - zoom changes in factors of two, so can 
	// have zooms at levels 1,2,4,8,16,32 with 1 being the max zoom in, and 32 being the 
	// max zoom out
	private static final int DEFAULTZOOM = 8;         
	private static final int MAXZOOM = 64; 
	private static final int MINZOOM = 1; 
	private static final int SCREENHEIGHT = 400; 
	private static final int SCREENWIDTH = 500; 
	
	public static final int CONSTRAINT_CIRCLE_RADIUS = 15;
	
	private static int mintZoom = 8;
	private static boolean blnThumbnail;
	private static boolean mblnDrawPoints=true;
	private static boolean mblnShowAxes=true;
	// the co-ordinates of the centre of the screen in terms of the page co-ord system
	
	private static SCoords mpageTopLeftCoord;
	
	// the fonts used
	public static Font FONT_USED_FOR_CONSTRAINT = new Font("Arial", Font.BOLD, 24);
	public static Font FONT_USED_FOR_INK = new Font("Arial", Font.PLAIN, 32);
	public static Font FONT_USED_FOR_TEXT = new Font("Arial", Font.PLAIN, 12);
	
	/**
	 * This private constructor prevents any more SDraw classes being created.
	 */
	private SDrawSketch(){};

	/*
	 * This method is responsible for drawing the 
	 * actual sketch on the screen. It will retrieve 
	 * data from the file system about which "files" are
	 * currently  displayed on screen. Sketchpad is 
	 * object oriented so large drawings can be made up 
	 * of smaller drawings (or instances of drawings to 
	 * be precise), which can  in turn also be made up 
	 * of smaller drawings. This method uses recursive 
	 * drawing to traverse these files until the
	 * constituent lines, arcs, points and text have been
	 * found and can be rendered. The size of the page 
	 * on which the user can draw is huge, so instead of
	 * drawing to the whole page every time the drawing
	 * is refreshed, the method will simply be drawing 
	 * to the area of the screen that the user is able to
	 * see. Depending on their current state, e.g. 
	 * selected, in the vicinity of a snapping point etc,
	 * the different constituents of the drawings will
	 * be displayed in different colours
	 * 
	 * The arguments give the Graphics object to be drawn
	 * upon, the zoom level (width of a pixel), the 
	 * coordinates of the top left of the page, whether 
	 * to diplay constraints, and whether this is a
	 * thumbnail which may need us to query the data
	 * engine to get a suitable scale so that the entire
	 * object is shown.
	 *
	 * @param g
	 * @param instanceToDraw
	 * @param blnDrawConstraints
	 * @param blnThumbnail
	 */
	public static void drawSketch(Graphics ge, SFile instanceToDraw, boolean blnDrawConstraints, boolean blnThumbnail){
		//Called by the GUI
		
		//First clear the graphics object
		Graphics2D g = (Graphics2D)ge;
		g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);

		
		g.setColor(Color.WHITE);
		g.fillRect(0,0,SCREENWIDTH,SCREENHEIGHT);
		
		//We don't need this line but it might prevent bugs (nothing will draw until we change the colour)
		g.setColor(SShapeState.SHAPE_DEFAULT.getShapeColor());

	    if(blnThumbnail){	    
	        int intSysHeight=instanceToDraw.getMax_x()-instanceToDraw.getMin_x();
	        int intSysWidth=instanceToDraw.getMax_y()-instanceToDraw.getMin_y();

	        Rectangle rectArea=g.getClipBounds();
			//This won't work until we know the height of the thumbnail graphics object
			
	        //Store zoom and page-top-left coords so we can set it to 1 and don't have to work round it
	        final int intStoreZoom=mintZoom;
	        mintZoom=1;
	        final SCoords scStoreTLCoord=mpageTopLeftCoord;
	        setPageCoordinates(new SCoords(0,0));
	        
			final int THUMBHEIGHT=140;
			final int THUMBWIDTH=140;

			//Calculate a scale factor which will fit the entire drawing on the thumbnail
	        double dblThumbScale=Math.min(THUMBWIDTH/(float)intSysWidth,THUMBHEIGHT/(float)intSysHeight);//the scale which will fit the sfile into the graphics object
	  
	        //Calculate the (system) coordinaes of the top-left corner
	        SCoords scCentre=new SCoords(-(int)(dblThumbScale*instanceToDraw.getMin_x()),-(int)(dblThumbScale*instanceToDraw.getMin_y()));
	        //scCentre=new SCoords((int)(dblThumbScale*instanceToDraw.getMax_x()),(int)(dblThumbScale*instanceToDraw.getMax_y()));
	        drawSFileObject(g,instanceToDraw,0,scCentre,dblThumbScale,false,true,false,null,true);
	        
	        //Reset zoom
	        mintZoom=intStoreZoom;
	        setPageCoordinates(scStoreTLCoord);
	    }else{
	        //TC long t=System.currentTimeMillis();
	        if(mblnShowAxes){
	        	//Draw axes
				g.setColor(Color.lightGray);
	        	drawLine(g,new SCoords(0,-100000),new SCoords(0,100000));
				drawLine(g,new SCoords(100000,0),new SCoords(-100000,0));
				//Why doesn't integer.max_value work???
				//drawLine(g,new SCoords(Integer.MIN_VALUE,0),new SCoords(Integer.MAX_VALUE,0));
	        }
	        drawSFileObject(g,instanceToDraw,0,new SCoords(0,0),1,blnDrawConstraints,true,false,null,false);
	        //TC System.out.println(""+(System.currentTimeMillis()-t)+"ms");
			if(instanceToDraw.myShapes().isEmpty()){
			    //write INK to the screen if there is nothing in the instance
			  	//Draw the text
				g.setColor(SShapeState.SHAPE_DEFAULT.getShapeColor());
				
				g.setFont(FONT_USED_FOR_INK);	
	
			    g.drawString("INK",SCREENWIDTH/2-10,SCREENHEIGHT/2+10);
			  	//System.out.println("INK");
			}
	    }
	}
	
	/**
	 * draws each instance recursively
	 * 
	 * @param g
	 * @param instanceToDraw
	 * @param dblAngle the angle of this instance wrt the screen, not its parent
	 * @param instancePosition the location of the instance within the SYSTEM coordinate system
	 * @param dblScale the scale factor of this instance relative to the screen
	 * @param isInstance true if we're drawing an instance within a master file
	 * @param stateIfInstance The state of an instance
	 */
	private static void drawSFileObject(Graphics g,SFile instanceToDraw,double dblAngle,SCoords instancePosition, double dblScale,boolean blnDrawConstraints,boolean blnDrawSubInstances, boolean isInstance, SShapeState stateIfInstance,boolean blnThumbnail){
		
		//dblScale and dblAngle give the angle and scale factor to transform this instance by.
		//We can apply this rotation and scaling directly to the system coordinates WITHIN the instance
		//We then ADD the instancePosition point to each transformed coordinate
		//For each instance we multiply the scale factor of this and the next instance, and add the angles
		//We transform their position like we do every other coordinate.

	    SIntPair fileTopLeft = null;
	    if (isInstance&&!blnThumbnail){
	        fileTopLeft = new SIntPair(instanceToDraw.getMin_x(), instanceToDraw.getMin_y());
	    } else if(isInstance&&blnThumbnail)
	    {
	    	int tLX = new Double(instanceToDraw.getMin_x()*dblScale).intValue();
	    	int tLY = new Double(instanceToDraw.getMin_y()*dblScale).intValue();
	    	fileTopLeft = new SIntPair(tLX,tLY);
	    }
	    else
	    {
	        fileTopLeft = new SIntPair(0, 0);
	    }
    
		//Get the collection of shapes belonging to this instance 
		if (instanceToDraw==null) return; 
		Collection colShapes = instanceToDraw.myShapes();
		Iterator itShapes = colShapes.iterator();
		
		//Cycle through them drawing each in turn
		while (itShapes.hasNext()){
			SShape ss = (SShape) itShapes.next();
			
			if(isInstance && stateIfInstance != null){
				ss.setState(stateIfInstance);
			}
			
			SShapeState shapeState = ss.getState();
			if(blnThumbnail)g.setColor(Color.BLACK);
			else g.setColor(shapeState.getShapeColor());

			
			if((ss instanceof SPoint) && mblnDrawPoints && !isInstance && !blnThumbnail){
				SPoint sp = (SPoint) ss;
				SCoords sc=scNewPosition(sp, dblAngle, instancePosition, dblScale, fileTopLeft);
				drawPoint(g,sc,!sp.isFixed());
			}else if(ss instanceof SLine){		   		
				//Get the end-points of the line from the SLine object
				SLine sl=(SLine) ss;
				SPoint spLineEnds[]=sl.getEnds();
				
				//Get the transformed COORDINATE (SCoord) of a SPoint
				SCoords sc_a,sc_b;
				sc_a=scNewPosition(spLineEnds[0], dblAngle, instancePosition, dblScale, fileTopLeft);
				sc_b=scNewPosition(spLineEnds[1], dblAngle, instancePosition, dblScale, fileTopLeft);
			   	//Draw the line
				//System.out.println("drawing file "+instanceToDraw.getName());
			   	drawLine(g,sc_a,sc_b);
			   	
			}else if(ss instanceof SArc){
				//Get the points of the arc from the SArc object
			  	SArc sa=(SArc)ss;
			  	
			   	//Get the transformed COORDINATE (SCoord) of a SPoint
			  	SCoords scCentre,scStart,scEnd;
			  	scCentre=scNewPosition(sa.getCentre(), dblAngle, instancePosition, dblScale, fileTopLeft);
			  	scStart=scNewPosition(sa.getStart(), dblAngle, instancePosition, dblScale, fileTopLeft);
			  	scEnd=scNewPosition(sa.getEnd(), dblAngle, instancePosition, dblScale, fileTopLeft);
			  	boolean blnIsAnticlockwise=sa.isAnticlockwise();	   	
			  	
			   	//Draw the arc
			   	drawArc(g,scCentre,scStart,scEnd,blnIsAnticlockwise);
			  	
			}else if(ss instanceof SLabel){
				//DO WE DRAW TEXT ON INSTANCES WITHIN THE MAIN INSTANCE??
				//This should catch SNumber as well I think--NEEDS TESTING!!
				//Get the position of the text from the SLabel object
			  	SLabel slab=(SLabel)ss;
			  	
			   	//Get the transformed COORDINATE (SCoord) of a SPoint
			  	SCoords scTopLeft;
			  	scTopLeft=scNewPosition(slab.getTopLeft(), dblAngle, instancePosition, dblScale, fileTopLeft);
				
			  	//Draw the text
			  	drawText(g,scTopLeft,slab.getText());
				
			}else if (ss instanceof SNumber){
			    //Get the position of the text from the SNumber object
			    SNumber snum=(SNumber)ss;
			  	
			   	//Get the transformed COORDINATE (SCoord) of a SPoint
			  	SCoords scTopLeft;
			  	scTopLeft=scNewPosition(snum.getTopLeft(), dblAngle, instancePosition, dblScale, fileTopLeft);
				
			  	//Draw the text
			  	drawText(g, scTopLeft, ""+(snum.getIntValue()));
			  	
			}else if((ss instanceof SInstance) && blnDrawSubInstances){
				
				SInstance si=(SInstance)ss;
				
				//Work out the angle of the instance relative to the screen:
				SVariable svDegAng=si.getAngle();
				int intDegreeAngle=svDegAng.getValue();
				double dblInstAngle=dblAngle+Math.toRadians((double)intDegreeAngle);
				
				//Work out the position relative to the screen
			  	SCoords scInstPos;
			  	if(isInstance) scInstPos = instancePosition;
			  	scInstPos=scNewPosition(si.getLocation(), dblAngle, instancePosition, dblScale, fileTopLeft);
				
			  	//Work out the scale factor
				double dblInstScale=dblScale*si.getScaleFactor();
				//We do not draw constraints on sub-instances, it would be too cluttered
				drawSFileObject(g,si.getFile(),dblInstAngle,scInstPos,dblInstScale,false,true,true,si.getState(),blnThumbnail);
			}
			
			// if user wants to see constraints draw them
			else if ((ss instanceof SConstraint) && (blnDrawConstraints==true)) {
				// Display the constraint on the screen 
				DrawConstraint(g, (SConstraint)ss);				
			}
		}
	}

	/**
	 * 
	 * draws a line on g from sc1 to sc2
	 * 
	 * @param g
	 * @param sc1
	 * @param sc2
	 */
	
	private static void drawPoint(Graphics g, SCoords sc,boolean filled){
		SIntPair pointPos = SDrawSketch.pageToScrCoord_new(sc);
		if(filled){
			g.fillOval(pointPos.x-2,pointPos.y-2,4,4);
		}
		else
		{
			g.drawOval(pointPos.x-2,pointPos.y-2,4,4);
		}
	}
	private static void drawLine(Graphics g, SCoords sc1,SCoords sc2){
		//This translates into screen coordinates and draws a line on the screen
		SIntPair sipEnd1, sipEnd2;
		sipEnd1=SDrawSketch.pageToScrCoord_new(sc1);
		sipEnd2=SDrawSketch.pageToScrCoord_new(sc2);
		//System.out.println("Drawing line from:"+sipEnd1.x+","+sipEnd1.y);
		//System.out.println("Drawing line to:"+sipEnd2.x+","+sipEnd2.y);
		g.drawLine(sipEnd1.x,sipEnd1.y,sipEnd2.x,sipEnd2.y);
	}
	/**
	 * Draws a circle arc on g with centre at scCentre, starting at scStart, and finishing in line with scEnd
	 * 
	 * @param g
	 * @param scCentre
	 * @param scStart
	 * @param scEnd
	 * @param blnIsAnticlockwise
	 */
	private static void drawArc(Graphics g,SCoords scCentre,SCoords scStart,SCoords scEnd,boolean blnIsAnticlockwise){
		//This translates into screen coordinates and draws a circle-arc on the screen
	    int temp=scCentre.getX().getValue();
	    
		SIntPair sipCentre;
		sipCentre=SDrawSketch.pageToScrCoord_new(scCentre);
				
		int intRadius=intPageDistance(scStart,scCentre);

		//Calculate the coordinates of the box enclosing the circle containing the arc
		int intTopLeftX=sipCentre.x-intRadius;
		int intTopLeftY=sipCentre.y-intRadius;//This might be -, it depends on which way the y axis goes
		int intSquareLength=2*intRadius;
		
		//Calculate the angles of the start of the arc, and the total angle
		double dblAng[]=dblArcAngles(scCentre,scStart,scEnd,blnIsAnticlockwise);
		int intStartAngle=(int)Math.toDegrees(dblAng[0]);
		int intArcAngle=(int)Math.toDegrees(dblAng[1]);
		
		g.drawArc(intTopLeftX,intTopLeftY,intSquareLength,intSquareLength,intStartAngle,intArcAngle);
	}

	/**
	 * Writes strText to g with BOTTOM LEFT coordinate scTopLeft
	 * 
	 * @param g
	 * @param scTopLeft
	 * @param strText
	 */
	private static void drawText(Graphics g,SCoords scTopLeft,String strText){
		//This translates into screen coordinates and draws the text on the screen
		SIntPair sipTopLeft,sipBottomLeft;
		sipTopLeft=SDrawSketch.pageToScrCoord_new(scTopLeft);
		sipBottomLeft=sipTopLeft;//##THIS NEEDS FIXING!!
		//The drawstring function takes a bottom left coordinate
		//System.out.println(sipBottomLeft.x+", "+sipBottomLeft.y+"");
		//Write the text!
		g.setFont(FONT_USED_FOR_TEXT);
		g.drawString(strText,sipBottomLeft.x+5,sipBottomLeft.y+5);
	}
	
	/**
	 * Rotates and translates points within sub-instances until they are in the correct position within this instance
	 * 
	 * @param sp
	 * @param dblAngle
	 * @param instancePosition
	 * @param dblScale
	 * @param topLeft The top left coord of the file we're drawing an instance of.
	 * @return the SYSTEM coordinate of the point translated and rotated to its correct position relative to the instance currently being displayed
	 */
	private static SCoords scNewPosition(SPoint sp,double dblAngle,SCoords instancePosition, double dblScale, SIntPair topLeft){
		SCoords sc=sp.getLocation();
		sc=scScale(sc,dblScale,new SCoords(topLeft));//Scale about 0,0 (the instance's centre)
		//If the rotation and scale where composed into a matrix first the error might be smaller.
		
		sc=scRotate(sc,dblAngle,new SCoords(0,0));//Rotate about 0,0 (the instance's centre)
		
		//sc=scAdd(sc,instancePosition);//Move to the position within the parent instance
		//System.out.println(sc.getX().getValue());
		//sc = scSubtract(sc,instancePosition);
		int xDifference = getIntX(instancePosition)-topLeft.x;
		int yDifference = getIntY(instancePosition)-topLeft.y;
		
		sc = scAdd(sc, new SCoords(topLeft));
		sc = scAdd(sc,new SCoords(xDifference,yDifference));
		//sc = scSubtract(sc, instancePosition);
		//System.out.println(sc.getX().getValue());
		return sc;
	}
	
	/**
	 * Rotates a SCoord relative to (0,0) in the ANTICLOCKWISE direction
	 * The rotation matrix corresponds to one for rotating by -dblAngle
	 * (so the signs are the wrong way round)
	 * @param sc
	 * @param dblAngle
	 * @return
	 */
	private static SCoords scRotate(SCoords sc,double dblAngle,SCoords offset){
		int trueX = -getIntX(offset)+getIntX(sc);
		int trueY = -getIntY(offset)+getIntY(sc);
		int intNewX=(int)(trueX*Math.cos(dblAngle)+trueY*Math.sin(dblAngle));
		int intNewY=(int)(-trueX*Math.sin(dblAngle)+trueY*Math.cos(dblAngle));
		return new SCoords(intNewX,intNewY);
	}
	
	/**
	 * Scale a SCoords relative to (0,0)
	 * @param sc
	 * @param dblScale
	 * @return 
	 */
	private static SCoords scScale(SCoords sc,double dblScale,SCoords offset){
		int trueX = -getIntX(offset)+getIntX(sc);
		int trueY = -getIntY(offset)+getIntY(sc);
		int intNewX=(int)(trueX*dblScale);
		int intNewY=(int)(trueY*dblScale);
		return new SCoords(intNewX,intNewY);
	}
	
	/**
	 * Adds 2 SCoord points (vector addition)
	 * @param sc1
	 * @param sc2
	 * @return sc1+sc2
	 */
	private static SCoords scAdd(SCoords sc1,SCoords sc2){
		return new SCoords(getIntX(sc1)+getIntX(sc2),getIntY(sc1)+getIntY(sc2));
	}
	
	/**
	 * Subtracts 2 SCoord points (vector addition)
	 * @param sc1
	 * @param sc2
	 * @return sc1-sc2
	 */
	private static SCoords scSubtract(SCoords sc1,SCoords sc2){
		return new SCoords(getIntX(sc1)-getIntX(sc2),getIntY(sc1)-getIntY(sc2));
	}
	
	/**
	 * Retrieves the x coordinate from a SCoords object
	 * @param sc
	 * @return The x coordinate
	 */
	private static int getIntX(SCoords sc){
		SVariable sv=sc.getX();
		return sv.getValue();
	}
	
	/**
	 * Retrieves the y coordinate from a SCoords object
	 * @param sc
	 * @return The y coordinate
	 */
	private static int getIntY(SCoords sc){
		SVariable sv=sc.getY();
		return sv.getValue();
	}
	
	/************************************************************************************
	* Gets the screen to zoom in on the current drawing by changing the scale factor.
	* Nothing will change on the screen until the DrawSketch function is next called. 
	*
	* @return true if can zoom in any more after this zoom
	************************************************************************************/
	public static boolean ZoomIn() 
      { 
        mintZoom = mintZoom / 2;               // set current zoom level
		if (mintZoom<MINZOOM) mintZoom = MINZOOM;
		// System.out.println(mintZoom);
        if(mintZoom == 1) return false;	     // check if at maxiumum zoom in	
        return true;
      }
      

	/************************************************************************************
	* Gets the screen to zoom out of the current drawing by changing the scale factor. 
	*
	* @return true if can zoom out any more after this zoom
	************************************************************************************/
    public static boolean ZoomOut() 
      {
      	mintZoom = mintZoom * 2;
		if (mintZoom>MAXZOOM)mintZoom = MAXZOOM;
		if(mintZoom == MAXZOOM) return false;  // check if at max zoom out
		return true;
	  }
	
	/**
	 * Sets the coordinates of the Top Left corner of the screen. You need to call DrawSketch as well
	 * to redraw the screen.
	 * 
	 * @param new pageTopLeftCoord
	 */
	public static void setPageCoordinates(SCoords newpageTopLeftCoord){
		//Sets the current page TOP LEFT coordinates
		mpageTopLeftCoord=newpageTopLeftCoord;
		//TC System.out.println("Setting page coords to "+getIntX(mpageTopLeftCoord)+" " +getIntY(mpageTopLeftCoord));
		//mpageTopLeftCoord=new SCoords(0,0);
	}
	/* The GUI sets these properties when it wants to
	 * zoom in or out or move.
	 */
	
	/**
	 * This turns the drawing of points on or off. It's for debugging.
	 * 
	 * @param blnNewShowPoints true=on, false=off
	 */
	public static void setShowPoints(boolean blnNewShowPoints){
		mblnDrawPoints=blnNewShowPoints;
	}
	
	/**
	 * Gets the state of show points
	 * 
	 */
	public static boolean getShowPoints(){
		return mblnDrawPoints;
	}
	
	/**
	 * Turns the drawing of axes on or off. It's for debugging but might turn into a feature later...
	 * 
	 * @param blnNewShowPoints
	 */
	public static void setShowAxes(boolean blnNewShowAxes){
		mblnShowAxes=blnNewShowAxes;
	}
	
	/**
	 * Finds out whether the drawing axis are shown
	 * 
	 * @returns true or false
	 */
	public static boolean getShowAxes(){
		return mblnShowAxes;
	}

	
	/* The screen co-ordinates are different from the page
	 * co-ordinates. The screen co-ordinates refer to the 
	 * co-ordinates of the portion of the page that the 
	 * user is currently able to see on screen. The page 
	 * co-ordinates are the co-ordinates of the graphics 
	 * object on which all the drawing is taking place.  
	 * The functions in this class convert between the two. 
	 * These functions are used by both the this class and 
	 * classes from the GUI
	*/

	/**
	 * Converts from system coordinates to the coordinates of points
	 * on the screen
	 * 
	 * @param scFrom
	 * @return the page coordinate corresponding to the system coordinate SCFrom
	 */
	public static SIntPair pageToScrCoord_new(SCoords scFrom){
		// Converts System coordinates to Display Coordinates
		// This uses integer division, which is ok I think,
		// except this function is NOT the inverse of pageToScrCoords
		int intDisplayX = (scFrom.getX().value()+mpageTopLeftCoord.getX().value())/mintZoom;
		int intDisplayY = (scFrom.getY().value()+mpageTopLeftCoord.getY().value())/mintZoom;
		
		return new SIntPair(intDisplayX,intDisplayY);
	}
	/**
	 * Calculate the distance in PIXELS between two points in system coordinates
	 * 
	 * @param scFrom
	 * @param scTo
	 * @return the distance in pixels from scFrom to scTo
	 */
	private static int intPageDistance(SCoords scFrom,SCoords scTo){
		SCoords sc;
		sc=scSubtract(scFrom,scTo);
		int intX=getIntX(sc);
		int intY=getIntY(sc);
		return (int)(Math.sqrt(intX*intX+intY*intY)/mintZoom);
	}
	
	/**
	 * Calculates the angle of a point ANTICLOCKWISE 
	 * from the horizontal +ve x axis.
	 * 
	 * @param sc
	 * @return the angle between the horizontal x axis and sc
	 * in the ANTICLOCKWISE direction
	 */
	private static double dblAngle(SCoords sc){
		
		//Math.atan2(y,x) returns the angle of point (x,y) CLOCKWISE from
		//the +ve x axis...
		double dblAng = Math.atan2((double)getIntY(sc),(double)getIntX(sc));
		
		//...so negate it
		if (dblAng < 0) return -dblAng;
		else return 2*Math.PI-dblAng;
		//Force it to be positive and return 2*Math.PI if the angle is zero (so circles should draw ok)
	}
	
	/**
	 * Calculates the angle of the start point and total angle of
	 * a circle arc. g.drawArc needs the angle anticlockwise
	 * from the +ve x axis.
	 * 
	 * @param scCentre
	 * @param scTo
	 * @param scFrom
	 * @param blnIsAnticlockwise
	 * @return A pair consisting of the angle of the start point from
	 * the circle centre, and the angle from SCFrom to SCTo about SCCentre
	 * in the blnIsAnticlockwise direction. 
	 */
	private static double[] dblArcAngles(SCoords scCentre,SCoords scTo,SCoords scFrom,boolean blnIsAnticlockwise){
		if(scTo==scFrom){
			double dblCircleAngles[]={0.0,2*Math.PI};
			return dblCircleAngles;
		}
		SCoords scStart,scEnd,sc;
		
		//THIS MIGHT NEED SWAPPPING AGAIN!
		if(blnIsAnticlockwise){
			scStart=scFrom;
			scEnd=scTo;
		}else{
			scStart=scTo;
			scEnd=scFrom;
		}
		//Translate points so we can calculate their angle realtive to (0,0)
		scStart=scSubtract(scStart,scCentre);
		scEnd=scSubtract(scEnd,scCentre);
		
		//Calculate the angle of the start point
		double dblStartAngle=dblAngle(scStart);
		
		//rotate scEnd clockwise by dblStartAngle
		//The angle of sc gives the angle between scStart and scEnd
		sc=scRotate(scEnd,-dblStartAngle,new SCoords(0,0));
		double dblArcAng[]={dblStartAngle,dblAngle(sc)};
		return dblArcAng;
	}
	
	/**
	 * Converts from coordinates of points on the screen to
	 * system coordinates 
	 *  
	 * @param sipFrom
	 * @return the system coordinate of a point on the page
	 */
	public static SCoords scrToPageCoord_new(SIntPair sipFrom){
		// Converts System coordinates to Display Coordinates
		// This uses integer division, which is ok I think
		int intSystemX = sipFrom.x * mintZoom - mpageTopLeftCoord.getX().value();
		int intSystemY = sipFrom.y * mintZoom - mpageTopLeftCoord.getY().value();
		//TC System.out.println("Top left is "+mpageTopLeftCoord.getX().value()+", "+mpageTopLeftCoord.getY().value());
		return new SCoords(intSystemX,intSystemY);
	}
	

	/*****************************************************************************************
	* The user may want a graphical display of constraints on the drawing. This method 
	* iterates through the collection of points that a constraint applies to and then draws 
	* lines from these points to the point at which the constraint is to be centered. Since 
	* a constraint is of type SConstraint which is a subclass of SShape, it has its own 
	* co-ordinates. The user will be able to drag the constraints around on the screen to make
	* the drawing more organised, hence changing these co-ordinates. A circle is drawn at the
	* constraint's handle co-ordinates containing a letter that signifies the type of the 
	* constraint applied to the points
	*	
	* @param  g The graphics object on which to draw the file
	* @param f The sketchpad file to be displayed
	*****************************************************************************************/
    private static void DrawConstraint(Graphics g, SConstraint sc)
    {
		if (sc.getState() != SShapeState.SHAPE_DEFAULT)
				g.setColor(sc.getState().getShapeColor());
		else 
			    g.setColor(Color.lightGray); // constraints are displayed in a light grey colour
	    
		g.setFont(FONT_USED_FOR_CONSTRAINT);	
	
		// variables to hold the handle co-ordinates and the location of a point to which
		// the constraint applies
		SCoords hc, ploc; 
	
		Iterator pts = (sc.getLinkedPoints()).iterator(); // get the points to which the constraint applies
		hc = sc.getHandleLocation(); // get the constraint handle

		int hcx = pageToScrCoord_new(hc).x;
		int hcy = pageToScrCoord_new(hc).y;

		while(pts.hasNext())	
		  {
		     ploc = ((SPoint) pts.next()).getLocation();
			 g.drawLine(pageToScrCoord_new(ploc).x, pageToScrCoord_new(ploc).y, hcx, hcy); // draw lines between points and handle 
		  }

		CircleByRadius(g, pageToScrCoord_new(hc), CONSTRAINT_CIRCLE_RADIUS, g.getColor()); // draw a circle of radius 20 for the constraint symbol        

		 // String s = s.getConsName(); // get the symbol to display for this particular constraint
		 String s = sc.getSymbolDyn();
		
		 FontMetrics fm   = g.getFontMetrics(FONT_USED_FOR_CONSTRAINT);

		 // get the dimensions of the symbol based on the font in which it will be displayed
		 Rectangle2D rect = fm.getStringBounds(s, g);
		 int textHeight = (int)(rect.getHeight()); 
		 int textWidth  = (int)(rect.getWidth());

		 // Work out screen co-ord positions based on the dimensions
		 int x = hcx - (textWidth/2);
		 int y = hcy - (textHeight/2) + fm.getAscent();

		if ( (s.equals("q")) || (s.equals("p")) ) {
			g.drawString(s, x, y - FONT_USED_FOR_CONSTRAINT.getSize() / 6);  // Draw the symbol with an offset upwards
		} else {
			g.drawString(s, x, y);  // Draw the symbol
		}       
 	  
		} // end inner while

	/** 
	 * Helper method for displaying the constraint circle using just its center co-ordinates
	 * and its radius
	 */
	private static void CircleByRadius(Graphics g, SIntPair sc, int r, Color Outline) {
   		g.setColor(Color.white); // clear the area where the symbol will be displayed
   		g.fillOval(sc.x - r, sc.y - r, 2*r, 2*r);       
         
   		// draw an outline to the circle 	
   		g.setColor(Outline);
   		g.drawOval(sc.x -(r-1), sc.y -(r-1), 2*(r-1), 2*(r-1));
  }

	/**
	 * Colours the pixels that, if clicked on in the GUI, would result in
	 * s being selected.
	 * @author Tom Craig
	 */
	public static void drawCoordRange(Graphics g, SShape s)
	{
	    g.setColor(Color.green);
	    Collection coordRange = s.getCoordinateRange();
	    Iterator i = coordRange.iterator();
	    while (i.hasNext())
	    {
	        SIntPair c = (SIntPair) i.next();
	        
	        int x = c.x + mpageTopLeftCoord.getX().value();
	        int y = c.y + mpageTopLeftCoord.getY().value();
	        
	        g.drawLine(x / mintZoom, y / mintZoom, x / mintZoom, y / mintZoom);
	    }
	}
	
	/**
	 * Get the zoom level.
	 * @author Tom Craig
	 */
	public static int getZoom()
	{
	    return mintZoom;
	}

}
