import java.awt.*;
import java.awt.image.*;

/**
 * The function of this class is to render maps that show the error values for the position of a point under one or several constraints.
 * This is purely for constraint system algorithm analysis purposes, and as such, this class need not be included in the final shipping version.
 *
 * @author David Stark
 * @version 1.0
 */
public class SErrorMapRenderer {


	/**
	 * @param thePoint the point for which the error map should be rendered
	 * @param theConstraints[] the constraints whose errors should be displayed
	 * @param top the topmost co-ordinate of the area considered, in units
	 * @param left the leftmost co-ordinate of the area considered, in units
	 * @param width the width of the image in pixels
	 * @param height the height of the image in pixels
	 * @param zoomfactor the factor by which the image is zoomed: units / zoomfactor = pixels
	 * @param jxp array of x-positions of jumps
	 * @param jyp array of y-positions of jumps
	*/
	public static BufferedImage renderError(	SPoint thePoint,
												SConstraint theConstraints[],
												int top, int left, int width, int height,
												double zoomfactor,
												int jxp[], int jyp[]
	) {
		//RETRIEVE THE VARIABLES WE'LL BE WORKING WITH
		SVariable xVar = thePoint.getX();
		SVariable yVar = thePoint.getY();
		SVariable[] vars = {xVar, yVar};
		
		//CREATE THE ARRAY THAT IS GOING TO BE PAINTED (oo memory munchy munchy)
		int data[][] = new int[width][height];
		int maxerr = 0;
		int minerr = 2000000;
		int pos[] = new int[2];
		int err = 0;
		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {
				//calculate the current position in units
				pos[0] = pixel2unit(x, left, zoomfactor);
				pos[1] = pixel2unit(y, top, zoomfactor);
				data[x][y] = 1;
				for (int i = 0; i < theConstraints.length; i++) {
					err = theConstraints[i].errorForValues(vars, pos);
					data[x][y] += err;
				}
				//check if this error is a new maximum
				maxerr = java.lang.Math.max(maxerr, data[x][y]);
				minerr = java.lang.Math.min(minerr, data[x][y]);
			}
		}
		
		
		//HAVING DONE SO, WE NOW CREATE AN IMAGE TO FILL WITH THE SCALED VALUES OF THE ARRAY
		//COMMENTED-OUT LINES ARE VARIATIONS
		BufferedImage img = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
		ColorModel cmodl = img.getColorModel();
		
		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {
				//img.setRGB(x, y, (data[x][y] * 255) / maxerr);
				
				//double maxerr2 = (10 * java.lang.Math.log((double) maxerr));
				
				//int q = (int) (( (10 * java.lang.Math.log((double) data[x][y])) * 255) / maxerr2);
				
				int maxerr2 = (int) java.lang.Math.sqrt((double) (maxerr * 100));
				
				int q = 255 - ((int) java.lang.Math.sqrt((double) (data[x][y] * 100)) * 1024) / maxerr2;
				
				int[] i = {q,q,q};
				
				//int q = error2colour(data[x][y], maxerr);
				
				/* <<= NORMAL MODE!
				int maxerr2 = (int) java.lang.Math.sqrt((double) (maxerr * 100));

				if (data[x][y] < (minerr + 1)) {
					int[] i = {255, 0, 0};
					img.setRGB(x, y, cmodl.getDataElement(i, 0));
				} else {
					int q = 255 - ((int) java.lang.Math.sqrt((double) (data[x][y] * 100)) * 255) / maxerr2;
					int[] i = {q,q,q};
					img.setRGB(x, y, cmodl.getDataElement(i, 0));
				}
				*/

				
				img.setRGB(x, y, cmodl.getDataElement(i, 0));
				//img.setRGB(x, y, 0);//-2147483648);//(x + y*width) * 10 - 1800000);
			}
		}
		
		//NOW WE (MAYBE) RENDER THE JUMPS
		if ((jxp.length > 1) && (jxp.length == jyp.length)) {
			Graphics2D g = img.createGraphics();
			
			g.setColor(Color.RED);
		
			for (int i = 0; i < jxp.length -1; i++) {
				g.drawLine(		unit2pixel(jxp[i], left, zoomfactor),
								unit2pixel(jyp[i], top, zoomfactor),
								unit2pixel(jxp[i+1], left, zoomfactor),
								unit2pixel(jyp[i+1], top, zoomfactor)
							);
			}
		}
		
		return img;
		//return null;
	}
	
	/**
	 * Converts pixel co-ordinates into unit co-ordinates.
	 * @param pixel the pixel co-ordinate to convert
	 * @param cornerpos the corner position, or offset ("top" or "left")
	 * @param zoomfactor the zoom factor
	*/
	private static int pixel2unit(int pixel, int cornerpos, double zoomfactor) {
		return (int)(((double) pixel) * zoomfactor) + cornerpos;
	}
	
	/**
	 * Converts unit co-ordinates to pixel co-ordinates.
	 * @param unit the unit co-ordinate to convert
	 * @param cornerpos the corner position, or offset ("top" or "left")
	 * @param zoomfactor the zoom factor
	*/
	private static int unit2pixel(int unit, int cornerpos, double zoomfactor) {
		return (int) ((double) (unit - cornerpos) / zoomfactor);
	}
	
	/**
	 * Re-curves the error values so that the first 10% account for 50% of the range,
	 * and the other 90% for the other 50%.
	 *
	 * @param errValue the error value to be re-curved
	 * @param maxValue the maximum error value, required for scale
	*/
	private static int error2colour(int errValue, int maxValue) {
		if (errValue < (maxValue / 10)) {
			return (errValue * 128) / (maxValue / 10);
		} else {
			return ((errValue - maxValue / 10) * 127) / (maxValue * 9 / 10) + 128;
		}
	}
}
