/**
 * A file, or master. Instances of a file may be created in other drawings.
 * Remember, a file, like all other D&C classes, knows nothing about screen co-ordinates - it deals in page co-ordinates only.
 *
 * @author Tom Craig
 * @see SInstance
 */

import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;

public class SFile implements Cloneable, Serializable
{
    
    
	/**
	 * The collection of SShapes in this file. Use a HashSet to make adding and removing quick.
	 */
	private HashSet myShapes = new HashSet();
	
	/**
	 * A map of (SIntPair, LinkedList of SShape) pairs, with an SIntPair for each grid coordinate.
	 * 
	 * We store this information to make it easy to find the shape under the cursor (which we need
	 * even when using handles, because the handles of a given shape are only displayed when
	 * the cursor is within a certain distance of that shape).
	 * 
	 * We use LinkedLists because ordering is important, and removing should be quick.
	 * Unfortunately, it does mean there's no automatic checking for duplicates.
	 */
	private HashMap shapesByCoord = new HashMap();
	
	/**
	 * The undo button object that can undo constraint satisfaction.
	*/
	private SCSolverUndo undoButton;
	
	/**
	 * The current cursor position.
	 */
	private SIntPair cursorPosition = new SIntPair(0, 0);
	
	/**
	 * The name of the file.
	 */
	private String name;
	
	/**
	 * A collection of SPoints that are the attacher points.
	 */
	private ArrayList attachers = new ArrayList();
	
	/**
	 * For the sake of sanity, we say that a file may have no more than this number of attachers.
	 * Otherwise, imagine the potential problems when snapping-to!
	 */
	public static int MAX_NUM_ATTACHERS = 2;
	
	/**
	 * The shape under the cursor. Is updated in setCursorPosition.
	 */
	private SShape shapeUnderCursor = null;
	
	/**
	 * The selection threshold value in page units. When getting what is under the cursor,
	 * a circle of radius (THRESHOLD_VALUE * file's zoom level) is searched, centred on
	 * the exact cursor position.
	 */
	private int THRESHOLD_VALUE = 10;
	
	/**
	 *  The top left co-ordinates of the bounding box of shapes that are copied 
	 */
 	private SIntPair topLeftOfCopyBox = new SIntPair(0,0);


	/**
	 * Top left of copying box used to calculate offset when pasting objects
	 * @param SIntPair - the coordinates of the top left corner of the copy box
	 */
	public void setTopLeftOfCopyBox(SIntPair tl)
	{
		topLeftOfCopyBox.x = tl.x;
		topLeftOfCopyBox.y = tl.y;
	}

	//
	// NAVIGATION
	//
	
	/**
	 * Determines whether some co-ordinates are within the bounding box of a shape.
	*/
	private boolean isWithinBoundingBox(SIntPair pos, SShape theShape, int threshold) {
		SIntPair topLeft = new SIntPair(0,0);
		SIntPair bottomRight = new SIntPair(0,0);
		theShape.getBoundingBox(topLeft, bottomRight);
		return
				(pos.x > topLeft.x - threshold) &&
				(pos.y > topLeft.y - threshold) &&
				(pos.x < bottomRight.x + threshold) &&
				(pos.y < bottomRight.y + threshold);
	}
	
	/**
	 * Finds and returns the shape that is closest to being under the cursor, within some threshold value.
	 */
	public SShape findWhatIsUnderCursor()
	{
		deleteMarkedShapes();
	    int zoomThreshold = THRESHOLD_VALUE * SDrawSketch.getZoom();
		
		//first we need a list of items where the cursor is within the bounding box
		ArrayList candidates = new ArrayList();
		Iterator shpIter = myShapes().iterator();
		while (shpIter.hasNext()) {
			SShape currentShape = (SShape) shpIter.next();
			if ( !(currentShape instanceof SConstraint) || SketchUI.showConstraints ) {
				if (isWithinBoundingBox(cursorPosition, currentShape, zoomThreshold)
					&& (currentShape.getState() != SShapeState.SHAPE_BEING_CREATED)
					&& (currentShape.getState() != SShapeState.SHAPE_BEING_MOVED)) 
				{
					candidates.add(currentShape);
				}
			}
		}
		
		
		
		if (candidates.size() == 0) {
			return null; //nothing found
		}
		
		
		//having found a list of initial candidates, find the nearest object
		//note: points have precedence over everything else, and everything has precedence over instances unless their distance is literally 0
		shpIter = candidates.iterator();
		SShape bestShape = null;
		double[] bestDistance = {100000,100000,100000,100000}; //set this to a high enough initial value
		int bestShapeRating = 2; // 2 is an instance, 1 an arc or line and 0 a point
		
		while (shpIter.hasNext()) {
			SShape currentShape = (SShape) shpIter.next();
						
			if (currentShape instanceof SPoint) {	
				//the new arrival is a point and wins by default
				double currentDistance = currentShape.minDistanceFromCoords(cursorPosition);
				if ((currentDistance<bestDistance[0] && currentDistance <= zoomThreshold) || bestShapeRating>0)
				{
					bestDistance[0] = currentDistance;
					bestShape = currentShape;
					bestShapeRating = 0;
				}
			}
			else if(bestShapeRating>0)
			{
				if(currentShape instanceof SLine
				|| currentShape instanceof SArc
				|| currentShape instanceof SNumber
				|| currentShape instanceof SLabel
				)
				{
					double currentDistance = currentShape.minDistanceFromCoords(cursorPosition);
					if ((currentDistance<bestDistance[1] && currentDistance <= zoomThreshold) && bestShapeRating>=1)
					{
						bestDistance[1] = currentDistance;
						bestShape = currentShape;
						bestShapeRating = 1;
					}	
				}
				else if (bestShapeRating>1)
				{
					//compare
					double dist = currentShape.minDistanceFromCoords(cursorPosition);
					if (dist == 0) {
						bestShape = currentShape;
						bestShapeRating = 2;
						bestDistance[2] = dist;
					}
				}	
			}
		}
		
		return bestShape;
	}
	
	/**
	 * Gets the shape that is closest to being under the cursor, within some threshold value.
	 * 
	 * <b>Implementation detail:</b>
	 * Assumes the private function findWhatIsUnderCursor has already been called at this cursor position.
	 * This is done in setCursorPosition.
	 * 
	 * @return The shape under the cursor, or null if none is found.
	 */
	public SShape whatIsUnderCursor()
	{
	    if (shapeUnderCursor != null && shapeUnderCursor.getState() != SShapeState.SHAPE_BEING_CREATED)
	        return shapeUnderCursor;
	    else
	        return null;
	}
	
	/**
	 * Selects the shape supplied as a parameter.
	 * @param ref The shape to be selected.
	 */
	public void selectShape(SShape ref)
	{
	    if (ref != null)
	    {
	        System.out.println("TC: Selecting shape: " + ref);
	        ref.setState(SShapeState.SHAPE_IS_SELECTED);
	    }
	}
	
	/**
	 * Deselects the shape supplied as a parameter.
	 * @param ref The shape to be deselected.
	 */
	public void deselectShape(SShape ref)
	{
	    if (ref != null && ref.getState() == SShapeState.SHAPE_IS_SELECTED)
	    {
	        System.out.println("TC: Deselecting shape: " + ref);
	        ref.setState(SShapeState.SHAPE_DEFAULT);
	    }
	}
	
	/**
	 * Deselects all selected shapes.
	 */
	public void deselectAll()
	{
	    Iterator i = myShapes.iterator();
	    while (i.hasNext())
	        deselectShape((SShape) i.next());
	}
	
	/**
	 * @return The shapes in the file that are selected.
	 */
	public ArrayList getSelectedShapes()
	{
	    return getShapesInState(SShapeState.SHAPE_IS_SELECTED);
	}
	
	/**
	 * @param state The state of shapes we want.
	 * @return The shapes in the supplied state.
	 */
	public ArrayList getShapesInState(SShapeState state)
	{
	    ArrayList ret = new ArrayList();
	    
	    Iterator i = myShapes.iterator();
	    while (i.hasNext())
	    {
	        SShape s = (SShape) i.next();
	        if (s.getState() == state)
	            ret.add(s);
	    }
	    
	    return ret;
	}
	
	//
	// CREATING NEW SHAPES
	//
	
	/**
	 * Do the stuff that needs doing for every new shape:
	 * 		<li> add to myShapes
	 * 		<li> set its state to "being created";
	 * 		<li> work out which co-ordinates it occupies.
	 * 
	 * The last can't be done in the SShape constructor because it needs the shape to be fully
	 * constructed first. So we may as well do the first two here, too.
	 * 
	 * All the create* functions below should call initialiseShape. 
	 */
	private void initialiseShape(SShape shape)
	{
	    myShapes.add(shape);
	    shape.setState(SShapeState.SHAPE_BEING_CREATED);	    
	    updateCoordinateRange(shape, true);
	}
	
	/**
	 * Commits a shape to the canvas.
	 */
	private boolean dismemberPoint = false;
	public SShape commitShape(SShape shape)
	{
	    System.out.println("Committing " + shape);
	    
		boolean allowSnap = false;
		if(shape.getState()==SShapeState.SHAPE_BEING_MOVED) allowSnap = true;
		if(shape.getState()==SShapeState.SHAPE_BEING_CREATED &! dismemberPoint) allowSnap = true;

		if(shape instanceof SPoint && allowSnap){
			//int zoomThreshold = THRESHOLD_VALUE * SDrawSketch.getZoom();
			//LinkedList allShapesUnderCursor = getShapesHere(cursorPosition, zoomThreshold, 1 );
			//if (allShapesUnderCursor.size()!=0){
			if (shapeUnderCursor != null) {
				System.out.println("do I snap?");
				//SShape cShape = (SShape) allShapesUnderCursor.get(0);
				if (shapeUnderCursor instanceof SPoint){
					SShape oldShape = shape;
					shape = shapeUnderCursor;
					oldShape.setState(SShapeState.SHAPE_HAS_BEEN_DELETED);
					System.out.println(oldShape);
					System.out.println(shape);
				}
			}	
		}
		
		// if we're committing the endpoint of an arc, keep it on the arc
	    if (shape instanceof SPoint)
	        keepPointOnArc((SPoint) shape);
		
	    // if we're committing an instance, generate its attachers
	    if (shape instanceof SInstance)
	        ((SInstance) shape).getAttachers();
 
       
	    
		shape.setState(SShapeState.SHAPE_DEFAULT);
		return shape;
	}
	
	/**
	 * Creates a new point at the current cursor position.
	 * @return A reference to the point just created.
	 */
	public SPoint createPoint()
	{
	    return createPoint(false);
	}
	
	/**
	 * Creates a new point at the current cursor position, specifying whether it is the attacher of
	 * some instance.
	 * 
	 * @return A reference to the point just created.
	 */
	public SPoint createPoint(boolean isAttacher)
	{
	    SPoint point = new SPoint(new SCoords(cursorPosition.x, cursorPosition.y), this, isAttacher);
	    initialiseShape(point);
	    return point;
	}
	
	/**
	 * Creates a new line.
	 * 
	 * @param p1 An endpoint of the line.
	 * @param p2 The other endpoint.
	 * @return A reference to the line just created.
	 */
	public SLine createLine(SPoint p1, SPoint p2)
	{
	    SLine line = new SLine(p1, p2, this);
	    
	    initialiseShape(line);
	    return line;
	}
	
	/**
	 * Creates a new arc. Forces p2 to be the same distance from centre as p1 is, if this is not already the case.
	 * 
	 * @param centre The centre point of the arc.
	 * @param p1 An endpoint of the arc.
	 * @param p2 The other endpoint.
	 * @param direction Whether the arc goes anticlockwise.
	 * @return A reference to the arc just created.
	 */
	public SArc createArc(SPoint centre, SPoint p1, SPoint p2, boolean anticlockwise)
	{
	    SArc arc = new SArc(centre, p1, p2, anticlockwise, this);
	    initialiseShape(arc);
	    return arc;
	}
	
	/**
	 * Creates a new label.
	 * 
	 * @param ref The shape this label is labelling.
	 * @param text The text of the label.
	 * @return A reference to the label just created.
	 */
	public SLabel createLabel(SPoint topLeft,String text)
	{
	    // create at the current cursor position the point that will represent this label
	    
	    // now create the label there
	    SLabel label = new SLabel(topLeft, text, this);
	    initialiseShape(label);
	    return label;
	}
	
	/**
	 * Creates a new number.
	 * 
	 * @param ref The shape this number is describing.
	 * @param type The type of information this number is displaying. (For example, the length of a line.)
	 * @param value The value of the number, if applicable.
	 * @param maybe If true, the point is not created, but the renderer may display the consequences of creating it.
	 * @return A reference to the label just created.
	 */
	public SNumber createNumber(SPoint topLeft, int value /*, SNumberValueType type*/)
	{
		// Create point in which to place SNumber
	
		SNumber number = new SNumber((SPoint)topLeft, value,this);
	    initialiseShape(number);
	    return number;
	}
	
	/**
	 * Adds a new constraint.
	 * 
	 * @param constraintClass The type of constraint to create.
	 * @param refs An ArrayList of shapes that the constraint refers to. (An ArrayList instead of a Collection is used because ordering may be important.)
	 * @return a reference to the constraint just created
	 * @throws SConstraintCouldNotBeCreatedException
	 */
	public SConstraint createConstraint(ArrayList refs, Class constraintClass) throws SConstraintCouldNotBeCreatedException
	{
	    SConstraint constraint = null;		// we'll return this
	    ArrayList moreShapes = new ArrayList();
	    // a constraint must not refer to any deleted or potential shapes
	    for (int i = 0; i < refs.size(); i++)
	    {
        	if (refs.get(i) instanceof SShape)
	        {
		        SShape s = (SShape) refs.get(i);
		        
		        if (	s.getState() == SShapeState.SHAPE_HAS_BEEN_DELETED
		             || s.getState() == SShapeState.SHAPE_BEING_CREATED)
					throw new SConstraintCouldNotBeCreatedException("Cannot refer to a shape that is being created or has been deleted");
	        }
	    }
	    
	    // now try to create the constraint
	    try
	    {
	    	
		    // get the constructor of constraintClass that takes a single ArrayList
		    Class[] constructorArgTypes = { SFile.class, ArrayList.class, ArrayList.class /* was refs.getClass() */ };
		    Constructor constraintConstructor = constraintClass.getConstructor(constructorArgTypes);
		    
		    // instantiate constraintClass using the constructor represented by constraintConstructor
		    Object[] constructorArgs = { this, refs, moreShapes };
		    constraint = (SConstraint) constraintConstructor.newInstance(constructorArgs);
		    // now initialise the shape
		    initialiseShape(constraint);
	    }
	    catch (Exception x)
	    {
	        System.out.println(x);
	        throw new SConstraintCouldNotBeCreatedException("Reflection failed in SFile.createConstraint");
	    }
	    
	    return constraint;
	}
	
	/**
	 * Creates a new instance of a given file.
	 * 
	 * @param file Reference to the file to make an instance of.
	 * @return A reference to the newly created instance.
	 * @throws SFileNotInLibraryException
	 */
	public SInstance createInstance(SFile f) throws SFileNotInLibraryException
	{
	    // create at the current cursor position the point that will represent this label
	    SPoint topLeft = createPoint();
	    commitShape(topLeft);
	    
	    // now create the instance there
	    /*
	    SIntPair tl = new SIntPair(0,0);
	    SIntPair br = new SIntPair(0,0);
	    // get top left of bounding box of instance
		getBoundingBox(tl, br, f.myShapes());
	    	    
	    tl.setValues(tl.x, tl.y);
	    */
	    	    
	    SInstance instance = new SInstance(f, topLeft, this);
	    
	    /* Create and commit the attachers.
	     * Beware, this code is quite hacky: 
	     */
	    /*Collection instanceAttachers = instance.getAttachers();
	    System.out.println("" + instance + " has " + instanceAttachers.size() + " attachers.");
	    Iterator i = instanceAttachers.iterator();
	    while (i.hasNext())
	    {
	        SPoint p = (SPoint) i.next();
	        initialiseShape(p);
	        p.setState(SShapeState.SHAPE_DEFAULT);	// a hack to stop commitShape merging p!
	        p.moveShape(0 - instance.getFile().getMin_x(), 0 - instance.getFile().getMin_y());
	        commitShape(p);
	    }*/
	    	     
	    // now, finally, initialise the instance
	    initialiseShape(instance);
	    return instance;
	}
	
	//
	// MODIFYING SHAPES
	//
	
	/**
	 * Delete the current co-ordinates of the parameter 'shape' from this.shapesByCoord,
	 * and add its new ones instead.
	 * 
	 * @param shape The shape whose co-ordinate range we're reading.
	 * @param isNew True if the shape is new and we therefore don't need to delete its current co-ordinates from shapesByCoord, because it's not in there.
	 */
	public void updateCoordinateRange(SShape shape, boolean isNew)
	{
	    if (!isNew)
	    {
		    // first delete the current co-ordinates by cycling through shapesByCoord (this is slooow!!)
		    Collection c = shapesByCoord.values();			// this is a collection of LinkedList
		    Iterator i = c.iterator();
		    while (i.hasNext())
		    {
		        LinkedList shapes = (LinkedList) i.next();	// this is a collection of shapes for a given co-ordinate
		        while (shapes.remove(shape))				// remove the shape passed as an argument, if present
		        { /* nothing */ }
		    }
	    }
	    
	    // now obtain the new co-ordinate range of the shape and add it to shapesByCoord 
	    Collection coordRange = shape.getCoordinateRange();
	    if (coordRange != null)
	    {
		    Iterator i = coordRange.iterator();
		    while (i.hasNext())
		    {
		        SIntPair coordUnderConsideration = (SIntPair) i.next();
		        
		        // obtain the LinkedList of shapes already at this co-ordinate (creating it if necessary)
		        LinkedList shapesAlreadyAtThisCoord = (LinkedList) shapesByCoord.get(coordUnderConsideration);
		        if (shapesAlreadyAtThisCoord == null)
		        {
		            shapesAlreadyAtThisCoord = new LinkedList();
		            shapesByCoord.put(coordUnderConsideration, shapesAlreadyAtThisCoord);
		        }
		        
		        // now add this shape to the set of shapes at this co-ordinate
		        shapesAlreadyAtThisCoord.add(shape);
		        
		        // extend the bounds of the picture if this co-ordinate is beyond them
		        /*if (coordUnderConsideration.x < min_x)
		            min_x = coordUnderConsideration.x;
		        if (coordUnderConsideration.x > max_x)
		            max_x = coordUnderConsideration.x;
		        if (coordUnderConsideration.y < min_y)
		            min_y = coordUnderConsideration.y;
		        if (coordUnderConsideration.y > max_y)
		            max_y = coordUnderConsideration.y;*/
		    }
	    }
	}
	
	/**
	 * Constrains the point to be on at least one of the arcs (if any) of which it is an endpoint.
	 * 
	 * @param flexibleEndpoint The point that should lie on the arc.
	 */
	public void keepPointOnArc(SPoint flexibleEndpoint)
	{
	    // loop through the shapes that are dependent on flexibleEndpoint, and do stuff for each that is an arc
	    // (N.B. if more than one is an arc, this may not keep the point on every arc)
	    Iterator d = flexibleEndpoint.getDependentShapes(this.myShapes).iterator();
	    while (d.hasNext())
	    {
	        SShape s1 = (SShape) d.next();
	        if (s1 instanceof SArc)
	        {
	            SArc arc = (SArc) s1;
		        
			    // do nothing if 'point' isn't an endpoint of the 'arc'
			    if (/*arc.getStart() == flexibleEndpoint ||*/ arc.getEnd() == flexibleEndpoint)
			    {			    
				    // fixedEndPoint := the endpoint of the arc that is NOT 'point'
				    SPoint fixedEndpoint = arc.getStart();
				    SPoint centre = arc.getCentre();
				    
					// check distance from centre to each of the endpoints is equal
				    if (   (int) centre.getLocation().distanceBetweenCoords(flexibleEndpoint.getLocation())
				        != (int) centre.getLocation().distanceBetweenCoords(fixedEndpoint.getLocation()))
				    {
				        // if we get in here, then it's not
				        // 	so make flexibleEndpoint same distance from centre as fixedEndpoint is,
				        // 	but without changing the angle
				        //		(i.e. keeping flexibleEndpoint on line between the old flexibleEndpoint and the centre)
				        
				        // r1 := distance from centre to fixed
				        // r2 := distance from centre to flexible
				        // sf := scale factor
				        // dx := diff in x-coords between flexible and centre
				        // dy := diff in y-coords between flexible and centre
				        double r1_sqd = SIntPair.cartesianDistanceSquared(centre.getLocation().makeIntPair(), fixedEndpoint.getLocation().makeIntPair());
				        double r2_sqd = SIntPair.cartesianDistanceSquared(centre.getLocation().makeIntPair(), flexibleEndpoint.getLocation().makeIntPair());
				        double sf = Math.sqrt(r1_sqd / r2_sqd);
				        double dx = flexibleEndpoint.getX().value() - centre.getX().value();
				        double dy = flexibleEndpoint.getY().value() - centre.getY().value();
				        
				        // set the new x- and y-coords of flexible
				        double xnew = dx * sf + centre.getX().value();
				        double ynew = dy * sf + centre.getY().value();
				        flexibleEndpoint.getLocation().getX().setValue((int) xnew);
				        flexibleEndpoint.getLocation().getY().setValue((int) ynew);
				        
				        // update the coordinate ranges of everything depending on flexibleEndpoint
				        Iterator i = myShapes.iterator();
				        while (i.hasNext())
				        {
				            SShape s2 = (SShape) i.next();
				            if (s2.getPoints().contains(flexibleEndpoint))
				                updateCoordinateRange(s2, true);
				        }
				    }
				}
	        }
	    }
	}
	
	/**
	 * Returns whether the user may move this shape. False if shape is a point and is an attacher.
	 */
	private boolean mayBeMoved(SShape p)
	{	    
	    boolean ret = (!(p instanceof SPoint) || !((SPoint) p).isAttacher());
	    if (!ret)
	        System.out.println("" + p + " may not be moved!");
	    return ret;
	}
	
	/**
	 * Moves the given shape to new co-ordinates. Once the shape is fully moved, the GUI
	 * should call dropShape(ref).
	 * 
	 * @param ref The shape to move.
	 * @param dx The amount of horizontal movement, in integer units.
	 * @param dy The amount of vertical movement, in integer units.
	 */
	public void moveShape(SShape ref, int dx, int dy)
	{
	    if (mayBeMoved(ref))
	    {
		    // set shape state to "being moved"
		    ref.setState(SShapeState.SHAPE_BEING_MOVED);
		    
		    // move the shape
		    if (dx != 0 || dy != 0)
		        ref.moveShape(dx, dy);
	    }
	    
	    /* N.B. We could call "updateCoordinateRange(ref, false);" here, but we don't because:
	     * 		(a) it is inefficient (as this method may be called many times for each move)
	     * 		(b) the coordinate range is only needed for finding out what is under the cursor, but:
	     * 			 (i)  once it has begun a move, the GUI should keep track of the shape it is moving
	     * 			 (ii) and if it didn't, and instead kept calling whatIsUnderCursor, it might not get
	     * 				  the same shape back each time! 
	     * 		(c) it is done upon dropping the shape.
	     */
	}
	
	/**
	 * Moves the given shape to new co-ordinates. Once the shape is fully moved, the GUI
	 * should call dropShape(ref).
	 * 
	 * @param ref The shape to move.
	 * @param x The new x-coord.
	 * @param y The new y-coord.
	 */
	public void moveShapeAbs(SShape ref, int x, int y)
	{
	    if (mayBeMoved(ref))
	    {
		    // set shape state to "being moved"
		    ref.setState(SShapeState.SHAPE_BEING_MOVED);
		    
		    // move the shape
		    ref.moveShapeAbs(x, y);
	    }
	}
	
	/**
	 * Resizes the instance.
	 */
	public void resizeShape(SInstance ref, double scaleFactor)
	{
		if(scaleFactor>0)ref.setScaleFactor(scaleFactor);
	    // set shape state to "being moved"
	    ref.setState(SShapeState.SHAPE_BEING_MOVED);
	}
	
	/**
	 * Rotates the given shape by the angle dt, where the shape's handle h is the centre of rotation
	 * and dt is the anticlockwise angle from p1 to p2 about h.
	 * 
	 * @param ref The shape to rotate.
	 */
	public void rotateShape(SShape ref, SIntPair p1, SIntPair p2)
	{
	    // set shape state to "being moved"
	    ref.setState(SShapeState.SHAPE_BEING_MOVED);
	    
	    // use the cosine rule to work out the angle A between p1 and p2 about ref's handle, where:
	    //		h = coords of handle
	    //		a = sq'd dist from p1 to p2 (we want cos A)
	    //		b = sq'd dist from p1 to h
	    //		c = sq'd dist from p2 to h
	    //		x_sqd = x^2 for all variables x
	    // and then convert A to degrees
	    //
	    double A = Math.atan2(p1.x-p2.x,p1.y-p2.y);
	    int angle = (int) (A / (2 * Math.PI) * (double) SInstance.MAX_ANGLE);
	    angle = angle+180;
	    // now (finally!) rotate the shape and update its coord range
	    ref.rotateShape(angle);
	    System.out.println(angle);
	    updateCoordinateRange(ref, true);
	}
	
	/**
	 * Called by the GUI to signal that the shape has finished being moved.
	 * 
	 * @param ref The shape to drop.
	 * @author Andrew Smee
	 */
	public void dropShape(SShape ref)
	{
	    ref.setState(SShapeState.SHAPE_DEFAULT);
	    updateCoordinateRange(ref, false);
	     
    
	    Collection pts = ref.getPoints();
	    
  	    // if we're moving the endpoint of an arc, keep it on the arc

		Iterator ptsit = pts.iterator();
		while(ptsit.hasNext())
		{
	     SPoint pt =  (SPoint) ptsit.next();
	    
	     keepPointOnArc(pt);
	        
	        // need to keep end point on Arc also, so find it if this point
	        // is on an arc and put it there
   	        Iterator i = pt.getDependentShapes(myShapes).iterator();
	        while(i.hasNext())
	        {
	        	SShape s = (SShape) i.next();
	        	
	        	if(s instanceof SArc)
	        	  	keepPointOnArc(((SArc) s).end);
   	     	}
	        
	    }
	 	
	 	    
	    // merge with first shape in state BEING_SNAPPED_TO (if any)
	    ArrayList mergeCandidates = getShapesInState(SShapeState.SHAPE_BEING_SNAPPED_TO);
	    if (mergeCandidates.size() > 0)
	    {
	        SShape mergeWithMe = (SShape) mergeCandidates.get(0);
	        
	        // can only merge unequal shapes that are of the same type
	        if (	mergeWithMe != ref
	            &&	mergeWithMe.getClass() == ref.getClass()
	            	// don't merge instances: too complicated
	            &&	mergeWithMe.getClass() != SInstance.class)
	        {
	            System.out.println("Probably going to merge " + ref + " with " + mergeWithMe);
	            
	            SShape s = ref;
	            
	            if(mergeWithMe!=s&&s instanceof SPoint && mergeWithMe instanceof SPoint)
				{
	                System.out.println("Case 1");
	                
					SPoint point = (SPoint) s;
					Collection refs = ((SPoint)mergeWithMe).getDependentShapes(myShapes());
					Iterator iter = refs.iterator();
					while(iter.hasNext()){
						SShape iterShape = (SShape) iter.next();
						if (!(iterShape instanceof SPoint)){
							iterShape.replacePoints((SPoint)mergeWithMe,(SPoint) s);
						}	
					}
					
					this.deleteShape(mergeWithMe);
				}
	            
				if (mergeWithMe !=s &&(s instanceof SLabel ||s instanceof SNumber)&& mergeWithMe instanceof SPoint)
				{
				    System.out.println("Case 2");
				    
					Collection refs = ((SPoint)mergeWithMe).getDependentShapes(myShapes());
					Iterator iter = refs.iterator();
					while(iter.hasNext()){
						SShape iterShape = (SShape) iter.next();
						if (!(iterShape instanceof SPoint)){
							iterShape.replacePoints((SPoint)mergeWithMe,(SPoint) s);
						}
					}
					
					this.deleteShape(mergeWithMe);
				} 
				
				if (mergeWithMe !=s &&!(s instanceof SPoint)&&mergeWithMe instanceof SLabel)
				{
				    System.out.println("Case 3");
				    
					Collection refs = ((SLabel)mergeWithMe).getTopLeft().getDependentShapes(myShapes());
					Iterator iter = refs.iterator();
					while(iter.hasNext()){
						SShape iterShape = (SShape) iter.next();
						if (!(iterShape instanceof SPoint)){
							iterShape.replacePoints(((SLabel)mergeWithMe).getTopLeft(),(SPoint) s);
						}
					}
					
					this.deleteShape(mergeWithMe);
				}
				
				if (mergeWithMe !=s &&!(s instanceof SPoint)&&mergeWithMe instanceof SNumber)
				{
				    System.out.println("Case 4");
				    
					Collection refs = ((SNumber)mergeWithMe).getTopLeft().getDependentShapes(myShapes());
					Iterator iter = refs.iterator();
					while(iter.hasNext()){
						SShape iterShape = (SShape) iter.next();
						if (!(iterShape instanceof SPoint)){
							iterShape.replacePoints(((SNumber)mergeWithMe).getTopLeft(),(SPoint) s);
						}
					}
					
					this.deleteShape(mergeWithMe);
				}
	        }
	    }

	}
	
	public void dismember(SPoint p)
	{
		System.out.println(p);
		Collection shapes = p.getDependentShapes(myShapes);
		Iterator sIter = shapes.iterator();
		System.out.println(shapes.size());
		while(sIter.hasNext())
		{
			SShape s = (SShape)sIter.next();
			SPoint newPoint =  createPoint();
			dismemberPoint = true;
			commitShape(newPoint);
			dismemberPoint = false;
			System.out.println(newPoint);
			s.replacePoints(p,newPoint);
		}
		deleteShape(p);
	}
	
	
	/**
	 * Deletes the given shape and anything associated with it.
	 * @param ref The shape to delete.
	 */
	public void deleteShape(SShape ref)
	{
	    ArrayList a = new ArrayList();
	    a.add(ref);
	    deleteShapes(a);
	}
	
	/**
	 * Deletes the given shapes and anything associated with them.
	 * @param refs The shapes to delete.
	 */
	public void deleteShapes(Collection refs)
	{
	    markShapesAsDeleted(refs);
	    deleteMarkedShapes();
	}

	/**
	 * Marks as deleted all shapes in refs and anything directly associated with them.
	 * A private function used only by deleteShapes
	 * 
	 * @param refs The shapes to mark as deleted.
	 */
	private void markShapesAsDeleted(Collection refs)
	{
	    // things that are deleted because of shapes in the parameter
	    // set being deleted	    
	    HashSet dependentDeletes = new HashSet();
	    

		// go through and delete all of the shapes
	    Iterator i = refs.iterator();
	    while (i.hasNext())
	    {	    	    
	    	    
	        SShape s = (SShape) i.next();
	        // delete the shape
	        s.setState(SShapeState.SHAPE_HAS_BEEN_DELETED); 
	                      
	        // get the shapes points - these will need to be deleted if they 
	        // are not connected to anything else
	        Collection s_pts = s.getPoints();
	        if (s instanceof SConstraint){
	        	s_pts = ((SConstraint) s).getLinkedPoints();
	        }
	        
	        // go through the points
   	        Iterator j = s_pts.iterator();
	        while (j.hasNext())
	        {
	        	
	        	// get the shapes this point is connected to
	            SPoint p = (SPoint) j.next();
                Collection deps = p.getDependentShapes(this.myShapes);
	                       
	                        
	            // if the only thing dependent on this point is the point
	            // being deleted then delete the point 	            
	            if (deps.size() == 0 || (deps.size() == 1 && deps.contains(s)))
	            {
	                p.setState(SShapeState.SHAPE_HAS_BEEN_DELETED);
	            }
	            else
	            {
	            	// otherwise, check for constraints that may need to be deleted
	            		            		            	
	            	Iterator h = deps.iterator();
	            	while(h.hasNext())
	            	{
	            		
	            		SShape pShape = (SShape) h.next();

	            		
	            		//if this is the point being deleted, then delete its associated
	            		//constraints
	            		if ((pShape instanceof SConstraint) && (s instanceof SPoint))
	            		{      				            			
	            			
	            			pShape.setState(SShapeState.SHAPE_HAS_BEEN_DELETED);	
	            			
	            		}
	            		
	            		if((pShape instanceof SConstraint) && (s instanceof SLine))
	            		{
	            			            			
	            			// if the shape being deleted is a line, check special constraint delete cases
	            			SLine sl = (SLine) s;
	            			SConstraint sc = (SConstraint) pShape;
	            		
	            			// see if the constraint applies to the line
	            			ArrayList scShapes = sc.myShapes(); 
	            			
	            			for(int k = 0; k < scShapes.size(); k++)
	            			{
	            				SShape ss = (SShape) scShapes.get(k);
	            				if(ss instanceof SLine) 
	            				{
	            					
	            				    SLine checkLine = (SLine) ss; 
	            		  			if ((checkLine.end_a.getX().value() == sl.end_a.getX().value()) &&
	            				   (checkLine.end_a.getY().value() == sl.end_a.getY().value()) &&
	            				   (checkLine.end_b.getX().value() == sl.end_b.getX().value()) &&
	            				   (checkLine.end_b.getY().value() == sl.end_b.getY().value()))
	            				   {
	            				   		// if the constraint applies to the line being deleted
	            				   		// then delete it
	            				   	
	            						pShape.setState(SShapeState.SHAPE_HAS_BEEN_DELETED);
	            						break;
	            				   }
	            				
	            				}	   
	            				   
	            				   
	            			}
	            			
	            			
	            			
	            			
	            		}
	            		
	            		
	            		if(s instanceof SPoint)
	            		{	            				            			
	            			// need to delete all dependent shapes of a point that is being 
	            			// deleted
	            			if(pShape.getState() != SShapeState.SHAPE_HAS_BEEN_DELETED)
	            				dependentDeletes.add(pShape);
	            		}
	            		            		         		           			            		
	            		            		
	            	 }
	            	 
	            	 
	            	 // check if the only dependent shapes were the constraints and s itself maybe
	            	 boolean constraintsOnly = true;
	            	 h = deps.iterator();
	            	 while(h.hasNext())
	            	 {
	            	 	SShape temp = (SShape) h.next();
	            	 	if(!(temp instanceof SConstraint) && (temp != s))
	            	 	  {            	 	  	            	 	  	
	            	 	  	   	 	  	
	            	 	    constraintsOnly = false; break; 
	            	 	  }
	            	 }
	            	 
	            	 // iff there are no other dependent shapes, delete this point
	            	 if(constraintsOnly) p.setState(SShapeState.SHAPE_HAS_BEEN_DELETED);
	            	 	            	 
	            }
	        }
	        
	        // delete the shape from the set if it has been added
	        
	      
	      dependentDeletes.remove(s);
	        
	    }
	    
	    if(!(dependentDeletes.isEmpty()))
	        markShapesAsDeleted(dependentDeletes);
	    
	}
	
	/**
	 * Removes from myShapes and potentialShapes all shapes that are marked as deleted.
	 * A private function used only by deleteShapes.
	 */
	private void deleteMarkedShapes()
	{	    
	    // delete from myShapes all shapes with state SHAPE_HAS_BEEN_DELETED
   	    Iterator i = myShapes.iterator();
	    while (i.hasNext())
	    {
	        SShape shape = (SShape) i.next();
	        if (shape.getState() == SShapeState.SHAPE_HAS_BEEN_DELETED)
	            i.remove();
	    }
    }
	
	/**
	 * Copies the given shapes to be centred under the cursor position.
	 * 
	 * @param refs The shapes to copy.
	 */
	public void copyAndPasteTheseShapes(Collection refs)
	{
		// copy location should set top left of copy box
		// therefore can't really do paste with this, need keyboard
				
		SIntPair tl = new SIntPair(0,0);
		SIntPair br = new SIntPair(0,0);
	    getBoundingBox(tl, br, refs);		
	    
	    // set offset based on top left of copy box and mouse location	   
	    SPoint.setOffset(calculateOffset());
	    	    
	    // Clear the copy points hash map 	     
	    SPoint.emptyCpyPts();
	    
	         	       	         	       	         	       
	    Iterator i = refs.iterator();
	    // go through and clone all of the shapes passed as parameters
	    while(i.hasNext())
	    {
	    	if(i instanceof SPoint)
	    	{
	    		((SPoint) i.next()).clone();
	    	}
	    	if (!(i instanceof SConstraint))   // as long as i isn't a constraint, clone it
	    	{
	    	   	SShape cln = (SShape)  (((SShape) i.next()).clone());
	    		initialiseShape(cln);
	    		commitShape(cln);
	    	}
	    }
	    
	    
		// get all of the points that have been copied - these were not added first because
		// points can be shared between shapes, and these only need to be added once
		// between the multiple shapes
	    Collection points = SPoint.getCopyPoints();
	    
	    Iterator ptsit = points.iterator();
	    while(ptsit.hasNext())
	    {
	    	SPoint pt = (SPoint) ptsit.next();
	    	initialiseShape(pt);
	    	commitShape(pt);
	    }
           
	}

	/**
	 * Makes an exact copy of the shapes passed as a parameter - this is so that 
	 * if the user changes the original stuff that has been copied, then the 
	 * shapes in the "clipboard" are still the same as the original ones rather
	 * than the changed ones
	 * 
	 * @ param the shapes to copy
	 * @ returns an ArrayList containing the shapes to copy
	 */
	public static ArrayList preliminaryCopy(Collection refs)
	{
		    
		SPoint.setOffset(new SIntPair(0,0)); // no offset for exact copy
		
	     // Clear the copy points hash map 	     
	    SPoint.emptyCpyPts();
	    ArrayList copyShapes = new ArrayList();     	       
	         	       	         	       	         	       
	  	// copy the shapes
	    Iterator i = refs.iterator();
	    while(i.hasNext())
	    {
	    	if(i instanceof SPoint)
	    	{
	    		((SPoint) i.next()).clone();
	    	}
	    	if (!(i instanceof SConstraint))   // as long as i isn't a constraint, clone it
	    	{
	    	   	SShape cln = (SShape)  (((SShape) i.next()).clone());
	    		copyShapes.add(cln);
	    	}
	    }
	    
	    
	    // add the points into the copy
	    Collection points = SPoint.getCopyPoints();
	    
	    Iterator ptsit = points.iterator();
	    while(ptsit.hasNext())
	    {
	    	SPoint pt = (SPoint) ptsit.next();
	    	copyShapes.add(pt);
	    }
	    
	    	                  	
	    return copyShapes; // retunr the copied shapes
	}
	
	/**
	 * calculate the offset that all objects in a paste should be pasted at
	 * based on the current mouse position and the original position of the
	 * top left corner of the copy box
	 *
	 * @returns SIntPair which contains the x and y offset values
	 */
	public SIntPair calculateOffset()
	{		
		
		int x = cursorPosition.x - topLeftOfCopyBox.y; 
		int y = cursorPosition.y - topLeftOfCopyBox.y;
		
		return new SIntPair(x, y);
	}
	
	
	/**
	 * Modifies the attacher status of a given point.
	 * 
	 * @param point The point to modify.
	 * @param attacher Whether the point should be an attacher.
	 */
	public void setAttacherState(SPoint point, boolean attacher) throws STooManyAttachersException
	{
	    // TODO: Changing attachers means instances must change too
	    
	    if (attacher && !attachers.contains(point))
	    {
	        if (attachers.size() < SFile.MAX_NUM_ATTACHERS)
	            attachers.add(point);
	        else
	            throw new STooManyAttachersException("Too many attachers");
	    }
	    else if (!attacher)
	        attachers.remove(point);
	}
	
	/**
	 * Activated or deactivated the fixed state of a point.
	 * 
	 * @param point The point to modify.
	 * @param fixed Whether the point should be fixed.
	 */
	public void setFixedState(SPoint point, boolean fixed)
	{
	    point.setFixed(fixed);
	}
	
	//
	// FUNCTIONS FOR RENDERING
	//
	
	/**
	 * @return All shapes in the file.
	 */
	public Collection myShapes()
	{
	    return myShapes;
	}
	
	//
	// ACCESSORS AND MUTATORS
	//
	
	/**
	 * Sets the current scale of the drawing.
	 * (TC: I've forgotten why we need this. I'm sure I'll work it out soon :) )
	 * 
	 * @param scale The number of integer units per pixel.
	 */
	//public abstract void setScale(int scale);
	
	/**
	 * Sets the current position of the cursor.
	 * Note that the units are integer units in absolute co-ordinates, not pixels.
	 * 
	 * @param x The x-coord of the cursor.
	 * @param y The y-coord of the cursor.
	 */
	public void setCursorPosition(int x, int y)
	{
	    cursorPosition.x = x;
	    cursorPosition.y = y;
	    
	    SShape newShapeUnderCursor = findWhatIsUnderCursor();
	    
	    // the previous shape under the cursor is no longer being hovered over or snapped to
	    // (implication: only one or zero shapes at a time are being hovered over or snapped to)
	    if (	shapeUnderCursor != null
	         && shapeUnderCursor != newShapeUnderCursor
	         && (	shapeUnderCursor.getState() == SShapeState.SHAPE_IS_BEING_HOVERED_OVER
	             ||	shapeUnderCursor.getState() == SShapeState.SHAPE_BEING_SNAPPED_TO))
	    {
	        shapeUnderCursor.setState(SShapeState.SHAPE_DEFAULT);
	    }

	    if (newShapeUnderCursor != null)
	    {
	        ArrayList shapesBeingMoved = getShapesInState(SShapeState.SHAPE_BEING_MOVED);
	        
	        // the new shape under the cursor is being hovered over, unless it's in some other special (non-default) state
		    if (	shapesBeingMoved.size() == 0
		        &&	newShapeUnderCursor.getState() == SShapeState.SHAPE_DEFAULT)
		    {
		        newShapeUnderCursor.setState(SShapeState.SHAPE_IS_BEING_HOVERED_OVER);
		        
		        // is the point the handle of an instance? if so, hover over that too
		        // TODO
		    }
		    
		    // merge?
		    if (	shapesBeingMoved.size() > 0
		        &&	shapesBeingMoved.get(0).getClass() == newShapeUnderCursor.getClass())
		    {
		        newShapeUnderCursor.setState(SShapeState.SHAPE_BEING_SNAPPED_TO);
		    }

		    // Matthew's code for snapping (N.B. TC: this uses handle locn and so works properly only for points)
	    	/*cursorPosition.x = newShapeUnderCursor.getHandleLocation().getX().getValue();
	    	cursorPosition.y = newShapeUnderCursor.getHandleLocation().getY().getValue();*/
	    }
	    
	    // finally, update the variable holding the shape under the cursor with the new one
	    shapeUnderCursor = newShapeUnderCursor;
	}
	
	/**
	 * @return What the file believes is the current cursor position.
	 */
	public SIntPair getCursorPosition()
	{
	    return cursorPosition;
	}
	
    /**
     * @return Returns the name.
     */
    public String getName()
    {
        return name;
    }
    
    /**
     * @param name The name to set.
     */
    public void setName(String name)
    {
        this.name = name;
    }
    
    /**
     * @return A collection of co-ordinates paired with lists of shapes occupying them.
     */
    public HashMap getShapesByCoordinate()
    {
        return shapesByCoord;
    }
    
    /**
     * @return Returns the attachers.
     */
    public ArrayList getAttachers()
    {
        return attachers;
    }
    
    /**
	 * Computes and returns (in the paramteres) the co-ordinates of the bounding box of this file.
	 * 
	 * @param topLeft Method makes this equal to the coord of the top left corner of the file's bounding box.
	 * @param bottomRight Method makes this equal to the coord of the bottom right corner of the file's bounding box.
	 */
	public void getBoundingBox(SIntPair topLeft, SIntPair bottomRight, Collection shapes)
	{
		if (shapes.size() == 0)
		   	return;
					
		// set these to the most extreme values they could possibly be
		// (e.g. top left := lowermost, rightmost coord of file (!))
	    SIntPair topLeftSoFar = new SIntPair(Integer.MAX_VALUE, Integer.MAX_VALUE);
	    SIntPair bottomRightSoFar = new SIntPair(Integer.MIN_VALUE, Integer.MIN_VALUE);
	    
	    // these two are working variables
	    SIntPair currentTopLeft = new SIntPair(0, 0);
	    SIntPair currentBottomRight = new SIntPair(0, 0); 
	    
	    // loop through shapes, updating topLeftSoFar and bottomRightSoFar as appropriate
	    Iterator i = shapes.iterator();
	    while (i.hasNext())
	    {
	        SShape s = (SShape) i.next();
	        s.getBoundingBox(currentTopLeft, currentBottomRight);
	        
	        // (N.B. y-coords go down the page)
	        if (currentTopLeft.x < topLeftSoFar.x) topLeftSoFar.x = currentTopLeft.x;
	        if (currentTopLeft.y < topLeftSoFar.y) topLeftSoFar.y = currentTopLeft.y;
	        if (currentBottomRight.x > bottomRightSoFar.x) bottomRightSoFar.x = currentBottomRight.x;
	        if (currentBottomRight.y > bottomRightSoFar.y) bottomRightSoFar.y = currentBottomRight.y;
	    }
	    
	    topLeft.setValues(topLeftSoFar.x, topLeftSoFar.y);
	    bottomRight.setValues(bottomRightSoFar.x, bottomRightSoFar.y);
	    
	    /* OK, that's it - nothing more to do, because we've updated the references that
		 * we were passed as parameters.
		 */
	}
    
    /**
     * @return Returns the maximum x-coordinate of the file.
     */
    public int getMax_x()
    {
        SIntPair topLeft = new SIntPair(0, 0);
        SIntPair bottomRight = new SIntPair(0, 0);
        getBoundingBox(topLeft, bottomRight, this.myShapes);
        return bottomRight.x;
    }
    
    /**
     * @return Returns the maximum y-coordinate of the file.
     */
    public int getMax_y()
    {
        SIntPair topLeft = new SIntPair(0, 0);
        SIntPair bottomRight = new SIntPair(0, 0);
        getBoundingBox(topLeft, bottomRight, this.myShapes);
        return bottomRight.y;
    }
    
    /**
     * @return Returns the minimum x-coordinate of the file.
     */
    public int getMin_x()
    {
        SIntPair topLeft = new SIntPair(0, 0);
        SIntPair bottomRight = new SIntPair(0, 0);
        getBoundingBox(topLeft, bottomRight, this.myShapes);
        return topLeft.x;
    }
    
    /**
     * @return Returns the minimum y-coordinate of the file.
     */
    public int getMin_y()
    {
        SIntPair topLeft = new SIntPair(0, 0);
        SIntPair bottomRight = new SIntPair(0, 0);
        getBoundingBox(topLeft, bottomRight, this.myShapes);
        return topLeft.y;
    }
    
    //
    // FUNCTIONS FOR TESTING PURPOSES ONLY
    //
    
    /**
     * Collection of shapes that make up the bounding box.
     * For testing purposes only.
     */
    Collection boundingBoxShapes;
    
    /**
     * Draw the smallest box that encloses the picture.
     * For testing purposes only.
     */
    public void drawBoundingBox()
    {
        deleteBoundingBox();
        
        SIntPair originalCursorPos = new SIntPair(cursorPosition.x, cursorPosition.y);
        
        // bottom left
        setCursorPosition(getMin_x(), getMax_y());
        SPoint bottomLeft = createPoint();
        
        // bottom right
        setCursorPosition(getMax_x(), getMax_y());
        SPoint bottomRight = createPoint();
        
        // top left
        setCursorPosition(getMin_x(), getMin_y());
        SPoint topLeft = createPoint();
        
        // top right
        setCursorPosition(getMax_x(), getMin_y());
        SPoint topRight = createPoint();
        
        // now create the lines
        SLine left = createLine(bottomLeft, topLeft);
        SLine right = createLine(bottomRight, topRight);
        SLine top = createLine(topLeft, topRight);
        SLine bottom = createLine(bottomLeft, bottomRight);
        
        // add these shapes to the collection of bounding box shapes
        boundingBoxShapes = new ArrayList();
        boundingBoxShapes.add(bottomLeft);
        boundingBoxShapes.add(bottomRight);
        boundingBoxShapes.add(topLeft);
        boundingBoxShapes.add(topRight);
        boundingBoxShapes.add(left);
        boundingBoxShapes.add(right);
        boundingBoxShapes.add(top);
        boundingBoxShapes.add(bottom);
        
        setCursorPosition(originalCursorPos.x, originalCursorPos.y);
    }
    
    /**
     * Deletes the bounding box.
     */
    public void deleteBoundingBox()
    {
        deleteShapes(boundingBoxShapes);
    }

    /**
	 * Satisfies the constraints in this file.
	*/
    public void satisfyConstraints(){
    	//Build list of constraints
    	ArrayList cons = new ArrayList();
    	Iterator iter = myShapes.iterator();
    	while(iter.hasNext())
    	{
    		SShape possCons = (SShape) iter.next();
    		if(possCons instanceof SConstraint)
    		{
    			cons.add(possCons);
    		}
    	}
    	undoButton = SConstraintSolver.satisfyConstraints(cons, 15);
    }
	
	/**
	 * Undoes constraint satisfaction if possible.
	*/
	public void undoConstraintSatisfaction() {
		if (undoButton != null) {
			undoButton.undo();
			//undoButton = null;
		}
	}
    
}
