import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

/**
 * Sketchpad is an object-oriented drawing program. Any collection of SShape objects may be saved
 * as an SFile, and SInstance objects referring to this SFile may then be incorporated into further
 * drawings. When the SFile changes, all instances thereof automatically synchronise therewith.
 * 
 * @author Tom Craig
 * @see SFile
 */
public class SInstance extends SShape implements Cloneable, Serializable
{
    /**
     * Constructor taking file and location.
     */
    public SInstance(SFile file, SPoint location, SFile owner)
    {
        super(location.getLocation() /* handle locn is top left of instance */, owner);
        owner.deselectAll();
        
        this.file = file;
        this.location = location;
    }
    
	/**
	 * The angle of the instance to the horizontal. (angle == 0) is equivalent to (angle == MAX_ANGLE).
	 */
	private SVariable angle = new SVariable(0);

	/**
	 * The number of possible different angles to the horizontal that an instance may be at.
	 */
	public static final int MAX_ANGLE = 360;
	    
	/**
	 * Stores the page co-ordinates of the top left of the instance.
	 */
	public SPoint location = null;

	/**
	 * Reference to the file of which this SInstance is an instance. The SFile object should not be modified via this reference.
	 */
	private SFile file = null;
	
	/**
	 * An ArrayList of SPoint objects.
	 * 
	 * You absolutely MUST access this through getAttachers, because that checks they're correct
	 * and updates them if not.
	 */
	private ArrayList attachers = new ArrayList();
	
	/**
	 * This scale factor is an SVariable so that the constraints system can manipulate it.
	 * The instance is of size: <pre>scaleFactor.getValue() / 256 * (size of file)</pre> 
	 * Hence, when its value is SAME_SIZE_AS_FILE (the default), the instance is the same size as the file of which it is an instance.
	 */
	private final int SAME_SIZE_AS_FILE = 256;
	private SVariable scaleFactor = new SVariable(SAME_SIZE_AS_FILE);
	
	/**
	 * @return The instance's attachers, and its top left point.
	 */
	public Collection getPoints()
	{
	    ArrayList pts = new ArrayList();	// we'll return this
	    
	    // add the attachers
	    pts = (ArrayList) attachers.clone();
	    
	    // add the top left
	    pts.add(location);
	    
	    return pts;
	}
	
	public Collection getAttachers()
	{
	    return getAttachers(true);
	}
	
	/**
	 * Computes and gets the instance's attacher points.
	 * 
	 * @param writeToOwner Updates the owner's myShapes collection with the attachers, as well as returning them.
	 *                     Should usually be true.
	 * @return An array of points corresponding to the attachers of this file. The array does not exceed SFile.MAX_NUM_ATTACHERS in size.
	 * @see SFile
	 */
	public Collection getAttachers(boolean writeToOwner)
	{
	    ArrayList fileAttachers = file.getAttachers();
	    
	    boolean attacherConfirmed[] = new boolean[attachers.size()];
	    // initialise attacherConfirmed to all false (Java might actually do this automatically, but I don't know)
	    for (int i = 0; i < attacherConfirmed.length; i++)
	        attacherConfirmed[i] = false;
	    
	    // check attachers for consistency like so:
	    //
	    //	(1) for each file attacher
	    //			if has a parallel in instance attachers
	    //				where that parallel is attachers[i]
	    //					set attacherConfirmed[i] to true
	    //			else
	    //				add a suitable parallel to this.attachers 
	    //	(2) set n = 0 (n = "number of attachers removed")
	    //		loop through attacherConfirmed
	    //			if attacherConfirmed[i] is false
	    //				remove attachers[i - n]
	    //				delete point from owner
	    //				increment n
	    
	    // step (1)
	    for (int f = 0; f < fileAttachers.size(); f++)	// 'f' for file
	    {
	        SPoint fileAttacher = (SPoint) fileAttachers.get(f);
	        
	        // work out the co-ordinate in this instance that fileAttacher *should* correspond to
	        SIntPair correctInstanceCoord = fileToInstanceCoordTransform(
	                new SIntPair(fileAttacher.getX().value(), fileAttacher.getY().value()));
	        
	        System.out.print("File attacher at ");
	        fileAttacher.getLocation().print();
	        System.out.print(" corres. to instance attacher at ");
	        System.out.print(correctInstanceCoord.toString());
	        System.out.println();
	            
	        boolean correctInstanceCoordFound = false;
	        
	        for (int i = 0; i < attachers.size() && !correctInstanceCoordFound; i++)	// 'i' for instance
	        {
	            SPoint existingInstanceAttacher = (SPoint) attachers.get(i);
	            SIntPair existingInstanceCoord = fileToInstanceCoordTransform(
	                    new SIntPair(existingInstanceAttacher.getX().value(), existingInstanceAttacher.getY().value()));
	            
	            if (existingInstanceCoord.equals(correctInstanceCoord))
	                attacherConfirmed[i] = correctInstanceCoordFound = true;
	        }
	        
	        // add the instance coord corresponding to this file coord if it's not there
	        if (!correctInstanceCoordFound)
	        {
	            // save the owner's cursor position so we can reset it after creating the attacher
	            SIntPair savedCursorPosition = (SIntPair) owner.getCursorPosition().clone();
	            
	            System.out.print("correctInstanceCoord: ");
	            System.out.print(correctInstanceCoord.toString());
	            System.out.println();
	            
	            // create the attacher SPoint in this instance
	            owner.setCursorPosition(correctInstanceCoord.x, correctInstanceCoord.y);
	            SPoint correctAttacher = owner.createPoint(true);
	            this.attachers.add(correctAttacher);
	            
	            // register this attacher with the file
	            if (writeToOwner)
	            {
		            correctAttacher.setState(SShapeState.SHAPE_DEFAULT);	// a hack to stop commitShape merging correctAttacher with anything!
		            correctAttacher.moveShape(0 - file.getMin_x(), 0 - file.getMin_y());
		            owner.commitShape(correctAttacher);
	            }
	            
	            // restore the owner's cursor position
	            owner.setCursorPosition(savedCursorPosition.x, savedCursorPosition.y);
	        } 
	    }
	    
	    // step (2)
	    ArrayList deleteTheseFromFile = new ArrayList();
	    for (int i = 0, n = 0; i < attacherConfirmed.length; i++)
	        if (!attacherConfirmed[i])
	        {
	            deleteTheseFromFile.add(attachers.get(i - n));
	            attachers.remove(i - n);
	            n++;
	        }
	    owner.deleteShapes(deleteTheseFromFile);
	    
	    
	    // now make into an array, because that's what we defined in the interfaces so that's what the GUI team expects!!
	    return attachers;
	}
	
    /**
     * @return Returns the angle of the instance to the page.
     */
    public SVariable getAngle()
    {
        return angle;
    }
    
    /**
     * @param angle The angle to set.
     */
    public void setAngle(SVariable angle)
    {
        this.angle = angle;
    }
    
    /**
     * @return Returns the file.
     */
    public SFile getFile()
    {
        return file;
    }
    
    /**
     * @param file The file to set.
     */
    public void setFile(SFile file)
    {
        this.file = file;
    }
    
    /**
     * @return Returns the location.
     */
    public SPoint getLocation()
    {
        return location;
    }
    
    /**
     * @return Returns the scale factor as the GUI wishes to see it.
     */
    public double getScaleFactor()
    {
        double sf = ((double) scaleFactor.getValue()) / ((double) SAME_SIZE_AS_FILE);
        return sf;
    }
    
    /**
     * @param scaleFactor The scale factor to set.
     */
    public void setScaleFactor(double sf)
    {
        this.scaleFactor.setValue((int) (sf));
        owner.updateCoordinateRange(this, false);
    }
	
	/**
	 * @returns the scale factor SVariable, for use of the constraints system
	 * May not be in fact required in final version, but needed for testing.
	*/
	public SVariable getScaleFactorVariable() {
		return scaleFactor;
	}
    
    /**
     * @bug if scaleFactor > ~(SFile.whatIsUnderCursor.THRESHOLD_VALUE) because loads of intermediate
     * 		coordinates are missed out
     * @see SShape#getCoordinateRange()
     */
    public Collection getCoordinateRange()
    {
        // get the co-ordinate ranges of all of the file's shapes
        HashMap fileCoordRange = file.getShapesByCoordinate();
        
        // this will hold the co-ordinate ranges of all the instance's shapes
        // (use a HashSet so that duplicates are ignored - we may get duplicates if scaleFactor < 1,
        //  giving a one-to-many relationship between file coords and instance coords)
        HashSet instanceCoordRange = new HashSet();
        
        // to each, apply the appropriate transformation from file coords to instance coords
        Iterator i = fileCoordRange.keySet().iterator();
        while (i.hasNext())
        {
            // get the file co-ordinate i is pointing to, and the list of shapes at that coord
            SIntPair fileCoord = (SIntPair) i.next();
            
            // add the corresponding instance co-ordinate
            instanceCoordRange.add(fileToInstanceCoordTransform(fileCoord));
        }
        
        return instanceCoordRange;
    }
    
    /**
     * Given a file co-ordinate, return an instance co-ordinate.
     */
    private SIntPair fileToInstanceCoordTransform(SIntPair fileCoord)
    {
        // get the co-ordinates of the top left of both the file and the instance
        SIntPair fileTopLeft = new SIntPair(file.getMin_x(), file.getMax_y());
        SIntPair instanceTopLeft = new SIntPair(location.getX().getValue(), location.getY().getValue());
        
        // do the transformation: scale, then translate
        //	(N.B. I thought we'd need the fileTopLeft subtracted, but we don't seem to)
        SIntPair instanceCoord = new SIntPair(0, 0);
        instanceCoord.x = (int) (instanceTopLeft.x + this.getScaleFactor() * (double) (fileCoord.x/* - fileTopLeft.x*/));
        instanceCoord.y = (int) (instanceTopLeft.y + this.getScaleFactor() * (double) (fileCoord.y/* - fileTopLeft.y*/));
        
        /*System.out.print("In fileToInstanceCoordTransform, returning ");
        instanceCoord.print();
        System.out.println();*/
        
        return instanceCoord;
    }
    
    /**
     * Move the instance.
     */
    public void moveShape(int dx, int dy)
    {
        super.moveShape(dx, dy);
        
        Collection attachers = getAttachers();
        Iterator i = getAttachers().iterator();
        while (i.hasNext())
            ((SPoint) i.next()).moveShape(dx, dy);
    }
    
    /**
     * Rotate the instance.
     */
    public void rotateShape(int angle)
    {
    	System.out.println(location);
        super.rotateShape(angle);
        this.angle.setValue(angle);
    }
    
    /**
     * @see SShape#getBoundingBox(SIntPair, SIntPair)
     */
    public void getBoundingBox(SIntPair topLeft, SIntPair bottomRight)
    {
        // get the file's bounding box
        SIntPair fileTopLeft = new SIntPair(0, 0);
        SIntPair fileBottomRight = new SIntPair(0, 0);
        this.file.getBoundingBox(fileTopLeft, fileBottomRight, this.file.myShapes());
        
        // work out the instance's bounding box
        SIntPair instanceTopLeft = fileToInstanceCoordTransform(fileTopLeft);
        SIntPair instanceBottomRight = fileToInstanceCoordTransform(fileBottomRight);
        
        // now we assign to what we were passed, so that the caller sees it
        // 		(we have to assign in this manner, field by field, because plain assignment
        // 		 won't update the actual parameters we were passed.)
		int xOffset = getHandleLocation().getX().getValue();
		int yOffset = getHandleLocation().getY().getValue();
		topLeft.x=xOffset;
		topLeft.y=yOffset;
        bottomRight.x = instanceBottomRight.x-instanceTopLeft.x+xOffset;
        bottomRight.y = instanceBottomRight.y-instanceTopLeft.y+yOffset;
    }
	
	/**
	 * Returns the distance of the given position from this instance, which is considered to be box-shaped.
	*/
	public double minDistanceFromCoords(SIntPair position) {
		SIntPair topLeft = new SIntPair(0,0);
		SIntPair bottomRight = new SIntPair(0,0);
		double distance = 0;
		getBoundingBox(topLeft, bottomRight);
		if((position.x > topLeft.x)&&(position.y > topLeft.y)&&(position.x < bottomRight.x)&&(position.y < bottomRight.y))
		{
			return distance;
		}
		else
		{
			double minX = Math.min(Math.abs(topLeft.x - position.x), Math.abs(bottomRight.x - position.x));
			double minY = Math.min(Math.abs(topLeft.y - position.y), Math.abs(bottomRight.y - position.y));
			distance = Math.min(minX,minY);
		}
		return distance;
	}
    
    /**
     * @return object reference for cloned object
     */
    public Object clone()
    {
    	SInstance cln = null;
   		cln = (SInstance) super.clone();
    	
    	cln.angle = (SVariable) this.angle.clone();
    	cln.location = (SPoint) this.location.clone();
    	cln.file = this.file;
    	cln.scaleFactor = (SVariable) this.scaleFactor.clone();
    	
    	return cln;
    }
}