import java.io.*;	
import java.util.ArrayList;
import java.util.Iterator;
import java.util.*;

/**
 * The top-level class of the Data + Constraints module. Holds a collection of files, and
 * supplies services to the other modules.
 *
 * @author Amit Sarna, David Stark, Tom C
 * @version 1.0
 */
public class SLibrary
{
	/**
	 * The collection of all files in the system
	 */
	private HashMap files;
	
	public SLibrary ()
	{
		files = new HashMap();
		populateFiles();
	}
	

	/**
	 * Different files refer to each other, but in serialisation, the file object that
	 * an instance refers to will be serialised, and when the object is reloaded, a new 
	 * SFile object will be created instead of using a reference to the old one, so go 
	 * through and update the links to the master files for all of the instances in the
	 * various files
	 */
	public void updateInstances()
	{
	
		Iterator it = files.values().iterator();
		
		SShape fileshape;              // a particular shape in the file
		SInstance fileinstance;		   // a particular instance in the file	
		
		while(it.hasNext())
		{
			SFile s = (SFile) it.next(); // get the next file
			
			Collection shapes = s.myShapes();
			Iterator sit = shapes.iterator(); // iterate through the shapes in the file
			
			try {
			
				while(sit.hasNext())
				{
					fileshape =(SShape) sit.next();
				
					if(fileshape instanceof SInstance)
					{
						// update the link for this instance by loading the 
						// reference to the file object from the files collection
						fileinstance = (SInstance) fileshape;
						String fileName = fileinstance.getFile().getName();
						fileinstance.setFile(getFile(fileName));
					}
				
				}
			
			
			}
			catch(SFileNotInLibraryException E){
				// master file not found
				System.out.println("MASTER FILE NO LONGER IN LIBRARY");
			}
			
		}
		
	}
	
	/**
	 * unsaved files are not in the library so if the user has placed an instance
	 * in an unsaved file, and then closed the master file of the instance without
	 * saving changes then the instances in the unsaved file must be 
	 * updated also
	 *
	 * @param untitled - the file to update
	 * @param updated - the master file of the instance
	 */
	public void updateUntitled(SFile untitled, SFile updated)
	{
		
			SShape fileshape;              // a particular shape in the file
			SInstance fileinstance;		   // a particular instance in the file	
		
						
			Collection shapes = untitled.myShapes();
			Iterator sit = shapes.iterator(); // iterate through the shapes in the file
			
			try {
			
				while(sit.hasNext())
					{
						fileshape =(SShape) sit.next();
				
						if(fileshape instanceof SInstance)
						{
							// update the link for this instance by loading the 
							// reference to the file object from the files collection
							fileinstance = (SInstance) fileshape;
							String fileName = fileinstance.getFile().getName();
							fileinstance.setFile(getFile(fileName));
						}
				
					}
		
			}
			catch(SFileNotInLibraryException E){
				// master file not found
				System.out.println("MASTER FILE NO LONGER IN LIBRARY");
			}
			
	}
		
		
	/**	
	 * load all of the files into the library colleciton
	 */
				
	public void populateFiles()
	{
		// get list of all of the files in the library
		String [] fileNames = getLibFilesList();
		
		/** debug purposes - print list of files in library *
		for(int i = 0; i < fileNames.length; i++)
			System.out.println(fileNames[i]);
		**/	
		
		// try and open the files and then put them into a hash map with the key set to 
		// the name of the file, and the value set to an SFile object referring to the file
		for(int i = 0; i < fileNames.length; i++)
		{
			try 
			{
				files.put(fileNames[i], (openFile(fileNames[i])));
			}
			catch (Exception E)
			{
				System.out.println("Exception in populate files");
			}
	
		}
			
		/** debug purposes - print list of files opened *
		for(int i = 0; i < files.size(); i++)
		{
			System.out.println(((SFile) files.get(fileNames[i])).getName());
		}
		**/
		
		updateInstances();
	}
	
		
	//
	// FILE HANDLING
	//
	
	/**
	 * Creates a new file, and returns a reference to it.
	 * @return the new file
	 */
	public SFile newFile()
	{
	  SFile file = new SFile();
	  return file;
	}
	
	/**
	 * Saves the given file to disk. Updates all instances in all files.
	 * @param f the file to be saved to disk.
	 * @throws SFileNotInLibraryException
	 * @author Amit
	 */
	public void saveFile(SFile f) throws IOException
	{
		// serialised file saved with same name as the currently open file
		FileOutputStream fout = new FileOutputStream(getLibFileName(f));
		ObjectOutputStream sfout = new ObjectOutputStream(fout);
		sfout.writeObject(f);
		sfout.flush();
		sfout.close();
		
		// add it to library if it isn't already there
		
		/*if(!files.containsKey(f.getName()))
		  {
		   	files.put(f.getName(), f);
		   	System.out.println("ADDED NEW FILE: " + f.getName());
		  }
		  */
		  
		files.put(f.getName(), f);
	
	    // debug
	    // System.out.println("FILES: " + files.size());
	 			
	}
	
	/**
	 * Saves all files to disk. (In preparation for shutdown.)
	 */
	/*
	public void saveAllFiles()
	{
	    Iterator i = files.keySet().iterator();
	    while (i.hasNext())
	    {
	        try { saveFile((SFile) (files.get(((String) i.next())))); }
	        
	        catch (Exception E) {}
	    }
	}
	*/
	
	/**
	 * Removes the given file from the library permanently. (!)
	 * @param f the file to be removed
	 * @throws SFileNotInLibraryException
	 * @author Amit
	 */
	public void removeFile(SFile remFile) throws IOException
	{
		
		Iterator i = files.keySet().iterator();
		// go through the file
	    while (i.hasNext())
	    {
	        SFile temp = (SFile) (files.get(((String) i.next())));
	        
	        if(temp == remFile) continue; // if this is the file to remove skip it
	        
	        // go through shapes in file
	        Iterator it = temp.myShapes().iterator();
	        
	        while(it.hasNext())
	        {
	        	
	        	ArrayList delShapes = new ArrayList();
	        	
	        	// if contains instance of file to delete, remove it
	        	SShape fshape = (SShape) it.next();
	        	if(fshape instanceof SInstance)
	        	{
	        		if((((SInstance) fshape).getFile()) == remFile) 
	        		{
	        			delShapes.add(fshape);
	        		}
	        	}
	        	
	        	if (!delShapes.isEmpty())	
	        	    temp.deleteShapes(delShapes);
	        	
   	        }
	           		    
		          
	        
	    }
		
		files.remove(remFile.getName());
		
		//		Delete the file from the hard drive
		
		
		File delFile = new File(getLibFileName(remFile));
		// if cannot delete file on hard drive for some reason
		 if(!delFile.delete())
			throw new IOException();
			
	}
	
	
	// given a file name, return the file by reading it from the library files collectin
	public SFile getFile(String name) throws SFileNotInLibraryException
	{
		if (!(files.containsKey(name))) throw new SFileNotInLibraryException();
		else return (SFile) files.get(name);
	}
	
	
	/**
	 * Opens the file. That is, the currentlyOpenFile reference is changed to the passed file.
	 * @param name The name of the file to be opened.
	 * @throws SFileNotInLibraryException
	 * @throws SFileAlreadyOpenException
	 * @author Amit
	 */
	public SFile openFile(String name) throws SFileNotInLibraryException, SFileAlreadyOpenException, Exception
	{
		FileInputStream fin = new FileInputStream(getLibFileName(name));
	    ObjectInputStream sfin = new ObjectInputStream(fin);
	    SFile loadedFile = (SFile) sfin.readObject();
		sfin.close();
		return loadedFile;
	}
	

	/**
	 * @return path to library file for the file object that is passed
	 * @author Amit
	 */
	public String getLibFileName(SFile f)
	{
		
		File currDir = new File("."); 
		String path = currDir.getAbsolutePath(); // get path to program
					 
		path = (path.substring(0, path.length() -1)); // remove traling .
	    path = path + f.getName() + ".spf"; // add in full file name
	    
	    return path;
	}
	
	/**
	 * @param fn the name of the file 
	 * @return the path at which the file object should reside
	 */
	public String getLibFileName(String fn)
	{
	   File currDir = new File("."); // get path to program dir
	   String path = currDir.getAbsolutePath();
	   return path.substring(0, path.length() - 1) + checkExt(fn);
	}
	
	/**
	 * check if a file name has already been passed with a .spf extension, if not, add it
	 * @return file name with .spf extension
	 * @param string to check 
	 */
	
	public String checkExt(String fn)
	{
		System.out.println(fn);
	
		if(fn.endsWith(".spf")) return fn;
		else return fn + ".spf";
		
	}
	
	/**
	 * @return array of strings of files in the current directory (for populating lib panel)
	 */
	public static String[] getLibFilesList()
	{
		  File progDir = new File("."); // Get program directory

		  // set up filter for only sketchpad files
		  FilenameFilter filter = new FilenameFilter() {
			  	public boolean accept(File dir, String name) {
		  		return name.endsWith(".spf");
		  		}
		  };

		  // get a list of files with required extension
		  String spFiles[] = progDir.list(filter);

		  // get rid of .spf on the end of flie names
		  for(int i = 0; i < spFiles.length; i++)
			  spFiles[i] = spFiles[i].substring(0, spFiles[i].length()-4);

		  return spFiles;
	}
	
	public void updateFiles(SFile f)
	{
		files.put(f.getName(), f);
		updateInstances();
	}

	
}
