import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;

/**
 * Represents a line. Is defined by two endpoints.
 *
 * @author Tom Craig
 */
public class SLine extends SShape implements Cloneable, Serializable
{
	/**
	 * A reference to one of the points defining the line.
	 */
	public SPoint end_a;

	/**
	 * A reference to the other of the points defining the line.
	 */
	public SPoint end_b;
	
	/**
	 * Default constructor.
	 */
	public SLine(SPoint end_a, SPoint end_b, SFile owner)
	{
	          // handle location is midpoint of line
	    super(new SCoords((end_a.getX().value() + end_b.getX().value()) / 2,
	            		  (end_a.getY().value() + end_b.getY().value()) / 2),
	          owner);
	    
	    this.end_a = end_a;
	    this.end_b = end_b;
	}


	public boolean hasEnd(int x, int y)
	{
		if(((end_a.getX().value() == x) && (end_a.getY().value() == y)) ||
		   ((end_b.getX().value() == x) && (end_b.getY().value() == y)))
		return true;
		
		return false;
	}


	/**
	 * @see SShape#getBoundingBox(SIntPair, SIntPair)
	 */
	public void getBoundingBox(SIntPair topLeft, SIntPair bottomRight)
	{
	    // top left is point of min x and min y
	    topLeft.x = (end_a.getLocation().getX().value() < end_b.getLocation().getX().value())
	    		   ? end_a.getLocation().getX().value() : end_b.getLocation().getX().value();
	    topLeft.y = (end_a.getLocation().getY().value() < end_b.getLocation().getY().value())
	    		   ? end_a.getLocation().getY().value() : end_b.getLocation().getY().value();
	    		   
	    // bottom right is point of max x and max y
	    bottomRight.x = (end_a.getLocation().getX().value() > end_b.getLocation().getX().value())
	    		       ? end_a.getLocation().getX().value() : end_b.getLocation().getX().value();
	    bottomRight.y = (end_a.getLocation().getY().value() > end_b.getLocation().getY().value())
	    		       ? end_a.getLocation().getY().value() : end_b.getLocation().getY().value();
	}
	
	/**
	 * Distance of the indicated position to the line.
	*/
	public double minDistanceFromCoords(SIntPair position) {
		int x1 = end_a.getX().value();
		int y1 = end_a.getY().value();
		int x2 = end_b.getX().value();
		int y2 = end_b.getY().value();
		int xp = position.x;
		int yp = position.y;
		
		//calculate the distance in between p1 and p2 in both coordinates, and as an absolute distance
		int p1top2x = x2 - x1;
		int p1top2y = y2 - y1;
		double p1top2d = java.lang.Math.sqrt(p1top2x*p1top2x + p1top2y*p1top2y);
		
		//also, the distance in between p1 and pp in both coordinates
		int p1toppx = xp - x1;
		int p1toppy = yp - y1;
		
		//project the vector in between p1 and pp onto the vector in between p1 and p2
		double projd = (double) (p1top2x * p1toppx + p1top2y * p1toppy) / p1top2d;
		
		//from this we can calculate the point-to-line distance parallel to the line
		double alongdist;
		if (projd < 0) {
			alongdist = projd * (-1); //before the 1st point
		} else {
			if ((projd - p1top2d) > 0) {
				alongdist = projd - p1top2d; //after the 2nd point
			} else {
				alongdist = 0; //alongside the line
			}
		}
		
		//now calculate the position of the projection of pp onto p1-to-p2
		double projx = x1 + projd / p1top2d * p1top2x;
		double projy = y1 + projd / p1top2d * p1top2y;
		
		//at which point we can calculate the distance perpendicular to the line
		double projtoppx = (double) xp - projx;
		double projtoppy = (double) yp - projy;
		double perdist = java.lang.Math.sqrt(projtoppx*projtoppx + projtoppy*projtoppy);
		
		//at which point we can calculate the absolute distance, which is the combined parallel and perpendicular distance
		return java.lang.Math.sqrt(perdist*perdist + alongdist*alongdist);
	}
	
	/**
	 * @return A list of co-ordinates occupied by this line.
	 */
	public Collection getCoordinateRange()
	{
	    ArrayList coordRange = new ArrayList();		// we'll return this
	    
	    // TODO? Optimise for special case: horiz or vert line
	    
	    /* Work out gradient, m, of the line and call its absolute floored value fm.
	     * If m < 1, i.e. fm = 0, then we'll step along the x-axis to get our values, which means some y-coords will appear twice.
	     * If m >= 1, i.e. fm > 0, then we'll step along the y-axis, which means some x-coords will appear twice (unless m = 1).
	     */
	    int dx = end_b.getX().value() - end_a.getX().value();
	    int dy = end_b.getY().value() - end_a.getY().value();
	    double m = ((double) dy) / ((double) dx);
	    int fm = (dx == 0 ? Integer.MAX_VALUE : Math.abs(dy / dx));
	    
	    /* Now, if m < 1 then get our x-values by x = a + (1/m)(y - b),
	     * or if m >= 1 then we get our y-values by y = b + m(x - a),
	     * where (a,b) is one of the points on the line (wlog, end_a).
	     * 
	     * So let's say we get our p-values by p = c + n(q - d),
	     * where we assign q, c, d and n as appropriate.
	     */
	    double n = (fm > 0 ? 1.0 / m : m);
	    int c = (fm > 0 ? end_a.getX().value() : end_a.getY().value());
	    int d = (fm > 0 ? end_a.getY().value() : end_a.getX().value());
	    int start = (fm > 0 ? end_a.getY().value() : end_a.getX().value());
	    int end = (fm > 0 ? end_b.getY().value() : end_b.getX().value());
	    
	    // we're going to step forwards, so make sure start <= end
	    if (start > end)
	    {
	        int tmp = start;
	        start = end;
	        end = tmp;
	    }
	    
	    // obtain the co-ordinates (by stepping through the q-values and working out the p-values)
	    for (int q = start; q <= end; q++)
	    {
	        int p = c + (int) (n * (double) (q - d));
	        int x = (fm > 0 ? p : q);
	        int y = (fm > 0 ? q : p);
	        coordRange.add(new SIntPair(x, y));
	    }
	    
	    return coordRange;
	}
	
	/**
	 * @return A list of co-ordinates occupied by this shape.
	 */
	public Collection getPoints()
	{
	    ArrayList pts = new ArrayList();
	    pts.add(end_a);
	    pts.add(end_b);
	    return pts;
	}
	
	/**
	 * @return Both ends.
	 */
	public SPoint[] getEnds()
	{
	    SPoint ends[] = {end_a, end_b};
	    return ends;
	}
	
	/**
	 * Specify the ends of the arc.
	 * @param end_a One of the ends.
	 * @param end_b The other end.
	 */
	public void setEnds(SPoint end_a, SPoint end_b)
	{
	    this.end_a = end_a;
	    this.end_b = end_b;
	}
	
	/**
    * @returns object reference for cloned object
    */
    public Object clone()
    {
    	System.out.println("CLONING LINE");
    	
    	SLine cln = null;
   		cln = (SLine) super.clone();
    	
    	cln.end_a = (SPoint) this.end_a.clone();
    	System.out.println("CLONED a");
    	cln.end_b = (SPoint) this.end_b.clone();
    	System.out.println("CLONED b");
    	
    	return cln;
    }
}