import java.io.Serializable;
import java.lang.reflect.Field;
import java.util.Collection;
import java.util.Iterator;

/**
 * SShape is an abstract class from which inherit all classes representing physical elements
 * of a drawing, such as lines, arcs and instances.
 * 
 * @author Tom Craig
 */
public abstract class SShape implements Cloneable, Serializable
{
    /**
     * Constructor.
     */
    public SShape(SCoords handleLocation, SFile owner)
    {
        this.handleLocation = handleLocation;
        this.owner = owner;
    }
    
	/**
	 * Stores whether the shape has been modified since it was last rendered. This is so
	 * that the renderer need not waste time by redrawing unmodified shapes.
	 * 
	 * Default to true because defaulting to false is dangerous.
	 *
	 * N.B. We need the renderer to call a clearModifiedFlags() function once it has rendered but before the user modifies anything more. This is annoying because it implies serial execution (i.e. first do all rendering, then be told stuff by GUI).
	 */
	private boolean modified = true;
    
    /**
     * The file owning the shape. Under some circumstances this is needed (particularly in SInstance).
     * The owner is never changed and never needs to be known outside any class extending SShape, so we don't provide a getter or setter. 
     */
    public SFile owner;

	/**
	 * Stores the state of the object (for example, unselected, being moved, etc.).
	 * These states are mutually exclusive.
	 * 
	 * @see SShapeState
	 */
	private SShapeState state;

	/**
	 * Stores the location of the shape handle in page co-ordinates. Whenever the user moves
	 * within a certain distance of the shape's handle, the renderer highlights it in a different
	 * colour and it may be selected.
	 */
	protected SCoords handleLocation;

	/**
	 * The renderer calls this and only displays the handle if the return value is less than
	 * a certain (perhaps user-determined) threshold.
	 * 
	 * @param cursorPosition The cursor position in page co-ordinates.
	 * @return The distance of the shape handle from the cursor.
	 */
	//public int handleDistanceFromCursor(SIntPair cursorPosition);
	
	/**
	 * An abstract method.
	 * @return A collection of all points belonging to the shape.
	 */
	public abstract Collection getPoints();
	
	/**
	 * An abstract method.
	 * @return A list of co-ordinates occupied by this shape.
	 */
	public abstract Collection getCoordinateRange();
	
	/**
	 * An abstract method. Returns co-ordinates of the bounding box of this shape.
	 * 
	 * @param topLeft Method makes this equal to the coord of the top left corner of the shape's bounding box.
	 * @param bottomRight Method makes this equal to the coord of the bottom right corner of the shape's bounding box.
	 */
	public abstract void getBoundingBox(SIntPair topLeft, SIntPair bottomRight);
	
	/**
	 * Find the point on the shape which is the minimum distance from the co-ordinates supplied as
	 * an argument, and return that distance.
	 * 
	 * @param coords The co-ordinates from which we wish to compute distance.
	 * @return The minimum distance from this shape.
	 */
	public abstract double minDistanceFromCoords(SIntPair position);
	
	//
	// ACCESSORS AND MUTATORS
	//
	
	/**
	 * The renderer calls this to determine whether it needs to redraw the shape.  
	 * 
	 * @return Whether the shape has been modified since it was last rendered.
	 */
	public boolean getModified()
	{
	    return modified;
	}
	
	/**
	 * This is so that the renderer need not waste time redrawing unmodified shapes.
	 * The renderer is responsible for calling this function.
	 * 
	 * @param modified Whether the shape has been modified since it was last rendered.
	 */
	public void setModified(boolean modified)
	{
	    this.modified = modified;
	}
	
	/**
	 * The renderer calls this so that it can render the shape in a certain
	 * colour based upon the shape's state.
	 * 
	 * @return A constant representing the state of the shape.
	 * @see SShapeState
	 */
	public SShapeState getState()
	{
	    return state;
	}
	
	/**
	 * Set the state of the shape. This overrides the previous state, as states are mutually exclusive.
	 * 
	 * @param state A constant representing the state of the shape.
	 * @see SShapeState
	 */
	public void setState(SShapeState state)
	{
	    this.state = state;
	}
	
	/**
	 * @return The co-ordinates of the shape's handle.
	 */
	public SCoords getHandleLocation()
	{
	    return handleLocation;
	}
	
	public void replacePoints(SPoint toReplace, SPoint replacement)
	{
		Field[] ourFields = this.getClass().getDeclaredFields();
		for (int i=0;i<ourFields.length;i++)
		{
			try{
				if(ourFields[i].get(this)==toReplace)
				{
					ourFields[i].set(this,replacement);
				}
			}
			catch (Exception e)
			{
				System.out.println(e);
			}
			
		}
	}
	
	/**
	 * Move the shape by (dx, dy).
	 */
	public void moveShape(int dx, int dy)
	{
		// update the shape's handle location
		SVariable hndx = handleLocation.getX();
	    SVariable hndy = handleLocation.getY();
	    hndx.setValue(hndx.getValue() + dx);
	    hndy.setValue(hndy.getValue() + dy);
	    //System.out.println("moved handle");
	    
	    // update the defining points of the shape
	    Collection pts = getPoints();
	    Iterator i = pts.iterator();
	    while (i.hasNext())
	    {
	        SPoint p = (SPoint) i.next();
	        if (p.getX()!=handleLocation.getX()){
		        SVariable x = p.getX();
		        SVariable y = p.getY();
		        x.setValue(x.getValue() + dx);
		        y.setValue(y.getValue() + dy);
	        }
	    }
	}
	
	/**
	 * Move the shape to (x, y).
	 */
	public void moveShapeAbs(int x, int y)
	{
	    SVariable hndx = handleLocation.getX();
	    SVariable hndy = handleLocation.getY();
	    int dx = x - hndx.getValue();
	    int dy = y - hndy.getValue();
	    
	    // N.B. it's important that moveShape is called, because it's overridden in some classes
	    //      to do special things (e.g. SInstance) and moveShapeAbs isn't
	    moveShape(dx, dy);
	}
	
	/**
	 * Rotate the shape about its handle.
	 * (N.B. The handle is not always the most intuitive place to rotate around, but it is general
	 *       because every shape has a handle, and the user can always translate after rotation
	 *       to get the desired effect.)
	 */
	public void rotateShape(int angle)
	{

	    SVariable hndx = handleLocation.getX();
	    SVariable hndy = handleLocation.getY();
	   
	    double angleInRadians = (double) angle / (double) SInstance.MAX_ANGLE * 2 * Math.PI;
	    double cosT = Math.cos(angleInRadians);
	    double sinT = Math.sin(angleInRadians);
	    
	    // apply the following transformation to each point:
	    // (x, y) := rot * (x - hndx, y - hndy) + (hndx, hndy)
	    Iterator i = getPoints().iterator();
	    while (i.hasNext())
	    {
	        SPoint p = (SPoint) i.next();
	        
            SVariable x = p.getX();
	        SVariable y = p.getY();
	        
	        // work out the new values
	        double xnew = (double) (x.value() - hndx.value()) * cosT - (double) (y.value() - hndy.value()) * sinT + (double) hndx.value();
	        double ynew = (double) (x.value() - hndx.value()) * sinT + (double) (y.value() - hndy.value()) * cosT + (double) hndy.value();
	        
	        x.setValue((int) xnew);
	        y.setValue((int) ynew);
	    }
	}
	
	/**
    * @returns object reference for cloned object
    */
    public Object clone()
    {
    	SShape cln = null;
    	try {
    		cln = (SShape) super.clone();
    	}
    	catch (CloneNotSupportedException E) {
    		// SConstraint has a clone method so should never reach here
    		System.out.println("Unexpected Error Cloning SShape");
    	}
    	
    	// both belong to the current file
    	cln.owner = this.owner;
    	// clone handle location
    	cln.handleLocation = (SCoords) this.handleLocation.clone();
    	// give object the same state
    	cln.state = (SShapeState) this.state.clone();
    	
    	return cln;
    }
}