import java.awt.Color;
import java.io.Serializable;

/**
 * Each SShape object has a state field of type SShapeState that refers to one of the static fields
 * defined below. The renderer needs to know this information so that it can apply different colours
 * to shapes in the various states. The states are mutually exclusive.
 *
 * @author Tom Craig
 * @see SShape
 */
public class SShapeState implements Serializable, Cloneable
{
	/**
	 * The colour used to render shapes in this state.
	 */
	private Color shapeColor;
	
	/**
	 * A short description of the state, used for inspecting debug output.
	 */
	private String stateDescription;

    /**
     * This is a shape's default state.
     */
    public static final SShapeState SHAPE_DEFAULT = new SShapeState(Color.black, "def");

    /**
     * A shape has this state when being created (that is, before it has been committed to the canvas).
     * Once it has been created, the shape's state becomes {@link #SHAPE_DEFAULT SHAPE_DEFAULT}.
     */
	public static final SShapeState SHAPE_BEING_CREATED = new SShapeState(Color.green, "create");

	/**
	 * A shape has this state when it is about to be merged with another of the same type.
	 */
	public static final SShapeState SHAPE_BEING_SNAPPED_TO = new SShapeState(Color.orange, "snap");
	
	/**
	 * A shape has this state when it is being translated, rotated or resized.
	 */
	public static final SShapeState SHAPE_BEING_MOVED = new SShapeState(Color.blue, "mov");

	/**
	 * A shape has this state when it is selected but not being moved or created.
	 */
	public static final SShapeState SHAPE_IS_SELECTED = new SShapeState(Color.cyan, "sel");

	/**
	 * A shape has this state when it has been deleted but not yet garbage-collected.
	 */
	public static final SShapeState SHAPE_HAS_BEEN_DELETED = new SShapeState(Color.gray, "del");
	
	/**
	 * A shape has this state when the cursor is over it such that, if the mouse is clicked, then the shape will be selected.
	 */
	public static final SShapeState SHAPE_IS_BEING_HOVERED_OVER = new SShapeState(Color.red, "hov");
	
	/**
	 * This constructor is private so that no more SShapeStates may be defined.
	 */
	private SShapeState(Color shapeColor, String stateDescription)
	{
		setShapeColor(shapeColor);
		this.stateDescription = stateDescription;
	}

	/**
	 * @param The new colour to be used to render shapes in this state.
	 */
	public void setShapeColor(Color shapeColor)
	{
		this.shapeColor = shapeColor;
	}

	/**
	 * @return The colour used to render shapes in this state.
	 */
	public Color getShapeColor()
	{
		return shapeColor;
	}
	
	public String toString()
	{
	    return stateDescription;
	}
	
	/**
    * @returns object reference for cloned object
    */
    public Object clone()
    {
    	Object cln = null;
    	try {
    		cln = super.clone();
    	}
    	catch (CloneNotSupportedException E) {
    		// SConstraint has a clone method so should never reach here
    		System.out.println("Unexpected Error Cloning SCDistanceFactor");
    	}
    	
    	return cln;
    }
}