import javax.swing.*;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.lang.Integer;
import java.awt.image.*;
import java.awt.print.*;
import java.io.*;
import com.sun.image.codec.jpeg.*;

/*
 * Created on 27-Jan-2005
 *
 */

/**
 * The main GUI class of Sketchpad
 * Gets input from the user in the form of mouse movement, clicks and keyboard input
 * and uses SFile, SLibrary, SDrawSketch and the constraints system to manipulate 
 * shapes drawn to the screen on the SCanvas 
 * 
 * @author Matthew Brown, Andrew Smee and Amit Sarna
 *
 */


public class SketchUI extends JFrame 
{
	
	private static final boolean DISPLAY_DEBUGFRAME = false;
	
	/*
	 * All the major components to be put into the GUI
	 */
	
	private JTabbedPane tabPane;
	private SCanvas canvas;
	private SLibrary library;
	private SToolTips toolTip;
	private SToolBar toolBar;
	private SLibraryPanel lp;
	private SDebugFrame debugFrame;
	
	private JMenuBar menu_bar;
	private JCheckBoxMenuItem show_constraints;
	private JCheckBoxMenuItem show_points;
	private JCheckBoxMenuItem show_axis;
	
	private JScrollBar hScroll;
	private JScrollBar vScroll;
		
	private int numberOfClicksWithTool = 0;	
	private SPoint centrePoint; // for arcs
	private SPoint startPoint;
	private SPoint endPoint;
	private SPoint intStart;
	private boolean firstLine;
	private ArrayList copiedShapes;
	private ArrayList deletedShapes;
	private HashMap shapesBeingMoved;	// SShape objects stored along with the difference between
										// their handle location and the cursor position
		
	private Box c;
	
	
	//settings to be passed to rendering and d&c
	
	private int CURSOR_X;	
	private int CURSOR_Y;	
	private SIntPair newCoords;
	
	
	private String TOOL_SELECTED = "Cursor";
	private String CURRENT_TAB = "Untitled1";
	private static SFile CURRENT_FILE;
	private Class CONSTRAINT_SELECTED;
	private ArrayList OPEN_FILES;
	
	//internal settings
	
	private int hScrollValue;
	private int vScrollValue;
	
	private int leewayX;
	private int leewayY;
	private int offsetX;
	private int offsetY;
	private SIntPair dragLocation;
	//##TB 2-3-05 Made these constants
	private final int wspaceHeight = 25600;
	private final int wspaceWidth = 38400;
	
	private boolean dragging;
	private boolean rightClick;
	private int newPageNumber = 1;
	private boolean selectMore = false;
	public static boolean showConstraints = false;
	private boolean dismemberPoint = false;
	
	
	JMenuItem save_as;
	
	// Objects used for double buffering purposes
	BufferedImage offscreen;
	Graphics buffer;	
	    
    public static HashMap ConstraintsMap = new HashMap(17);
        
    /**
     * Mapping of constraint names to their classes    
     */  
	public void initialiseCMap()
    {
       	ConstraintsMap.put("SCCollinear", SCCollinear.class);
       	ConstraintsMap.put("SCParallel", SCParallel.class);
		ConstraintsMap.put("SCPerpendicular", SCPerpendicular.class);
		ConstraintsMap.put("SCDistanceFactor", SCDistanceFactor.class);
		ConstraintsMap.put("SCErect", SCErect.class);
		ConstraintsMap.put("SCFixedDistance", SCFixedDistance.class);
		ConstraintsMap.put("SCInstanceIsOriginalSize", SCInstanceIsOriginalSize.class);
		ConstraintsMap.put("SCInstanceIsXTimesOriginal", SCInstanceIsXTimesOriginal.class);
		ConstraintsMap.put("SCInstanceIsXTimesSmallerThanOriginal", SCInstanceIsXTimesSmallerThanOriginal.class);
		ConstraintsMap.put("SCInstanceScaleFactorRelation", SCInstanceScaleFactorRelation.class);
		ConstraintsMap.put("SCOnItsSide", SCOnItsSide.class);
		ConstraintsMap.put("SCSameLength", SCSameLength.class);
		ConstraintsMap.put("SCParallel", SCParallel.class);
		ConstraintsMap.put("SCSameX", SCSameX.class);
		ConstraintsMap.put("SCSameY", SCSameY.class);
		ConstraintsMap.put("SCScaleFactor", SCScaleFactor.class);
		ConstraintsMap.put("SCSetDistance", SCSetDistance.class);
		ConstraintsMap.put("SCSetSize", SCSetSize.class);	
		ConstraintsMap.put("SCVarIsDistance", SCVarIsDistance.class);			
		ConstraintsMap.put("SCSameDistance", SCSameDistance.class);			
    }


	/**
	 * most of the work done in this class is in the constructor,
	 * arranging other components, and adding functionality with
	 * mouse listeners 
	 */
	
	public SketchUI() {
		super("Sketchpad");

		initialiseCMap();
		
		Container cp = getContentPane();
		canvas = new SCanvas(this);
		canvas.setBackground(new Color(0xFFFFFF));
		canvas.setFocusable(true);
		canvas.setDoubleBuffered(true);
		
		tabPane = new JTabbedPane(JTabbedPane.TOP);
		library = new SLibrary();
		toolTip = new SToolTips();
		toolBar = new SToolBar(toolTip, this);
		lp = new SLibraryPanel(library,this);
		
		hScroll = new JScrollBar(Adjustable.HORIZONTAL);
		vScroll = new JScrollBar(Adjustable.VERTICAL);
		
		hScroll.setMaximum(1024);
		vScroll.setMaximum(1024);
		hScroll.setValue(512);
		vScroll.setValue(512);
		setZoomPosition();
		
		copiedShapes = new ArrayList();
		deletedShapes = new ArrayList();
		shapesBeingMoved = new HashMap();
		
		OPEN_FILES = new ArrayList();
		SDrawSketch.setPageCoordinates(new SCoords (200,200));
					
		
		/*
		 * stops the application running if the close button is pressed at 
		 * the top of the window 
		 */
		this.addWindowListener(new WindowAdapter() 
		{
			public void windowClosing(WindowEvent evt) 
			{
				quitApplication();
			}
		});
		
		
		/*
		 * tells filesystem when a different drawing should
		 * be displayed
		 * will also call refresh on the canvas when a tab is clicked on 
		 */
		tabPane.addMouseListener(new MouseAdapter()
		{
				public void mousePressed(MouseEvent m) 
				{
					System.out.println("page selected = " + tabPane.getTitleAt(tabPane.getSelectedIndex()) + ", refresh the drawing");
					CURRENT_TAB = tabPane.getTitleAt(tabPane.getSelectedIndex());
					CURRENT_FILE = (SFile)OPEN_FILES.get(tabPane.getSelectedIndex());
					updateDisplay(true, false);
				}
		});

		//tells file system when drawing is scrolled horizontally, giving
		//scroll amount.  also calls refresh on canvas 	
		hScroll.addAdjustmentListener(new AdjustmentListener()
		{
			public void adjustmentValueChanged(AdjustmentEvent e)
			{
				setZoomPosition();
			}
		});
		
		//as above, for vertical scroll
		vScroll.addAdjustmentListener(new AdjustmentListener()
		{
			public void adjustmentValueChanged(AdjustmentEvent e)
			{
				setZoomPosition();
			}
		});
		
				
		//arranging the GUI using nested containers
				
		//box keeping the v-scroll bar in the right place
		Box v = new Box(BoxLayout.Y_AXIS);
			v.setSize(new Dimension(10, 400));
			v.setMinimumSize(new Dimension(10, 400));
			v.add(Box.createVerticalStrut(24));
			v.add(vScroll);
			v.add(Box.createVerticalStrut(18));

		Box c = new Box(BoxLayout.Y_AXIS);	
			c.setPreferredSize(new Dimension(400, 400));
			c.setBorder(BorderFactory.createMatteBorder(1,1,1,1, new Color(0x666666)));
			c.add(canvas, BorderLayout.WEST);
			c.add(hScroll, BorderLayout.SOUTH);
	
		//box containing the canvas, the tabs box, and the h-scroll bar
		Box m = new Box(BoxLayout.Y_AXIS);	
			m.setSize(new Dimension(450, 375));
			m.add(tabPane, BorderLayout.NORTH);
			m.add(c, BorderLayout.SOUTH);

		//another box, keeping the tabs, canvas and scroll bars all together
	    Box l = new Box(BoxLayout.X_AXIS);
			l.add(m,BorderLayout.WEST);
			l.add(v, BorderLayout.EAST);
		
		Box lib = new Box(BoxLayout.Y_AXIS);
			lib.add(Box.createVerticalStrut(5));
			lib.add(new JLabel("Library Files"));
			lib.add(lp);
			
		toolTip.setPreferredSize(new Dimension(400, 50));
		
		tabPane.setPreferredSize(new Dimension(500, 25));
		tabPane.setMaximumSize(new Dimension(500, 25));
		tabPane.setMinimumSize(new Dimension(500, 25));

		canvas.setPreferredSize(new Dimension(450, 400));	
		
		//adding everything to the main container
		cp.add(toolBar, BorderLayout.WEST);
		cp.add(lib, BorderLayout.EAST);
		cp.add(l, BorderLayout.CENTER);
		cp.add(toolTip, BorderLayout.SOUTH);
		
				
//menu system 
		menu_bar = new JMenuBar();
			
		/*
		 * menu headings
		 */
			JMenu file_menu = new JMenu("File");
			JMenu view_menu = new JMenu("View");	
			JMenu edit_menu = new JMenu("Edit");
			JMenu help_menu = new JMenu("Help");
				
			/*
			 * 'file' menu items
			 */
				JMenuItem exit = new JMenuItem ("Exit");
					exit.addActionListener(new ActionListener() 
					{
						public void actionPerformed(ActionEvent evt) 
						{
							quitApplication();
						}
					});
					exit.setToolTipText("Exits Sketchpad");
					exit.setMnemonic(KeyEvent.VK_Q);

				JMenuItem open = new JMenuItem("Open");
					open.addActionListener(new ActionListener()
					{
						public void actionPerformed(ActionEvent evt) 
						{
							openFile();
						}
					});
					open.setToolTipText("Open a file to modify it");
					open.setMnemonic(KeyEvent.VK_O);

				JMenuItem new_page = new JMenuItem("New");
					new_page.addActionListener(new ActionListener()
					{
						public void actionPerformed(ActionEvent evt) 
						{
							newFile();
						}
					});
					new_page.setToolTipText("Opens a new blank page");
					new_page.setMnemonic(KeyEvent.VK_N);

				JMenuItem close_page = new JMenuItem("Close");
					close_page.addActionListener(new ActionListener()
					{
						public void actionPerformed(ActionEvent evt) 
						{
							removePage();
						}
					});
					close_page.setToolTipText("Close a page");
					close_page.setMnemonic(KeyEvent.VK_C);

				JMenuItem print = new JMenuItem("Print");
					print.addActionListener(new ActionListener()
					{
						public void actionPerformed(ActionEvent evt) 
						{
							printDisplay();
						}
					});
					print.setMnemonic(KeyEvent.VK_P);
					print.setToolTipText("Print current Drawing");
					
				JMenuItem save = new JMenuItem("Save");
					save.addActionListener(new ActionListener()
					{
						public void actionPerformed(ActionEvent evt)
						{
							save();
							
						}
					});
					save.setMnemonic(KeyEvent.VK_S);
					save.setToolTipText("Save current Drawing");
		
					
				//JMenuItem
				save_as = new JMenuItem("Save As...");
					save_as.addActionListener(new ActionListener()
					{
						public void actionPerformed(ActionEvent evt)
						{
							saveAs();
							lp.updateList();
						}
					});
					save_as.setMnemonic(KeyEvent.VK_V);
					save_as.setToolTipText("Save current Drawing");
					
			/*	JMenuItem save_all = new JMenuItem("Save All");
					save_all.addActionListener(new ActionListener()
					{
						public void actionPerformed(ActionEvent evt)
						{
							saveAll();
						}
					});
					save_all.setMnemonic(KeyEvent.VK_A);
					save_all.setToolTipText("Save all open Drawings");
			*/
					
					
			/*
			 * 'view' menu items
			 */
				JMenuItem zoom_in = new JMenuItem("Zoom In");
					zoom_in.addActionListener(new ActionListener()
					{
						public void actionPerformed(ActionEvent evt) 
						{
							SDrawSketch.ZoomIn();
							setZoomPosition();
						}
					});
					zoom_in.setMnemonic(KeyEvent.VK_I);
					zoom_in.setToolTipText("Zoom in on Drawing");
			
				JMenuItem zoom_out = new JMenuItem("Zoom Out");
					zoom_out.addActionListener(new ActionListener()
					{
						public void actionPerformed(ActionEvent evt) 
						{
							SDrawSketch.ZoomOut();
							setZoomPosition();
						}
					});
					zoom_out.setToolTipText("Zoom out on Drawing");
			
				JMenuItem satisfy_constraints = new JMenuItem("Satisfy Constraints");
					satisfy_constraints.addActionListener(new ActionListener()
					{
						public void actionPerformed(ActionEvent evt) 
						{
							CURRENT_FILE.satisfyConstraints();
							updateDisplay(showConstraints, false);							
						}
					});
					satisfy_constraints.setToolTipText("Satisfy constraints in the drawing");
					
				JMenuItem undo_satisfy_constraints = new JMenuItem("Undo Constraint Satisfaction");
					undo_satisfy_constraints.addActionListener(new ActionListener()
					{
						public void actionPerformed(ActionEvent evt) 
						{
							CURRENT_FILE.undoConstraintSatisfaction();
							updateDisplay(showConstraints, false);							
						}
					});
					undo_satisfy_constraints.setToolTipText("Undo the changes constraint satisfaction made");
					
				show_constraints = new JCheckBoxMenuItem("Show constraints");
					show_constraints.addActionListener(new ActionListener()
					{
						public void actionPerformed(ActionEvent evt) 
						{
							showConstraints = !showConstraints;
							
							System.out.println("show constraints");
							updateDisplay(showConstraints, false);							
						}
					});
					show_constraints.setToolTipText("Show constraints on Drawing");
					
				show_axis = new JCheckBoxMenuItem("Show Axis");
					show_axis.addActionListener(new ActionListener()
					{
						public void actionPerformed(ActionEvent evt) 
						{
							SDrawSketch.setShowAxes(!SDrawSketch.getShowAxes());
							updateDisplay(showConstraints, false);							
						}
					});
					show_axis.setToolTipText("Show constraints on Drawing");
			
			/*
			 *	Edit menu items
			 */
				JMenuItem copy = new JMenuItem("Copy");
					copy.addActionListener(new ActionListener()
					{
						public void actionPerformed(ActionEvent evt) 
						{
							copiedShapes.clear();
							ArrayList selectedShapes = CURRENT_FILE.getSelectedShapes();
							for (int i = 0; i < selectedShapes.size(); i++)
							{
								copiedShapes.add(selectedShapes.get(i));
							}
							
							SIntPair tl = new SIntPair(0,0);
							SIntPair br = new SIntPair(0,0);
			
							CURRENT_FILE.getBoundingBox(tl,br,copiedShapes);
							
														
							System.out.println("TOP LEFT: " + tl.x + " " + tl.y);
							
							
							CURRENT_FILE.setTopLeftOfCopyBox(tl);
														
							System.out.println("copy shapes");
							
							
							copiedShapes = SFile.preliminaryCopy(copiedShapes);
							
							
							updateDisplay(showConstraints, false);
						}
					});
				copy.setToolTipText("Copy selected shapes");
				copy.setMnemonic(KeyEvent.VK_C);
				
				JMenuItem paste = new JMenuItem("Paste");
					paste.addActionListener(new ActionListener()
					{
						public void actionPerformed(ActionEvent evt) 
						{
							CURRENT_FILE.copyAndPasteTheseShapes(copiedShapes);
							System.out.println("paste shapes");
							updateDisplay(showConstraints, false);
						}
					});
				paste.setToolTipText("Paste copied shapes");
				paste.setMnemonic(KeyEvent.VK_V);
				
				JMenuItem del = new JMenuItem("Delete");
				del.addActionListener(new ActionListener()
				{
					public void actionPerformed(ActionEvent evt) 
					{
						CURRENT_FILE.deleteShapes(CURRENT_FILE.getSelectedShapes());
						System.out.println("delete shapes");
						updateDisplay(showConstraints, false);
					}
				});
				del.setToolTipText("Delete selected shapes");
				
				show_points = new JCheckBoxMenuItem("Show points");
				show_points.addActionListener(new ActionListener()
				{
					public void actionPerformed(ActionEvent evt) 
					{
						SDrawSketch.setShowPoints(!SDrawSketch.getShowPoints());
						updateDisplay(showConstraints, false);
					}
				});
				show_constraints.setToolTipText("Show points on Drawing");
				

				JMenuItem hide_points = new JMenuItem("Hide points");
				hide_points.addActionListener(new ActionListener()
				{
					public void actionPerformed(ActionEvent evt) 
					{
						SDrawSketch.setShowPoints(false);
						System.out.println("hide points");
						updateDisplay(showConstraints, false);
					}
				});
				hide_points.setToolTipText("Hide points on Drawing");

				JMenuItem dismember = new JMenuItem("Dismember");
				dismember.addActionListener(new ActionListener()
				{
					public void actionPerformed(ActionEvent evt) 
					{
						toolBar.resetTool();
						dismemberPoint = true;
						updateDisplay(showConstraints, false);
					}
				});
				dismember.setToolTipText("Separate merged points from each other");
				
				
				JMenuItem about = new JMenuItem("About");
				about.addActionListener(new ActionListener()
				{
					public void actionPerformed(ActionEvent evt) 
					{
						JOptionPane.showMessageDialog(new JFrame(), "Created as Part IB group project, 2005 by \n \n Tom Botterill, Matthew Brown, Tom Craig,\n Easlyn Kirupairajah, Amit Sarna,\n Andrew Smee & David Stark", "About Sketchpad", JOptionPane.INFORMATION_MESSAGE, null);
						System.out.println("about sketchpad");
						updateDisplay(showConstraints, false);
					}
				});
				about.setToolTipText("About Sketchpad");
				about.setMnemonic(KeyEvent.VK_F1);
			
			
			
			file_menu.add(new_page);
			file_menu.add(open);	
			file_menu.add(save);
			file_menu.add(save_as);
			file_menu.add(new JSeparator());
			file_menu.add(close_page);
			file_menu.add(print);	
			file_menu.add(new JSeparator());
			file_menu.add(exit);
			
			file_menu.setMnemonic(KeyEvent.VK_F);
			
			
			view_menu.add(zoom_in);
			view_menu.add(zoom_out);
			view_menu.add(new JSeparator());
			view_menu.add(satisfy_constraints);
			view_menu.add(undo_satisfy_constraints);
			view_menu.add(new JSeparator());
			view_menu.add(show_constraints);
			view_menu.add(show_points);
			view_menu.add(show_axis);
			view_menu.add(new JSeparator());
			view_menu.add(dismember);
			
			edit_menu.add(copy);
			edit_menu.add(paste);
			edit_menu.add(del);
			
			
			help_menu.add(about);
			
			
								
		menu_bar.add(file_menu);
		menu_bar.add(edit_menu);
		menu_bar.add(view_menu);
		menu_bar.add(help_menu);
		
		
		setJMenuBar(menu_bar);
				
		pack();
		
		// initialise buffering objects
		System.out.println(canvas.getWidth() + "askdfj;lasdjflkajsldf");
		System.out.println(canvas.getHeight() + "dsakfa;sdjfajsdf");
		
		offscreen = (BufferedImage) createImage(canvas.getWidth(), canvas.getHeight());
  	    buffer = offscreen.getGraphics();
  	    
  	    
  	    // nasty hack! we have to do this because SLabel.getBoundingBox and SNumber.getBoundingBox
  	    // require a Graphics object in order to work out their font metrics
  	    SLabel.g = buffer;
  	    SNumber.g = buffer;
  	    
		if(DISPLAY_DEBUGFRAME) debugFrame = new SDebugFrame();
	}
		
		
		
//private methods start here
	
	/**
	 * Updates the screen display with double buffering
	 *
	 * @param showconstraints - whether or not to show the constraints
	 * @param isthumbnail - if the image to display is a thumbnail
	 */
	private void updateDisplay(boolean showconstraints, boolean isthumbnail)
	{
		show_constraints.setState(showconstraints);
		show_points.setState(SDrawSketch.getShowPoints());
		show_axis.setState(SDrawSketch.getShowAxes());
		
		// draw the sketch offscreen
		SDrawSketch.drawSketch(buffer, CURRENT_FILE, showconstraints, isthumbnail);
		// render the image to the canvas
		canvas.getGraphics().drawImage(offscreen, 0, 0, this);
		menu_bar.setSelected(canvas);
				
		// update the instance thumbanil
		lp.ip.updateThumb();
	}
	
	
	protected void mouseClick(MouseEvent m,boolean rightClick)
	{
		if(!menu_bar.isSelected()){
			numberOfClicksWithTool++;
			drawingActionPerformed(m.getX(), m.getY(), rightClick, true);
			setScroll();
			setZoomPosition();
			updateShapeInfo();
		}
		menu_bar.setSelected(canvas);
		setZoomPosition();
		updateShapeInfo();
		canvas.grabFocus();
	}
	
	protected void mouseDrag(MouseEvent m)
	{
		setCursorPosition(m);
		if (TOOL_SELECTED.equals("Navigate workspace"))
		{
			setZoomPosition(dragLocation.x-CURSOR_X,dragLocation.y-CURSOR_Y);
			dragLocation = new SIntPair(CURSOR_X,CURSOR_Y);	
		}
		else{
			drawingActionPerformed(m.getX(), m.getY(), false, false);
		}

	}
	
	protected void mouseMove(MouseEvent m)
	{
		setCursorPosition(m);
			
		if (CURRENT_FILE != null&&!menu_bar.isSelected())
		{
			//System.out.println("Under the cursor: " + CURRENT_FILE.whatIsUnderCursor());
			drawingActionPerformed(m.getX(), m.getY(), false, false);
		}
		updateShapeInfo();
	}
	
	protected void selectMore(boolean selectMore){
		this.selectMore = selectMore;
	}
	
	/**
	 *	try to satisfy all constraints
	 */
	protected void doConstraints(){
		CURRENT_FILE.satisfyConstraints();
		updateDisplay(showConstraints,false);
	}
	
	/**
	 * delete the currently selected shapes
	 */	 
	protected void deleteShapes(){
		CURRENT_FILE.deleteShapes(CURRENT_FILE.getSelectedShapes());
		updateDisplay(showConstraints,false);
	}
	
	
	private void checkForInstances(SFile dropping, SFile target) throws SCyclicInstancesException
	{
		Iterator it = dropping.myShapes().iterator();
		while(it.hasNext())
		{
			SShape s = (SShape) it.next();
			
			if (s instanceof SInstance)
			{
				SInstance si = (SInstance) s;
				if(si.getFile().getName().equals(target.getName())) throw new SCyclicInstancesException();
				else checkForInstances(si.getFile(), target);
			}
			
		}
	}
	
	
	/**
	 * Add the file as an instance in the currently open file as long as this
	 * will not cause a cycle
	 * @param x and y - the coords of the drop operation
	 * @param file - a string which is the name of the file
	 **/
	 
	protected void fileDropped(int x, int y, String file) throws SCyclicInstancesException
	{
		setCursorPosition(x,y);
		// The user has dragged a file name from the library panel 
		// onto the actual screen area, so draw this to the 
		// screen at the required location
					
		try {
						
			SFile f = library.getFile(file);
						
			System.out.println(f.getName());
			System.out.println(CURRENT_FILE.getName());
		
			// before creating the instance in the file, go through the file and check to see
			// if either the file is the same file as the instance, or if the file is contained by
			// the instance to prevent endless cycling
			if(CURRENT_FILE.getName() != null)
			 {
			  	if(f.getName().equals(CURRENT_FILE.getName())) throw new SCyclicInstancesException();
								
				// loop thorough and check if instance to be dropped contains CURRENT_FILE
				checkForInstances(f, CURRENT_FILE);
			 }
			 
			 					
			// insert file into this file, at current mouse position (i.e. where dropped)
			SInstance si = CURRENT_FILE.createInstance(f);
									
			CURRENT_FILE.commitShape(si);
								
			updateDisplay(false, false);
		}
		catch(SCyclicInstancesException e)
		{
			JOptionPane.showMessageDialog(new JFrame(), "Cannot drop this instance in this file as it would cause a cycle.", "Error", JOptionPane.ERROR_MESSAGE, null);
		}
				
		catch(Exception E)
		{
			JOptionPane.showMessageDialog(new JFrame(), "Sketchpad was unable to open the file for dropping.", "Error", JOptionPane.ERROR_MESSAGE, null);
		}
		
	}
	
	
	/**
	 * ends program properly
	 */
	private void quitApplication() 
	{
		System.exit(0);
	}
	
	/** 
	 * converts mouse coordinate on screen to actual drawing coordinate
	 * and updates this position to rendering and file system
	 * 
	 * @param m - mouse event that triggers update of these cursor coords 
	 */

	public void setCursorPosition(MouseEvent m){
		CURSOR_X = m.getX();
		CURSOR_Y = m.getY();
		setCursorPosition();
	}
	public void setCursorPosition (int x,int y)
	{
		CURSOR_X = x;
		CURSOR_Y = y;
		setCursorPosition();
	}
	public void setCursorPosition()
	{
		SCoords scrCoords = SDrawSketch.scrToPageCoord_new(new SIntPair(CURSOR_X,CURSOR_Y));
		newCoords = new SIntPair(scrCoords.getX().getValue(),scrCoords.getY().getValue());
		toolTip.setText("(" + newCoords.x + ", " + newCoords.y + ")");
		if(CURRENT_FILE!=null){
			CURRENT_FILE.setCursorPosition(newCoords.x, newCoords.y);
			newCoords = CURRENT_FILE.getCursorPosition();
		}
	}

	
	private void setZoomPosition(){
	
		// leewayX and leewayY represent the available movement
		leewayX = wspaceWidth - SDrawSketch.getZoom()*canvas.getWidth();
		leewayY = wspaceHeight - SDrawSketch.getZoom()*canvas.getHeight();
		
		if(!canvas.isDragging())
		{
			hScrollValue = hScroll.getValue();
			vScrollValue = vScroll.getValue();
			
			//##TB 2-3-05
			offsetX = wspaceWidth/2-(hScrollValue*leewayX)/1014;
			offsetY = wspaceHeight/2-(vScrollValue*leewayY)/1014;
			//offsetX = wspaceWidth/2-(hScrollValue*leewayX)/1014;
			//offsetY = wspaceHeight/2-(vScrollValue*leewayY)/1014;
			//I'm sure we don'r need a setScroll here, I've tested it without and it seems to work ok
			// setScroll();
		}else{
			//TC System.out.println("Scroll override");
		}

		if((TOOL_SELECTED.equals("Zoom In")||TOOL_SELECTED.equals("Zoom Out"))&&canvas.isDragging())
		{
			offsetX = -newCoords.x+SDrawSketch.getZoom()*canvas.getWidth()/2;
			offsetY = -newCoords.y+SDrawSketch.getZoom()*canvas.getHeight()/2;
			SDrawSketch.setPageCoordinates(new SCoords (offsetX,offsetY));
			setScroll();
		}
		else
		{
			SDrawSketch.setPageCoordinates(new SCoords (offsetX,offsetY));
		}
		
		if(canvas.getGraphics()!=null){
			updateDisplay(showConstraints, false);
		}
	}
	
	private void setZoomPosition(int x, int y)
	{
		if(canvas.isDragging())
		{
			offsetX -= x*SDrawSketch.getZoom();
			offsetY -= y*SDrawSketch.getZoom();
			setZoomPosition();
		}
	}
	private void setScroll(){
	    //##TB 2-3-05
		hScrollValue = 1+(1014*(wspaceWidth/2-offsetX))/leewayX;
		vScrollValue = 1+(1014*(wspaceHeight/2-offsetY))/leewayY;
    
//		hScrollValue = 1+(1014*(wspaceWidth-offsetX))/leewayX;
//		vScrollValue = 1+(1014*(wspaceHeight-offsetY))/leewayY;
		hScroll.setValue(hScrollValue);
		vScroll.setValue(vScrollValue);
	}
	
	/**
	 * add a tab for a new page
	 * 
	 * @param filename - the title of the tab to be added
	 */

	private void setCurrent (SFile newFile)
	{
				CURRENT_FILE = newFile;
		
	
		
		/*	
		if(newFile.getName() != null)
		   this.save_as.setEnabled(false);
		   
		else
		   this.save_as.setEnabled(true);
		*/
		
	}

	/**
	 * adds a new tab and initialises the associated file for drawing into
	 * 
	 * @param filename
	 */
	private void addPage(String filename)
	{
		OPEN_FILES.add(CURRENT_FILE);
		tabPane.addTab(filename, null, Box.createVerticalStrut(0), "Click to view " + filename);
		tabPane.setSelectedIndex(tabPane.getTabCount()-1);
		
		tabPane.setPreferredSize(new Dimension(tabPane.getWidth(), ((tabPane.getTabRunCount() - 1)*19) + 25));
		tabPane.setMinimumSize(new Dimension(tabPane.getWidth(), ((tabPane.getTabRunCount() - 1)*19) + 25));
		tabPane.setMaximumSize(new Dimension(tabPane.getWidth(), ((tabPane.getTabRunCount() - 1)*19) + 25));
		canvas.setPreferredSize(new Dimension(canvas.getWidth(), 387 - (tabPane.getTabRunCount()-1)*19));
		canvas.setMaximumSize(new Dimension(canvas.getWidth(), 387 - (tabPane.getTabRunCount()-1)*19));
		canvas.setMinimumSize(new Dimension(canvas.getWidth(), 387 - (tabPane.getTabRunCount()-1)*19));
		
		setZoomPosition();
	}
	
	/**
	 * close currently selected page - remove its tab - brings up a dialog to confirm
	 */
	private void removePage()
	{
		String filename = tabPane.getTitleAt(tabPane.getSelectedIndex());
						
		if (OPEN_FILES.size() == 1)
		{
			JOptionPane.showMessageDialog(new JFrame(), 
										"Cannot close this page.\nYou must have one file open at all times.", "Forbidden", JOptionPane.INFORMATION_MESSAGE, null);
			return;
		}
		
		int check = JOptionPane.showConfirmDialog(this,"Do you want to save "+filename+" before closing?");
		
		if(check!=JOptionPane.CANCEL_OPTION){
		
			if(check==JOptionPane.NO_OPTION){
				
				if(CURRENT_FILE.getName() != null) 
				{
						
				// this file might be a master file so discarding the changes means that instances
				// of this file in other files need to be changed back to when the file was last saved
				try {
					
				   // don't save this file to the library and update instances in all other files
				   SFile old = library.openFile(CURRENT_FILE.getName());
				   
				   library.updateFiles(old);
				   
				   
				   // update untitled files
				   Iterator it = OPEN_FILES.iterator();
				   
				   
				   while(it.hasNext())
				   {
				   		SFile toUpdate = (SFile)it.next();
				   		
				   		// udpate the file 
				   		if(toUpdate.getName() == null)
				   		{
				   			library.updateUntitled(toUpdate, old);			   		   		
				   		}
				   }
					    
				}
				
				catch(Exception e)
				{
				}
			
	 		}				
							

			}else if (check==JOptionPane.YES_OPTION){
				save();
				filename = tabPane.getTitleAt(tabPane.getSelectedIndex());
			}
			try
			{
				System.out.println("OF" + OPEN_FILES.size());
				OPEN_FILES.remove(tabPane.indexOfTab(filename));
				System.out.println("OF" + OPEN_FILES.size());
				tabPane.remove(tabPane.indexOfTab(filename));
				
				tabPane.setSelectedIndex(0);
				// go back to first page
				CURRENT_FILE = (SFile) OPEN_FILES.get(0);
				
			}
			catch(IndexOutOfBoundsException i)
			{
				JOptionPane.showMessageDialog(new JFrame(), "File not open", "Error", JOptionPane.ERROR_MESSAGE, null);
			}
			
			tabPane.setPreferredSize(new Dimension(tabPane.getWidth(), ((tabPane.getTabRunCount() - 1)*19) + 25));
			tabPane.setMinimumSize(new Dimension(tabPane.getWidth(), ((tabPane.getTabRunCount() - 1)*19) + 25));
			tabPane.setMaximumSize(new Dimension(tabPane.getWidth(), ((tabPane.getTabRunCount() - 1)*19) + 25));
			canvas.setPreferredSize(new Dimension(canvas.getWidth(), 385 - (tabPane.getTabRunCount()-1)*19));
			canvas.setMaximumSize(new Dimension(canvas.getWidth(), 385 - (tabPane.getTabRunCount()-1)*19));
			canvas.setMinimumSize(new Dimension(canvas.getWidth(), 385 - (tabPane.getTabRunCount()-1)*19));
		
			setZoomPosition();
		}
		
	}
	
	
	/**
	 * @returns a reference to the currently open file
	 */
	public static SFile getCurrFile()
	{
		return CURRENT_FILE;
	}
	
	/**
	 * creates a new file
	 */
	private void newFile()
	{
		SFile f = library.newFile();
		setCurrent(f);
		addPage("Untitled" + newPageNumber);
		newPageNumber++;
	}
	/**
	 * opens a saved file for editing
	 * 
	 * @param filename - the name of the file to be opened
	 */
	protected void openFile(String filename){
		toolBar.resetTool();
		TOOL_SELECTED = "Cursor";

		if (filename!=null && filename.trim()!= "")
		{			
			try 
			{
				SFile newFile = library.getFile(filename);
												
				if(OPEN_FILES.contains(newFile)) throw new SFileAlreadyOpenException(filename);
												
				setCurrent(newFile);
				addPage(newFile.getName());
			}
			catch (SFileAlreadyOpenException e)
			{
				System.out.println(e.toString());
				System.out.println("EXCEPTION HERE");
				
				for(int i = 0; i < OPEN_FILES.size(); i++)
				  System.out.println((String) (((SFile)OPEN_FILES.get(i)).getName()));
			}
			catch(Exception e)
			{
				JOptionPane.showMessageDialog(new JFrame(), "Error opening file " + filename, "Error", JOptionPane.ERROR_MESSAGE);
				System.out.println("not in library" + e);
			}
			
		}
	}
	
	/**
	 * open a saved file for modification - NOT AN INSTANCE OF THE FILE.  
	 * opens a dialog to get the name of the file to open
	 */
	private void openFile()
	{
		String filename = (String) JOptionPane.showInputDialog(new JFrame(), "Enter Filename", "Open", JOptionPane.PLAIN_MESSAGE, null, null, null);
		openFile(filename);
	}
	
	/**
	 * Prints the image to a JPEG file - originally printed straight to the printer but there
	 * were problemns with image quality and amount of canvas that was displayed so instead used
	 * printing to file which allows user to use the drawing how he/she wants
	 */
	private void printDisplay()
	{
		// get the image currently drawn on screen
		BufferedImage toPrint = new BufferedImage(canvas.getWidth(), canvas.getHeight(), BufferedImage.TYPE_INT_RGB) ;
		Graphics2D buf = (Graphics2D) toPrint.getGraphics();
		SDrawSketch.drawSketch(buf, CURRENT_FILE, false, false);
		
		// draw an outline around the image
		buf.setColor(Color.black);
		buf.drawRect(0, 0, canvas.getWidth() - 1, canvas.getHeight() - 1);
		
		
		// save the image to disk with its filename and a jpg extension
		saveImageToDisk(toPrint, tabPane.getTitleAt(tabPane.getSelectedIndex()) + ".jpg");
		
		// refresh the display
		updateDisplay(showConstraints, false);
		
		
		/*PrinterJob pj = PrinterJob.getPrinterJob();
		
		pj.setPrintable(new Printable() {
		
		public int print(Graphics g, PageFormat pf, int pageIndex) {
				if (pageIndex > 0) return Printable.NO_SUCH_PAGE;
				Graphics2D g2d = (Graphics2D)g;
				g2d.translate(pf.getImageableX(), pf.getImageableY());
  			    SDrawSketch.drawSketch(g2d, CURRENT_FILE, false,false);
  			    

				return Printable.PAGE_EXISTS;
			}
		});

  	    if(pj.printDialog())
	    try {
			pj.print();
	    }
	    catch(PrinterException pe) {
			System.out.println("Error printing:" + pe);
	    }*/
    

	}
	
	
	/**
	 * save current file
	 */
	private void save()
	{
	    CURRENT_FILE.deselectAll();
	    
		System.out.println("Saving");
		String fileName = CURRENT_FILE.getName();
		if(fileName==null) {
			fileName = (String) JOptionPane.showInputDialog(new JFrame(), "Enter Filename", "Save", JOptionPane.PLAIN_MESSAGE, null, null, null);
			if(fileName!=null&&!fileName.equals("")){
				
				CURRENT_FILE.setName(fileName);
				tabPane.setTitleAt(tabPane.getSelectedIndex(), fileName);
				
				
			}
		}
		if(fileName!=null&&!fileName.equals("")){
			try
			{				
				library.saveFile(CURRENT_FILE);
			}
			catch (Exception e)
			{
				JOptionPane.showMessageDialog(new JFrame(), "Could not save file", "Error", JOptionPane.ERROR_MESSAGE, null);
			}
		}
		
		lp.ip.updateEntries();
	}
	/**
	 * save the currently open  file with a different filename from its current one 
	 */
	private void saveAs()
	{
		CURRENT_FILE.setName(null);
		save();
		
		if(CURRENT_FILE.getName() == null)
		   return;
		
		String filename = CURRENT_FILE.getName();
		tabPane.setTitleAt(tabPane.getSelectedIndex(), filename);
		tabPane.setToolTipTextAt(tabPane.getSelectedIndex(), "Click to view " + filename);
					
	}


	
	/**
	 * Shape debugging tool
	 * 
	 */
	private void updateShapeInfo(){
		if(CURRENT_FILE!=null){
			String shapeInfo = "";
			shapeInfo = "File top left: (" + CURRENT_FILE.getMin_x() + "," + CURRENT_FILE.getMin_y() + ")\n";
			int shapeCount = 0;
			Collection refs = CURRENT_FILE.myShapes();
			Iterator iter = refs.iterator();
			while(iter.hasNext())
			{
				SShape x = (SShape) iter.next();
				shapeInfo += "\n" + x.toString()
				          + ", state=" + x.getState().toString()
				          + (CURRENT_FILE.getAttachers().contains(x) ? ", is att" : "")
				          + ", pos=(" + x.getHandleLocation().getX().getValue() + "," + x.getHandleLocation().getY().getValue() + ")";
				shapeCount ++;
	
			}
			if(DISPLAY_DEBUGFRAME) debugFrame.setText(shapeInfo);
		}
	}
	
	/**	
	 *  test which tool is in use, then do the required action
	 *  also tests how many mouse clicks have been performed, for 
	 *  actions like drawing a line or an arc
	 * 
	 * @param x - the x coordinate of the action
	 * @param y - the y coordinate of the action
	 * @param rightClick - boolean which is true if the captured mouse action was a right-click
	 * @param clicked - boolean which is true if the captured mouse action was a click
	 */
	private void drawingActionPerformed(int x, int y,boolean rightClick, boolean clicked)
	{	
		if (TOOL_SELECTED.equals("Cursor"))
		{
			if (dismemberPoint)	dismember(x, y, clicked);
			select(x, y, clicked);
		}
		else if (TOOL_SELECTED.equals("Move Item"))
		{		
			moveItem(clicked);
		}
		else if (TOOL_SELECTED.equals("Draw Line"))
		{
			drawLine(x, y, rightClick, clicked);
		}
		else if (TOOL_SELECTED.equals("Draw Arc"))
		{
			drawArc(x, y, clicked,rightClick);
		} 			
		else if (TOOL_SELECTED.equals("Cursor"))
		{
			if (dismemberPoint)	dismember(x, y, clicked);
				
			select(x, y, clicked);
		}
		else if (TOOL_SELECTED.equals("Navigate workspace"))
		{
			dragLocation = new SIntPair(CURSOR_X,CURSOR_Y);
		}

		else if (TOOL_SELECTED.equals("Add Constraint"))
		{
			addConstraint(x, y, rightClick, clicked);
		}
		else if (TOOL_SELECTED.equals("delete"))
		{
			CURRENT_FILE.deleteShapes(CURRENT_FILE.getSelectedShapes());
		}
		else if (TOOL_SELECTED.equals("Add Text"))
		{
			addLabel(clicked);		
		}
		else if (TOOL_SELECTED.equals("Add Number Constraint")){
			addNumber(x,y,clicked);
		}		
		else if (TOOL_SELECTED.equals("Lock Point"))
		{
		    lockShape(clicked);
		}
		else if (TOOL_SELECTED.equals("Copy"))
		{
			copiedShapes.clear();
			ArrayList selectedShapes = CURRENT_FILE.getSelectedShapes();
			for (int i = 0; i < selectedShapes.size(); i++)
				{
					copiedShapes.add(selectedShapes.get(i));
				}
					
			SIntPair tl = new SIntPair(0,0);
			SIntPair br = new SIntPair(0,0);
			
			CURRENT_FILE.getBoundingBox(tl,br,copiedShapes);
							
							
					
			System.out.println("TOP LEFT: " + tl.x + " " + tl.y);
							
						
			CURRENT_FILE.setTopLeftOfCopyBox(tl);
														
			System.out.println("copy shapes");
							
							
			copiedShapes = SFile.preliminaryCopy(copiedShapes);
	
			
			System.out.println("copy shapes");

			updateDisplay(showConstraints, false);
			// get top left co-ordinate of bouding box of objects copied
			// for working out new positions of objects when pasted
		}
		else if (TOOL_SELECTED.equals("Paste"))
		{
			System.out.println(copiedShapes.size());
			if (clicked) CURRENT_FILE.copyAndPasteTheseShapes(copiedShapes);
		}	
		else if (TOOL_SELECTED.equals("Rotate Instance"))
		{
			rotate(x,y, rightClick, clicked);
		}
		else if (TOOL_SELECTED.equals("Resize Instance"))
		{
			resizeItem(rightClick, clicked);
		}
		else if (TOOL_SELECTED.equals("Set Attacher"))
		{	
		    setAttacher();
		}
		else if (TOOL_SELECTED.equals("Navigate workspace"))
		{
			dragLocation = new SIntPair(CURSOR_X,CURSOR_Y);
		}
		else if (TOOL_SELECTED.equals("Zoom In")&&clicked)
		{
			SDrawSketch.ZoomIn();
		}
		else if (TOOL_SELECTED.equals("Zoom Out")&&clicked)
		{
			SDrawSketch.ZoomOut();
		}
		
		if (!TOOL_SELECTED.equals("Navigate workspace")
				&&!TOOL_SELECTED.equals("Zoom In")
				&&!TOOL_SELECTED.equals("Zoom Out"))
		{
			setZoomPosition();
		}
	}
	
	private SLine latestLine = null;  //allows us to delete the line being created without recursively deleting lots of things
	
	/**
	 * creates lines on the canvas, starting at the first mouse click, and ending at 
	 * the second click, and starting a new line from this point
	 * 
	 * @param x
	 * @param y
	 * @param rightClick
	 * @param clicked
	 */
	private void drawLine(int x, int y, boolean rightClick, boolean clicked)
	{
		//before first click, don't make any new lines
		if(numberOfClicksWithTool==0){
			toolTip.setText("Click in on the canvas where you want the line to start.");
		}
		else if(numberOfClicksWithTool == 1 && rightClick)
		{				
		    numberOfClicksWithTool=0;
		}
		//start drawing a line on the first click
		else if (numberOfClicksWithTool == 1 && clicked)
		{
		    CURRENT_FILE.deselectAll();
			
			startPoint = CURRENT_FILE.createPoint();
			endPoint = CURRENT_FILE.createPoint();
			intStart = startPoint;

			System.out.println("current start"+startPoint);
			startPoint = (SPoint) CURRENT_FILE.commitShape(startPoint);
			System.out.println("current start"+startPoint);

			SLine newLine = CURRENT_FILE.createLine(startPoint, endPoint);
			this.latestLine = newLine;
			
			firstLine = true;
		}
		
		/*
		 * this is a separate state for fixing the start of the 
		 * line, and keeping the end attached to the cursor
		 */ 
		else if(numberOfClicksWithTool == 1)
		{
			toolTip.setText("To stop drawing lines, right click on the canvas.");
		    //TCM CURRENT_FILE.moveShapeAbs(endPoint,newCoords.x, newCoords.y);
		    endPoint.moveShape(newCoords.x - endPoint.getX().getValue(), newCoords.y - endPoint.getY().getValue());
			//TC System.out.println(startPoint + " , " + endPoint + " , " + intStart);
		}
		
		//stop drawing lines and delete the current one if a right-click is performed
		else if (numberOfClicksWithTool == 2 && rightClick)
		{
		    numberOfClicksWithTool=0;
			
		    // delete the latest line
			CURRENT_FILE.deleteShape(latestLine);
			
			endPoint = null;
			intStart = null;
			startPoint = null;
		}
		
		//otherwise the line was created properly, so complete it
		else if (numberOfClicksWithTool == 2&&clicked)
		{
		    System.out.println(startPoint + " , " + endPoint + " , " + intStart);
		    endPoint = (SPoint) CURRENT_FILE.commitShape(endPoint);
				
			intStart = startPoint;

			System.out.println(startPoint + " , " + endPoint + " , " + intStart);
			
			CURRENT_FILE.commitShape(intStart);
			
			CURRENT_FILE.commitShape(latestLine);			
			latestLine.setEnds(intStart, endPoint);
		
			// check to see if there already exists a different line with the same
			// start and end points and if so delete this one

			// get coords of new line        
	    	int x1 = latestLine.end_a.getX().value();
			int y1 = latestLine.end_a.getY().value();
			int x2 = latestLine.end_b.getX().value();
			int y2 = latestLine.end_b.getY().value(); 
	
			// check to see if a line with the same coords already exists, if not add it
			Iterator it = CURRENT_FILE.myShapes().iterator();
			SShape s;
			boolean delete = false;
			while(it.hasNext())
			{
				s = (SShape) it.next();
			
			
				if((s instanceof SLine) && (s!=latestLine))
				{
					System.out.println("Has Lines");
			
					SLine sl = (SLine) s;
					
					if(sl.hasEnd(x1, y1) && sl.hasEnd(x2,y2) && (sl.getState() != SShapeState.SHAPE_BEING_CREATED))
					 {
					 	delete = true;
 					 }
				}
				
			}
			
			// if the line does exist delete it
			if(delete) CURRENT_FILE.deleteShape(latestLine);	
		
			//making new line
			intStart = endPoint;
			startPoint = endPoint;
			endPoint = CURRENT_FILE.createPoint();
			System.out.println(startPoint + " , " + endPoint + " , " + intStart);			
			
			CURRENT_FILE.deselectAll();
			SLine newLine = CURRENT_FILE.createLine(startPoint, endPoint);
			this.latestLine = newLine;
			
			//TC System.out.println("selectedShapes size is " + selectedShapes.size());
			
			firstLine = false;
			numberOfClicksWithTool--;
		}
	}
	
	private SArc arcBeingCreated = null;
	
	/**
	 * create an arc on the canvas, with centre at the first mouse click location,
	 * starting at the second, and ending at the third mouse click location
	 * 
	 * @param x
	 * @param y
	 * @param clicked
	 * @param rightClick
	 */
	private void drawArc(int x, int y, boolean clicked,boolean rightClick)
	{
			
		if (rightClick){
			if(numberOfClicksWithTool>1){
				numberOfClicksWithTool = 0;
				
				CURRENT_FILE.deleteShape(arcBeingCreated);
				
				centrePoint = null;
				startPoint = null;
				endPoint = null;
			}	
		}
		else if (numberOfClicksWithTool == 1 && clicked)
		{
		    CURRENT_FILE.deselectAll();
			
			centrePoint = CURRENT_FILE.createPoint();
			startPoint = CURRENT_FILE.createPoint();
			endPoint = CURRENT_FILE.createPoint();
			
			centrePoint = (SPoint) CURRENT_FILE.commitShape(centrePoint);
		 
			arcBeingCreated = CURRENT_FILE.createArc(centrePoint, startPoint, endPoint, false);

			
			//System.out.println("possible new arc created with centre at " + s.getCentre() + " and going from");
			//System.out.println(s.getStart() + " to " + s.getEnd());	
		}
		
		/*
		 *it fixes the circle centre, and keeps the start and end
		 *points attached to the cursor
		 */ 
		else if(numberOfClicksWithTool == 1)
		{
			startPoint.moveShapeAbs(newCoords.x, newCoords.y);
			endPoint.moveShapeAbs(newCoords.x, newCoords.y);

			//System.out.println("possible new arc created with centre at " + arcBeingCreated.getCentre() + " and going from");

			//System.out.println(startPoint + " to " + endPoint);
		}
		
		/*this is also just a separate state.  It fixes the start 
		 *point of the arc, and leave the end poin attached to the 
		 *cursor
		 */

		else if(numberOfClicksWithTool == 2&&clicked)
		{
			startPoint = (SPoint) CURRENT_FILE.commitShape(startPoint);
			arcBeingCreated.setStart(startPoint);
		}
		else if (numberOfClicksWithTool == 2)
		{
			endPoint.moveShapeAbs(newCoords.x, newCoords.y);
			//System.out.println("possible new arc created with centre at " + arcBeingCreated.getCentre() + " and going from");
			//System.out.println(startPoint + " to " + endPoint);
		}

		//fixes the end point of the arc
		else if (numberOfClicksWithTool == 3)
		{
			endPoint = (SPoint) CURRENT_FILE.commitShape(endPoint);
			arcBeingCreated.setEnd(endPoint);
			CURRENT_FILE.commitShape(arcBeingCreated);
			CURRENT_FILE.commitShape(startPoint);
			CURRENT_FILE.commitShape(endPoint);
			CURRENT_FILE.commitShape(centrePoint);
			
			//CURRENT_FILE.deleteShapes(deletedShapes);
			deletedShapes.clear();
			
			numberOfClicksWithTool = 0;		
			
			//System.out.println("new arc created with centre at " + s.getCentre() + " and going from");
			//System.out.println(s.getStart() + " to " + s.getEnd());	
			
			CURRENT_FILE.deselectAll();
			numberOfClicksWithTool = 0;
			startPoint = null;
			endPoint = null;
			centrePoint = null;
		}
	}
	
	private void addConstraint(int x, int y, boolean rightClick, boolean clicked){
		
		int[] req = {-1,0,0,0};
		if(CONSTRAINT_SELECTED!=null)
		{
			try{
				
				//get the array of required shapes
				req = (int[])CONSTRAINT_SELECTED.getMethod("getRequired",null).invoke(CONSTRAINT_SELECTED,new Object[]{});
			}
			catch (Exception e)
			{
				System.out.println(e);
			}
		}
		System.out.println(CURRENT_FILE.whatIsUnderCursor());
		if (CURRENT_FILE.whatIsUnderCursor()!=null&&clicked) 
		{
		    CURRENT_FILE.selectShape(CURRENT_FILE.whatIsUnderCursor());
		}
		//now see how many of the shapes we have
		Iterator iter = CURRENT_FILE.getSelectedShapes().iterator();
		while(iter.hasNext())
		{
			SShape testShape = (SShape) iter.next();
			if (testShape instanceof SPoint) req[0]--;
			if (testShape instanceof SLine) req[1]--;
			if (testShape instanceof SArc) req[2]--;
			if (testShape instanceof SInstance) req[3]--;
			if (testShape instanceof SNumber) req[3]--;
		}
		String requirements="To use this constraint you must select ";
		if(req[0]>0) requirements+=req[0]+" more points,";
		if(req[1]>0) requirements+=req[1]+" more lines,";
		if(req[2]>0) requirements+=req[2]+" more arcs,";
		if(req[3]>0) requirements+=" and " +req[3]+ " more numbers or instances";
		
		if(CONSTRAINT_SELECTED!=null){
			toolTip.setText(requirements);
		}else{
			toolTip.setText("Please select a constraint from the list on the left");
		}
		if(clicked)
		{
			if (req[0]<0||req[1]<0||req[2]<0||req[3]<0){
				//must have selected wrong shapes, clear selection
				CURRENT_FILE.deselectAll();
			}
			else if (req[0]==0&&req[1]==0&&req[2]==0&&req[3]==0)
			{
				//all requirements fulfilled, add constraint
				try{
					SConstraint newCons = CURRENT_FILE.createConstraint(CURRENT_FILE.getSelectedShapes(), CONSTRAINT_SELECTED);
					CURRENT_FILE.commitShape(newCons);
					CONSTRAINT_SELECTED= null;
					toolTip.setText("Constraint applied succesfuly. Press the spacebar to satisfy all constraints.");
					toolBar.constraintsFrame.setActive(-1);
				}
				catch (Exception e)
				{
					System.out.println(e);
				}
				CURRENT_FILE.deselectAll();
				//toolBar.resetTool();
				toolTip.setText("");
			}
			System.out.println(req[0]+" "+req[1]+" "+req[2]);

		}
		numberOfClicksWithTool = 0;
		
		
	}
	
	/**
	 * selects a shape when the mouse is clicked on it, and user can select multiple
	 * shapes if SHIFT is pressed when clicking on shapes, otherwise only the latest
	 * shape that was clicked on will be selected
	 * 
	 * @param x
	 * @param y
	 * @param clicked
	 */
	private void select(int x, int y, boolean clicked)
	{
	    //ArrayList selectedShapes = CURRENT_FILE.getSelectedShapes();
	    
		/*
		 * to select a shape, the user clicks on that shape.
		 * the shape is added to the selectedShapes arrayList
		 */
		if ((clicked)&&(!selectMore))
		{
			CURRENT_FILE.deselectAll();
			CURRENT_FILE.selectShape(CURRENT_FILE.whatIsUnderCursor());
		}
		/*
		 * if SHIFT is pressed down, more shapes can be selected 
		 */
		//TC System.out.println(clicked+" "+ selectMore);
		if ((clicked)&&(selectMore))
		{
			SShape s = CURRENT_FILE.whatIsUnderCursor();
			if (s != null)
			{
				// if it's selected already, deselect it; otherwise, select it
			    if (s.getState() == SShapeState.SHAPE_IS_SELECTED)
			        CURRENT_FILE.deselectShape(s);
			    else
			        CURRENT_FILE.selectShape(s);
			}			
		}
		
		//TC System.out.println("number of items selected = " + CURRENT_FILE.getSelectedShapes().size());
	}
	
	/**
	 * adds a text label to the drawing wherever the user clicks on the canvas 
	 * 
	 * @param clicked
	 */
	private void addLabel (boolean clicked)
		{
			if (clicked)
			{
				String text = (String) JOptionPane.showInputDialog(new JFrame(), "Enter label text", "Add Text", JOptionPane.PLAIN_MESSAGE, null, null, null);
			
				SShape underCurs = CURRENT_FILE.whatIsUnderCursor();
				if (underCurs!=null && underCurs instanceof SLabel)
				{
					((SLabel)underCurs).setText(text);
				}
				else
				{
					startPoint = CURRENT_FILE.createPoint();
					startPoint = (SPoint) CURRENT_FILE.commitShape(startPoint);
					System.out.println("selecting position");
					// user selects position of number on screen
					SLabel newLabel = CURRENT_FILE.createLabel(startPoint, text);
					CURRENT_FILE.commitShape(newLabel);
					numberOfClicksWithTool = 0;
				}
			}
		}
	
	/**
	 * similar to above, but only allows a numerical label to be added
	 * 
	 * @param x
	 * @param y
	 * @param clicked
	 */
	private void addNumber (int x, int y, boolean clicked)
	{
		if (clicked)
		{
			String stringValue = (String) JOptionPane.showInputDialog(new JFrame(), "Enter value of number", "Add Number Constraint", JOptionPane.PLAIN_MESSAGE, null, null, null);
			int value = 0;
			try{
				value = Integer.parseInt(stringValue);
			}
			catch (NumberFormatException e){
				System.out.println("bad number");
			}
			
			SShape underCurs = CURRENT_FILE.whatIsUnderCursor();
			if (underCurs!=null && underCurs instanceof SNumber)
			{
				((SNumber)underCurs).setValue(value);
			}
			else
			{
				startPoint = CURRENT_FILE.createPoint();
				startPoint = (SPoint) CURRENT_FILE.commitShape(startPoint);
				System.out.println("selecting position");
				// user selects position of number on screen
				SNumber newNumber = CURRENT_FILE.createNumber(startPoint, value);
				CURRENT_FILE.commitShape(newNumber);
				numberOfClicksWithTool = 0;
			}
		}
	}

	private void setAttacher()
	{
		ArrayList selectedShapes = CURRENT_FILE.getSelectedShapes();
	    
	    Iterator i = selectedShapes.iterator();
	    while (i.hasNext())
	    {
	        SShape s = (SShape) i.next();
	        if (s instanceof SPoint)
	        {
				try
				{
					CURRENT_FILE.setAttacherState((SPoint) s, true);
				}
				catch(STooManyAttachersException e)
				{
					//what do we do here?
				    
				    System.out.println(e.toString());
				    toolTip.setText("Not all points could be made attachers. Maximum possible is " + SFile.MAX_NUM_ATTACHERS);
				}
			}
	    }
	}
	
	private void lockShape(boolean clicked)
	{
		if (numberOfClicksWithTool == 1)
		{
		    numberOfClicksWithTool = 0;
			if (CURRENT_FILE.whatIsUnderCursor() != null)
			{			    
			    SShape s = CURRENT_FILE.whatIsUnderCursor();
				if(s instanceof SPoint) ((SPoint) s).setFixed(true);
			} 
		}
	}
	
	/**
	 * user selects a shape, and moves it to another location on the canvas
	 * 
	 * @param x
	 * @param y
	 * @param clicked
	 */
	private void moveItem(boolean clicked)
	{
		//selecting shapes to move
		if (numberOfClicksWithTool == 1&&clicked)
		{
			if (CURRENT_FILE.whatIsUnderCursor() != null)
			{			    

			    Iterator i = shapesBeingMoved.keySet().iterator();
			    while (i.hasNext())
			        CURRENT_FILE.dropShape((SShape) i.next());

				SShape s = CURRENT_FILE.whatIsUnderCursor();
				s.moveShape(0, 0);	// start moving the shape
				
				// compute the offset of the cursor from the handle, and store it in shapesBeingMoved with the shape itself
				int xoff = s.getHandleLocation().getX().value() - newCoords.x;
				int yoff = s.getHandleLocation().getY().value() - newCoords.y;
				SIntPair offset = new SIntPair(xoff, yoff);
				shapesBeingMoved.put(s, offset);
			} 
		}
		
		//actually moving the selected shapes
		else if (numberOfClicksWithTool == 1)
		{
		    Iterator i = shapesBeingMoved.keySet().iterator();
		    while (i.hasNext())
		    {
		    	SShape s = (SShape) i.next();
		    	
		    	// get s's offset
		    	SIntPair offset = (SIntPair) shapesBeingMoved.get(s);
		    	if (offset == null)
		    	    offset = new SIntPair(0, 0);
		    	
				CURRENT_FILE.moveShapeAbs(s, newCoords.x + offset.x, newCoords.y + offset.y);
				
				if (s instanceof SPoint)
				{
					SPoint p = (SPoint) s;
					CURRENT_FILE.setFixedState(p, false);
				}
			}
		}
		
		else if (numberOfClicksWithTool == 2)
		{
		    Iterator i = shapesBeingMoved.keySet().iterator();
		    while (i.hasNext())
			{
				SShape s = (SShape) i.next();
				//TC s = CURRENT_FILE.commitShape(s);	-- NO! you don't need to commit s, as it's not just been created
			
				System.out.println("dropping " + s);
				CURRENT_FILE.dropShape(s);

			}
			numberOfClicksWithTool = 0;
			
			shapesBeingMoved.clear();
		}
	}

	private SInstance shapeBeingResized = null;
	private double size = 1.0;
	private void resizeItem(boolean rightClick, boolean clicked)
	{
		if (numberOfClicksWithTool == 1&&clicked)
		{
		    SShape underCursor = CURRENT_FILE.whatIsUnderCursor();
		    
			if (underCursor == null || !(underCursor instanceof SInstance))
			{
				numberOfClicksWithTool = 0;
			}
			else
			{
			    shapeBeingResized = (SInstance) underCursor;
				CURRENT_FILE.resizeShape(shapeBeingResized, 1.0);
				startPoint = shapeBeingResized.getLocation();
				SIntPair tL = new SIntPair(0,0);
				SIntPair bR = new SIntPair(0,0);
				shapeBeingResized.getBoundingBox(tL,bR);
				size = SIntPair.cartesianDistance(tL,bR);
			}
		}
		else if (numberOfClicksWithTool == 1)
		{
			
			double distance = SIntPair.cartesianDistance(new SIntPair(startPoint.getX().value(), startPoint.getY().value()), newCoords);
			System.out.println("resizing"+(distance/size));
			
			CURRENT_FILE.resizeShape(shapeBeingResized, distance);
			//size=distance;
			
		}
		else if (numberOfClicksWithTool == 2)
		{
			CURRENT_FILE.dropShape(shapeBeingResized);
			shapeBeingResized = null;
			startPoint = null;
			numberOfClicksWithTool = 0;
		}
		
	}
	
	private SShape shapeBeingRotated = null;
	private SCoords startCoords;
	/**
	 * rotates a shape selected by the user
	 * 
	 * @param x
	 * @param y
	 * @param rightClick
	 * @param clicked
	 */
	private void rotate(int x, int y,boolean rightClick, boolean clicked)
	{
		
		
		if (numberOfClicksWithTool == 1&&clicked)
		{
		    SShape underCursor = CURRENT_FILE.whatIsUnderCursor();
		    
			if (underCursor == null){
				numberOfClicksWithTool = 0;
			}
			else
			{
			    shapeBeingRotated = underCursor;
				startCoords = shapeBeingRotated.getHandleLocation();
			}
		}
		//state where we follow the cursor
		else if (numberOfClicksWithTool == 1)
		{
			CURRENT_FILE.rotateShape(shapeBeingRotated, new SIntPair(startCoords), newCoords);
		}
		
		else if (numberOfClicksWithTool == 2)
		{
			CURRENT_FILE.rotateShape(shapeBeingRotated, new SIntPair(startCoords), newCoords);
			CURRENT_FILE.dropShape(shapeBeingRotated);
			shapeBeingRotated = null;
			startPoint = null;
			numberOfClicksWithTool = 0;
		}
	}
	
	
	/**
	 * separates shapes with a common point
	 * 
	 * @param x
	 * @param y
	 * @param clicked
	 */
	private void dismember(int x, int y, boolean clicked)
	{
		if (clicked && CURRENT_FILE.whatIsUnderCursor() instanceof SPoint)
		{
			System.out.println("DISMEMBERING A POINT");
			CURRENT_FILE.dismember((SPoint) CURRENT_FILE.whatIsUnderCursor());
			updateDisplay(showConstraints, false);
			dismemberPoint = false;
		}
	}
	
		
	/**used by the file system to store scroll settings when 
	 * switching between tabs
	 */
	//or should be, but hasn't been implemented in SFile due to time constraints
	public int getHScroll()
	{
		return hScrollValue;
	}
		
	public int getVScroll()
	{
		return hScrollValue;	
	}
		
	/**used by SToolBar to tell UI which tool is being used
	 * 
	 * @param tool
	 */
	public void clickedOn(String tool)
	{
		this.TOOL_SELECTED = tool;
		numberOfClicksWithTool = 0;	
		if(TOOL_SELECTED.equals("Zoom"))
		{
			canvas.setMagnify(2);
		}
		else if(TOOL_SELECTED.equals("Zoom In"))
		{
			canvas.setMagnify(2);
		}
		else if(TOOL_SELECTED.equals("Zoom Out"))
		{
			canvas.setMagnify(1);
		}
		else
		{
			canvas.setMagnify(0);
		}
	}
	
	//select constraint
	public void constraintClickedOn(Class cons)
	{
		if(numberOfClicksWithTool==0){
			this.CONSTRAINT_SELECTED = cons;
			this.TOOL_SELECTED = "Add Constraint";
		}
	}

	/**
	 * paste an instance from the library into the drawing
	 * 
	 * @param file
	 */
	public void pasteInstance(SFile file)
	{
		try
		{
			CURRENT_FILE.createInstance(file);
		}
		catch(SFileNotInLibraryException e)
		{}	
	}
	
			
	//make rocket go now 	
	public static void main(String[] args)
	{
		int offset = -300;	
		
		try {
        	UIManager.setLookAndFeel(
        	UIManager.getCrossPlatformLookAndFeelClassName());
    	} catch (Exception e) { }
    	   	  	
    	 
    	 
    	  	
		SketchUI s = new SketchUI();		
		s.show();
		
		
		// create a default page untitled1
		s.newFile();
		
		
		SDrawSketch.setPageCoordinates(new SCoords(0,0));
				
	/*	SFile bulb = new SFile();
		bulb.setName("bulb");
		s.setCurrent(bulb);
		s.addPage("bulb");
		
		s.setResizable(false);
		
							   	
		bulb.setCursorPosition(-50 + offset, -111+ offset);
		
		SPoint bp1 = bulb.createPoint();
		bulb.commitShape(bp1);
		
		bulb.setCursorPosition(-1000+ offset, -111+ offset);
		SPoint bp2 = bulb.createPoint();
		bulb.commitShape(bp2);
		SLine bl1 = bulb.createLine(bp1, bp2);
		bulb.commitShape(bl1);
		*/
        
        /*
			   bulb.setCursorPosition(-300+ offset, -111+ offset);
			   SPoint bp3 = bulb.createPoint();
			   bulb.commitShape(bp3);
		
			   bulb.setCursorPosition(-450+ offset, -111+ offset);
			   SPoint bp4 = bulb.createPoint();
			   bulb.commitShape(bp4);
			   SLine bl2 = bulb.createLine(bp3, bp4);
			   bulb.commitShape(bl2);
        
        	   bulb.setCursorPosition(-250+ offset, -111+ offset);
			   SPoint centre = bulb.createPoint();
			   bulb.commitShape(centre);
			   SArc ba1 = bulb.createArc(centre, bp2, bp3, true);
			   bulb.commitShape(ba1);
        
        	   bulb.setCursorPosition(-175+ offset, -111+ offset);
			   SPoint bp5 = bulb.createPoint();
			   bulb.commitShape(bp5);
		
			   bulb.setCursorPosition(-325+ offset, -111+ offset);
			   SPoint bp6 = bulb.createPoint();
			   bulb.commitShape(bp6);
			   SArc ba2 = bulb.createArc(centre, bp5, bp5, true);
			   bulb.commitShape(ba2);
  

  
  		try 
		{
			bulb.setAttacherState(bp1, true);
				 bulb.setAttacherState(bp2, true);
  			  
			
			   // save the file to the library
			    s.library.saveFile(bulb);
														
							  							
	        	SFile circuit = new SFile();
				circuit.setName("circuit");
				s.setCurrent(circuit);
				s.addPage("circuit");
				
					
								
				CURRENT_FILE.setCursorPosition(-275 + offset,-400 + offset);
			  	SInstance ci1 = CURRENT_FILE.createInstance(s.library.getFile("bulb"));
			  	ci1.setScaleFactor(1);
			  	CURRENT_FILE.commitShape(ci1);
					   
			  }
			   
				

			   catch(Exception E)
			   {
			   	System.out.println("JJJJJ");
			   }
			   
			   
	*/

	}
	
	
	
  	/**
  	 * saves the passed buffered image as a JPEG to disk
  	 * 
  	 * @param image to save
  	 * @param filename to save with
  	 */ 
 
   public static void saveImageToDisk(BufferedImage bi, String str) {
    
    if (bi != null && str != null) {
 
      // save image as Jpeg
      FileOutputStream out = null;
      try {
        out = new FileOutputStream(str);
      }
      catch (java.io.FileNotFoundException fnf) {
        System.out.println("File Not Found");
      }
 	
 		
 	  // create a jpeg from the passed image	
      JPEGImageEncoder encoder = JPEGCodec.createJPEGEncoder(out);
      JPEGEncodeParam param = encoder.getDefaultJPEGEncodeParam(bi);
      param.setQuality(1.0f, false);
 
 
      try {
        encoder.encode(bi);
        out.close();
      }
      catch (java.io.IOException io) {
        System.out.println(io);
      }
    }
   }

	

	
}
